//---------------------------------------------------------------------------
// File: C:\cbproj\Remote_CW_Keyer\CircularFifo.c
// Date: 2024-02-20
// Author: Wolfgang Buescher (DL4YHF)
// Purpose: Simple, classic, lock-free, thread-safe CIRCULAR FIFO
//          with ONE WRITER (in thread A)
//          and MULTIPLE READERS (in threads B,C,D,..)
//          Used in various firmware projects, but since 2024
//          also in the 'Remote CW Keyer' (initially a windows
//          application), to pass data received from, or sent to,
//          a serial port from one thread to another
//          WITHOUT THE NEED FOR THREAD SYNCHRONISATION.
//---------------------------------------------------------------------------

#include "switches.h"  // project specific compiler switches ("options"),
                       // must be included before anything else !
#include "yhf_type.h"  // classic types like BYTE, WORD, DWORD, BOOL, ..
#include <string.h>    // no string functions used here, but memset()
#include "CircularFifo.h" // header for THIS module (lock-free circular FIFO)

//----------------------------------------------------------------------------
// Implementation of functions
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
void CFIFO_Init( T_CFIFO *pFifo )
  // If the application uses the 'parent-struct trick' to expand the
  // FIFO size, THE CALLER must adjust pFifo->dwSize (in bytes)
  // *after* the call. 
{
  memset( pFifo, 0, sizeof(T_CFIFO) );
  pFifo->iSize = SWI_CIRCULAR_FIFO_SIZE;
} // end CFIFO_Init()

//----------------------------------------------------------------------------
int CFIFO_Write( T_CFIFO *pFifo, BYTE *pbSource, int nBytes)
  // Returns the number of bytes actually written .
{
  int nBytesWritten = 0; 
  int iHeadIndex = pFifo->iHeadIndex;
  int iTailIndex = pFifo->iTailIndex;
  int iNewHeadIndex;
  while( nBytesWritten < nBytes )
   { iNewHeadIndex = iHeadIndex+1;
     if( iNewHeadIndex >= pFifo->iSize )
      {  iNewHeadIndex = 0;
      }
     if( iNewHeadIndex == iTailIndex )
      { break;  // buffer is full -> stop
      }
     pFifo->b[iHeadIndex] = *pbSource++;
     iHeadIndex = iNewHeadIndex;
     ++nBytesWritten;
   }
  pFifo->iHeadIndex = iHeadIndex;
  return nBytesWritten;
} // end CFIFO_Write()

//----------------------------------------------------------------------------
int CFIFO_Read( T_CFIFO *pFifo, BYTE *pbDest, int nBytesMax )
  // Returns the number of bytes actually read .
{
  int iHeadIndex = pFifo->iHeadIndex;
  int iTailIndex = pFifo->iTailIndex;
  int nBytesRead = 0;
  while( (nBytesRead < nBytesMax) && (iHeadIndex!=iTailIndex) )
   { *pbDest++ = pFifo->b[iTailIndex++];
     ++nBytesRead;
     if( iTailIndex >= pFifo->iSize )
      {  iTailIndex = 0;
      }
   }
  pFifo->iTailIndex = iTailIndex;
  return nBytesRead;
} // end CFIFO_Read()

//----------------------------------------------------------------------------
int CFIFO_GetNumBytesReadable( T_CFIFO *pFifo )
  // Returns the number of bytes currently available in the FIFO.
{
  int iResult = pFifo->iHeadIndex - pFifo->iTailIndex;
  if( iResult < 0 )
   {  iResult += pFifo->iSize;
   }
  return iResult;
} // end CFIFO_GetNumBytesReadable()

/* EOF < CircularFifo.c > */

