// File:  C:\cbproj\Remote_CW_Keyer\CwDSP.c
// Date:  2023-11-22
// Author:  Wolfgang Buescher (DL4YHF)
// Purpose: 'CW Digital Signal Processor' - initially just a sidetone generator,
//          later also a DSP with sample rate conversions and compression
//          for low-bandwidth audio streaming via network.
//

#include "switches.h" // project specific 'compilation switches' like SWI_USE_DSOUND

#include <string.h> // no string functions used here, but possibly memset()
#ifndef _WINDOWS    // obviously compiling for a PC, so ..
# include "math.h"  // generate our lookup-tables at runtime in RAM, instead of using a 'const table in ROM'
#endif

#include "Timers.h" // high-resolution 'current time'-function, T_TIM_Stopwatch(), etc.
#include "Elbug.h"  // header for the basic 'Elbug' functions (plain C)
#include "CwDSP.h"  // header for THIS module ('CW Digitial Signal Processor')

float CwDSP_fltCosTable[CW_DSP_COS_TABLE_SIZE];


//----------------------------------------------------------------------------
// Global variables for hardcore debugging / post-mortem crash analysis,
//        when the debugger's call stack shows nothing but garbage,
//        e.g. after an exception at the infamous address 0xFEEEFEEE .
//        ( C++Builder's debugger could still inspect SIMPLE GLOBAL
//          VARIABLES after most kinds of exceptions, but of course nothing
//          stack-based because the CPU- or task-stack was usually trashed.)
//----------------------------------------------------------------------------
#if(SWI_HARDCORE_DEBUGGING) // (1) = hardcore-debugging, (0)=normal compilation
 int DSP_iLastSourceLine = 0;  // WATCH THIS after crashing with e.g. "0xFEEEFEEE" !
# define HERE_I_AM__DSP() DSP_iLastSourceLine=__LINE__
     // (see complete list of other XYZ_iLastSourceLine varibles to watch
     //  in C:\cbproj\Remote_CW_Keyer\Keyer_Main.cpp, near GUI_iLastSourceLine)
#else
# define HERE_I_AM__DSP()
#endif // SWI_HARDCORE_DEBUGGING ?


//----------------------------------------------------------------------------
// "Internal function prototypes" (functions called before their implementation)
//----------------------------------------------------------------------------

static DWORD WINAPI DspThread( LPVOID lpParam );
static int CwDSP_FrequencyToPhaseIncrement( int iAudioFrequency_Hz );
#if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
 static void CwDSP_EnterCriticalSection_AudioIO( T_CwDSP *pCwDSP );
 static void CwDSP_LeaveCriticalSection_AudioIO( T_CwDSP *pCwDSP );
#endif

// implementation of functions:


//----------------------------------------------------------------------------
void CwDSP_InitInstanceWithDefaults( T_CwDSP *pCwDSP )
  // Called ****ONCE**** from the main application (or "GUI") before loading an actual
  // configuration FROM A FILE. The result (in *pConfig) are the "defaults".
{
  int i;

  if( pCwDSP->dwInitMagic != DSP_INIT_MAGIC )
   { memset( pCwDSP, 0, sizeof(T_CwDSP) );
#   if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
     InitializeCriticalSection( &pCwDSP->csAudioIO );
#   endif
     pCwDSP->dwInitMagic = DSP_INIT_MAGIC;
   }
  else // someone didn't read the above ("call ****ONCE****"), so bear with him...
   { // don't memset() because this would kill our creepy little critter :o)
   }
  pCwDSP->cfg.iSidetoneRiseTime_ms = 4; // Icom offered 2, 4, 6, or 8 milliseconds
  pCwDSP->iSidetonePhaseAccu   = 0;
  pCwDSP->cfg.iSidetoneFreq_Hz = 650;
  pCwDSP->cfg.iAudioFlags = DSP_AUDIO_FLAGS_ALLOW_NETWORK_AUDIO;

  // Not really "instance specific", but on this occasion fill the COSINE table:
  for(i=0; i<CW_DSP_COS_TABLE_SIZE; ++i)
   { CwDSP_fltCosTable[i] = cos((double)i * 2.0 * 3.14159265359 / (double)CW_DSP_COS_TABLE_SIZE);
   }

  SRConv_InitDecimator( &pCwDSP->Decimators[0],    3/*decimation_ratio*/ );
  SRConv_InitDecimator( &pCwDSP->Decimators[1],    2/*decimation_ratio*/ );
  SRConv_InitDecimator( &pCwDSP->Interpolators[0], 2/*upsampling_ratio*/ );
  SRConv_InitDecimator( &pCwDSP->Interpolators[1], 3/*upsampling_ratio*/ );
      // '---> That's not a typo .. the same T_SRCONV_DECIMATOR struct
      //       can be used for decimation (downsampling)
      //       as well as for interpolation (upsampling) !
      //  First DECIMATE BY TWO, then decimate by three;
      //  first INTERPOLATE BY THREE, then interpolate by two. Guess why..

} // end CwDSP_InitInstanceWithDefaults()


//---------------------------------------------------------------------------
float CwDSP_DecibelToVoltageGainFactor( int iGain_dB )
{ return pow( 10.0, (float)iGain_dB / 20.0 );
} // end CwDSP_DecibelToVoltageGainFactor()

//---------------------------------------------------------------------------
void CwDSP_Start( T_CwDSP *pCwDSP )                                    // API
  // Tries to open all configured audio input and -output devices,
  //       and calls whatever the OS'es 'audio API' requires
  //       to read audio input, and produce audio output with our 'DSP'.
{
  CwDSP_Stop(pCwDSP); // <- doesn't do harm if not previously STARTED

#if( SWI_USE_DSOUND ) // use "DirectSound" / dsound_wrapper.c ?
  // If not done yet, but configured, open DirectSound devices for in- and/or output.
  //  [in]  pCwDSP->sz255AudioOutputDevice
  //  [out] pCwDSP->pDSW (our 'DirectSound' wrapper instance)
  if( (pCwDSP->pDSW != NULL ) && (pCwDSP->dwInitMagic==DSP_INIT_MAGIC) )
   {
#   if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
     CwDSP_EnterCriticalSection_AudioIO( pCwDSP ); // prevent other threads from calling DSW_xyz() now
#   endif
     DSW_CloseInput(  pCwDSP->pDSW ); // <- doesn't do harm if NOT opened at all, but don't close while DspThread() runs.
     DSW_CloseOutput( pCwDSP->pDSW ); // <- doesn't do harm if NOT opened at all, but don't close while DspThread() runs.
#   if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
     CwDSP_LeaveCriticalSection_AudioIO( pCwDSP );
#   endif
     if(   (pCwDSP->cfg.sz255AudioOutputDevice[0] != '\0')
        && (stricmp( pCwDSP->cfg.sz255AudioOutputDevice, "None") != 0 ) )
      { // want audio output via DirectSound -> kick it alive ..
#      if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
        CwDSP_EnterCriticalSection_AudioIO( pCwDSP ); // prevent other threads from calling DSW_xyz()
#      endif
        if( ! DSW_OpenOutputDevice( pCwDSP->pDSW, pCwDSP->cfg.sz255AudioOutputDevice ) )
         { strcpy( pCwDSP->sz255LastError, pCwDSP->pDSW->sz255LastOutputError );
         }
        else // successfully opened the output device, ready for the next step:
        if( ! DSW_InitOutputBuffer( pCwDSP->pDSW, CW_DSP_SAMPLING_RATE/*nFrameRate*/,
             1/*nChannels*/, CW_DSP_OUTPUT_BUFFER_NSAMPLES*sizeof(short)/*bufSize*/ ) )
         { strcpy( pCwDSP->sz255LastError, pCwDSP->pDSW->sz255LastOutputError );
         }
        else // successfully initialized the output buffer (circular FIFO),
         { // ready for the next step (see DirectSound API summary in dsound_wrapper.c):
           if( ! DSW_StartOutput( pCwDSP->pDSW ) )
            { strcpy( pCwDSP->sz255LastError, pCwDSP->pDSW->sz255LastOutputError );
            }
           // If all works as planned, the rest happens in CwDSP_UpdateSidetone(),
           //                          periodically called from KeyerThread() .
         } // end if < successfully initialized a DirectSound OUTPUT BUFFER >
#      if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
        CwDSP_LeaveCriticalSection_AudioIO( pCwDSP ); // allow others to call DSW_xyz() again
#      endif
      }   // end if < want audio OUTPUT via 'Direct Sound' ? >
     if(   (pCwDSP->cfg.sz255AudioInputDevice[0] != '\0')
        && (stricmp( pCwDSP->cfg.sz255AudioInputDevice, "None") != 0 ) )
      { // want audio INPUT via DirectSound ->
#      if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
        CwDSP_EnterCriticalSection_AudioIO( pCwDSP ); // prevent other threads from calling DSW_xyz()
#      endif
        if( ! DSW_OpenInputDevice( pCwDSP->pDSW, pCwDSP->cfg.sz255AudioInputDevice ) )
         { strcpy( pCwDSP->sz255LastError, pCwDSP->pDSW->sz255LastOutputError );
         }
        else // successfully opened the input device, ready for the next step:
        if( ! DSW_InitInputBuffer( pCwDSP->pDSW, CW_DSP_SAMPLING_RATE/*nFrameRate*/,
             1/*nChannels*/, CW_DSP_INPUT_BUFFER_NSAMPLES*sizeof(short)/*bufSize*/ ) )
         { strcpy( pCwDSP->sz255LastError, pCwDSP->pDSW->sz255LastOutputError );
         }
        else // successfully initialized the input buffer (circular FIFO),
         { // ready for the next step (see DirectSound API summary in dsound_wrapper.c):
           if( ! DSW_StartInput( pCwDSP->pDSW ) )
            { strcpy( pCwDSP->sz255LastError, pCwDSP->pDSW->sz255LastOutputError );
            }
           // If all works as planned, the rest happens in CwDSP_ProcessAudioInput(),
           //                       periodically called from ... not sure from where.
         } // end if < successfully initialized a DirectSound OUTPUT BUFFER >
#      if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
        CwDSP_LeaveCriticalSection_AudioIO( pCwDSP ); // allow others to call DSW_xyz() again
#      endif        
      }   // end if < want audio INPUT via 'Direct Sound' ? >
   }     // end if( pCwDSP->pDSW != NULL )
#endif // SWI_USE_DSOUND ?

  //
  // Create the worker thread (again, using ancient WIN32 API functions only,
  //                           no "dot net" framework, no VCL, no Qt, etc)
  if( pCwDSP->hThread==NULL )
   {  pCwDSP->iThreadStatus = DSP_THREAD_STATUS_LAUNCHED;
      pCwDSP->hThread = CreateThread(
         NULL,    // LPSECURITY_ATTRIBUTES lpThreadAttributes = pointer to thread security attributes
         65536,   // DWORD dwStackSize  = initial thread stack size, in bytes
         DspThread, // LPTHREAD_START_ROUTINE lpStartAddress = pointer to thread function
         (LPVOID*)pCwDSP,    // LPVOID lpParameter = argument for new thread
         0,       // DWORD dwCreationFlags = creation flags
                  // zero -> the thread runs immediately after creation
         &pCwDSP->dwThreadId // LPDWORD lpThreadId = pointer to returned thread id
       );
      // > The thread object remains in the system until the thread has terminated
      // > and all handles to it have been closed through a call to CloseHandle.
   }
  if( pCwDSP->hThread==NULL ) // Check the return value for success.
   {
     strcpy( pCwDSP->sz255LastError, "CreateThread failed." );
     pCwDSP->iThreadStatus = DSP_THREAD_STATUS_NOT_CREATED;
     // Don't bail out without a worker thread. At least the SIDETONE GENERATOR
     // doesn't need the 'DSP Thread', because it's called every 2 milliseconds(!)
     // from the KEYER THREAD ... don't ask why.
   }
  else
   { // > Define the Thread's priority as required.
     SetThreadPriority( pCwDSP->hThread, // handle to the thread
                THREAD_PRIORITY_NORMAL); // thread priority level
   }

} // end CwDSP_Start()


//---------------------------------------------------------------------------
void CwDSP_Stop( T_CwDSP *pCwDSP )                                     // API
  // Stops worker threads (if any), closes audio devices (if any),
  // frees buffers, etc.
{

  int i;
  if(  (pCwDSP->iThreadStatus == DSP_THREAD_STATUS_LAUNCHED )
    || (pCwDSP->iThreadStatus == DSP_THREAD_STATUS_RUNNING  ) )
   { pCwDSP->iThreadStatus = DSP_THREAD_STATUS_TERMINATE; // politely ask the thread to terminate itself
     for(i=0; i<20; ++i )
      { Sleep(10);
        if( pCwDSP->iThreadStatus == DSP_THREAD_STATUS_TERMINATED ) // bingo..
         { break;  // .. the thread has terminated itself, so no need to kill it
         }
        // If the thread is unable to TERMINATE ITSELF (and we get stuck here),
        // watch DSP_iLastSourceLine (in the debugger). Sometimes, the thread
        // got stuck in the call to DSW_ReadBlock(), so also
        // watch DSW_iLastSourceLine (DSW=DirectSound Wrapper; dsound_wrapper.c).
        // 2024-01-06 : Indeed DSPThread() died in DSW_ReadBlock(), and
        //              DSW_ReadBlock() died in IDirectSoundCaptureBuffer_Unlock()
        //              for reasons yet to be discovered. WATCH the following via
        //              debugger (when compiled with SWI_HARDCORE_DEBUGGING) :
        //
      }
   }
  if( pCwDSP->hThread != NULL )
   { CloseHandle( pCwDSP->hThread );
     pCwDSP->hThread  =  NULL;
   }
  pCwDSP->iThreadStatus = DSP_THREAD_STATUS_NOT_CREATED;


#if( SWI_USE_DSOUND ) // use "DirectSound" / dsound_wrapper.c ?
  // If not done yet, but configured, open DirectSound devices for in- and/or output.
  //  [in]  pCwDSP->cfg.sz255AudioOutputDevice
  //  [out] pCwDSP->pDSW (our 'DirectSound' wrapper instance)
  if( (pCwDSP->pDSW != NULL) && (pCwDSP->dwInitMagic==DSP_INIT_MAGIC) )
   {
#   if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
     CwDSP_EnterCriticalSection_AudioIO( pCwDSP );
#   endif
     DSW_CloseInput(  pCwDSP->pDSW ); // <- doesn't do harm if NOT opened at all ..
     DSW_CloseOutput( pCwDSP->pDSW );
#   if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
     CwDSP_LeaveCriticalSection_AudioIO( pCwDSP );
#   endif
   }
#endif // SWI_USE_DSOUND ?

} // end CwDSP_Stop()

#if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
//---------------------------------------------------------------------------
static void CwDSP_EnterCriticalSection_AudioIO( T_CwDSP *pCwDSP ) // INTERN !
{
  // Avoid crashing without InitializeCriticalSection() [shouldn't happen, but..]
  if( pCwDSP->dwInitMagic==DSP_INIT_MAGIC )
   { EnterCriticalSection( &pCwDSP->csAudioIO );
     // Between now and the following CwDSP_LeaveCriticalSection_AudioIO(),
     // the calling thread may the DSW_xyz()-functions in dsound_wrapper.c .
     // THE TIME THE CALLER KEEPS THE CRITICAL SECTION OCCUPIED MUST BE
     // KEPT AS SHORT AS POSSIBLE - no waiting, no sleeping, no processing
     // of audio samples, etc between ..Enter.. and ..Leave.. !
   }
} // end CwDSP_EnterCriticalSection_AudioIO()

//---------------------------------------------------------------------------
static void CwDSP_LeaveCriticalSection_AudioIO( T_CwDSP *pCwDSP ) // INTERN !
{
  if( pCwDSP->dwInitMagic==DSP_INIT_MAGIC )
   { LeaveCriticalSection( &pCwDSP->csAudioIO );
   }
} // end CwDSP_EnterCriticalSection_AudioIO()
#endif // SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO ?


//---------------------------------------------------------------------------
void CwDSP_SwitchOutputState( T_CwDSP *pCwDSP,
    int iNewOutputState) // [in] DSP_OUTPUT_STATE_.. (off, sidetone, RX-audio)
                         // [out] pCwDSP->iOutputState
{
  // can we 'fulfil' the caller's wish ?
  if( pCwDSP->pDSW == NULL )  // no DirectSound device available ->
   { pCwDSP->iOutputState = DSP_OUTPUT_STATE_OFF; // the output can only be *OFF*
     return;
   }
  if( ! pCwDSP->pDSW->fOutputOpened )  // no audio-OUTPUT-device *open* ->
   { pCwDSP->iOutputState = DSP_OUTPUT_STATE_OFF; // the output can only be *OFF*
     return;
   }
  // Arrived here ? The DSP can principally 'generate an audio output' (locally),
  // so switch between whatever the caller has requested :
  if( iNewOutputState != pCwDSP->iOutputState ) // ... here, with a STATE TRANSITION
   { pCwDSP->iOutputState = iNewOutputState;
     // Note: DO NOT SLEEP HERE ! The caller may be the KEYER THREAD,
     //       which cannot even affort being blocked for more than a few millseconds !
     switch( iNewOutputState )
      { case DSP_OUTPUT_STATE_OFF : // start outputting 'silence'
           break;
        case DSP_OUTPUT_STATE_GENERATE_SIDETONE: // start generating a LOW-LATENCY SIDETONE (locally, with the min. possible latency)
           break;
        case DSP_OUTPUT_STATE_RECEIVER_AUDIO:    // start outputting the RECEIVER'S audio signal (with more buffering and latency)
           break;
      } // end switch( iNewOutputState )
   } // end if( iNewOutputState != pCwDSP->iOutputState )

} // end CwDSP_SwitchOutputState()

//---------------------------------------------------------------------------
void CwDSP_UpdateSidetone( T_CwDSP *pCwDSP, BOOL fKeyDown )
  // Does what the name implies. Called from the fast-running
  // "PLC-like" KEYER THREAD (see KeyerThread.c : KeyerThread() ),
  // possibly as fast as every 2 milliseconds. Must not block,
  // must not wait for other threads, must not call any 'slow' OS functions,
  // and must not interfere with the audio samples that are POSSIBLY being
  // sent to the output in DspThread() [with a higher output latency] !
  //
  //  [in]  fKeyDown : "Morse keying signal" for the sidetone.
  //                   TRUE = tone on, FALSE = tone off.
  //        How to generate the waveform, and how to add it
  //        to e.g. the audio signal from the remote receiver,
  //        depends on the low-level audio API (DirectSound ?)
  //        and on the configuration in T_CwDSP itself.
{
  long nBytesEmpty;
  short i16Samples[CW_DSP_SIDETONE_BUFFER_NSAMPLES]; // Generate up to <CW_DSP_SIDETONE_BUFFER_NSAMPLES> audio samples per call.
        // Note: The DirectSound output buffer must be large enough
        //    for AT LEAST 10 milliseconds of audio (don't ask why);
        //    but CW_DSP_SIDETONE_BUFFER_NSAMPLES only needs to provide
        //    2 milliseconds of audio, plus some headroom because the
        //    calling interval of CwDSP_UpdateSidetone() isn't too predictable.
  int nSamples, nSamplesOccupied, iSample;
  int iSidetonePhaseInc;
  float fltSample, fltFactor, fltSidetoneRampInc;
  BOOL  fOk;

  if( (pCwDSP->pDSW != NULL) && (pCwDSP->pDSW->fOutputOpened) ) // Can we produce 'audio output' at all ?
   { // Automatically 'change over' between SIDETONE OUTPUT
     //  and passing through from RECEIVER (or network) to the AUDIO OUTPUT:
     if( pCwDSP->iOutputState != DSP_OUTPUT_STATE_GENERATE_SIDETONE )
      { // Audio output currently NOT used by the sidetone generator -> Switch now ?
        if( fKeyDown ) // yes, switch to "sidetone output" on "key down"
         { CwDSP_SwitchOutputState( pCwDSP, DSP_OUTPUT_STATE_GENERATE_SIDETONE );
         }
      } // end if( pCwDSP->iOutputState != DSP_OUTPUT_STATE_GENERATE_SIDETONE )

     if( pCwDSP->iOutputState == DSP_OUTPUT_STATE_GENERATE_SIDETONE ) // audio output currently occupied by the SIDETONE ?
      { // How many BYTES (not *sample points*) can we stuff into DirectSound's OUTPUT buffer ?
        // (to keep the latency as low as possible, don't the fill the output buffer COMPLETELY.
        //  Only fill in as many samples as we expect until THE NEXT CALL,
        //  which -when made from KeyerThread()- happens every two milliseconds)
#      if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
        CwDSP_EnterCriticalSection_AudioIO( pCwDSP ); // prevent other threads
        // from interfering between DSW_QueryOutputSpace() and DSW_WriteBlock().
#      endif // SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO ?
        // Generating the samples only takes a few microseconds, so it's ok...
        if( DSW_QueryOutputSpace( pCwDSP->pDSW, &nBytesEmpty ) )
         { // Ideally, when entering CwDSP_UpdateSidetone(),
           // approximately <CW_DSP_SIDETONE_BUFFER_NSAMPLES> should still
           // remain in the DirectSOund buffer, to avoid output buffer underruns.
           // But Direct Sound seems to 'drain' its output buffer only once in 10 ms.
           //  Example:  CW_DSP_OUTPUT_BUFFER_NSAMPLES   = 2048  ~~ 42 milliseconds
           //            CW_DSP_SIDETONE_BUFFER_NSAMPLES = 1024  ~~ 21 milliseconds
           //            CW_DSP_SAMPLING_RATE            = 48000
           // When hitting the Morse key for the first time, the output buffer
           // may be completely empty, and a breakpoint below should fire
           // with nBytesEmpty = CW_DSP_OUTPUT_BUFFER_NSAMPLES ...
           //   but it did NOT; the largest value that DSW_QueryOutputSpace()
           //   ever returned was HALF of the buffer size. Strange.
           //
           // To keep the audio latency down, even though the buffer is
           // LARGE, only emit APPROXIMATELY as many samples as the
           // calling interval (from KeyerThread() to CwDSP_UpdateSidetone(),
           // synchronized by the "Multimedia"-timer, set to 2 milliseconds).
           nSamples/*empty*/ = nBytesEmpty / sizeof(short); // MAXIMUM number of
                       // samples to emit (but usually MUCH MORE than necessary)
           nSamplesOccupied  = CW_DSP_OUTPUT_BUFFER_NSAMPLES/2 - nSamples;
           //  ,---------------|_____________________________|
           //  '--> Kludge because DSW_QueryOutputSpace() NEVER reported
           //       an empty or almost empty buffer, even when
           //       DSW_WriteBlock() wasn't called at all. Got to clean up
           //       this mess one fine day; but for now, it works reliably.)
           #define L_WANTED_BUFFER_USAGE_MS 12 // If over <L_WANTED_BUFFER_USAGE_MS> of audio are already buffered ...
           if( nSamplesOccupied > (CW_DSP_SAMPLING_RATE/(1000/L_WANTED_BUFFER_USAGE_MS)) )
            { if( nSamples > (CW_DSP_SAMPLING_RATE/520) ) // .. then emit "slightly less than 2 ms"
               {  nSamples = (CW_DSP_SAMPLING_RATE/520);  // emit e.g. 48000/520 = 92 samples = ~~1.92 milliseconds
               }
            }
           else // less than 12 ms samples already in the DirectSound output buffer ->
            { if( nSamples > (CW_DSP_SAMPLING_RATE/480) ) // emit "slightly MORE than 2 ms" ...
               {  nSamples = (CW_DSP_SAMPLING_RATE/480);  // e.g. 48000/480 = 100 samples = ~~2.08 milliseconds
               }
              // The result is a kind of control loop to keep the DirectSound
              // output buffer JUST BELOW the point of a buffer *underrun*.
              // This gives the lowest latency that the "new" 10-millisecond
              // READ INTERVAL permits. If audio stutters or seems to "hum",
              // allow a few MORE milliseconds in the buffer.
              // When all worked as planned, the "report" on the 'Debug' tab showed:
              //  >  SidetoneSamples: 92 92 100 100 92 92 92 100 100 100 92 92 100 100 92 100
              // (which means ON AVERAGE, 'short' and 'slightly longer' sample blocks
              //  are equally distributed)
            }
           if( nSamples >= (CW_DSP_SAMPLING_RATE/1000) ) // only calculate new waveform when at least 1ms can be emitted
            {
              // Store the number of audio samples sent per 2-ms 'KeyerThread'
              //  in a short FIFO for debugging (this can be examined on the 'Debug'-tab):
              pCwDSP->i16SidetoneSamplesPerUpdate[ pCwDSP->dwSidetoneUpdates & 15 ] = nSamples;
              ++pCwDSP->dwSidetoneUpdates;
              fltFactor = CwDSP_DecibelToVoltageGainFactor( pCwDSP->cfg.iSidetoneGain_dB );
              fltFactor *= 3276.7; /* 32767 to scale into 16-bit integer "minus 20 dB" */
              iSidetonePhaseInc = CwDSP_FrequencyToPhaseIncrement( pCwDSP->cfg.iSidetoneFreq_Hz );
              if( pCwDSP->cfg.iSidetoneRiseTime_ms > 0 )  // use a "nicely shaped" sidetone ?
               { fltSidetoneRampInc = (1000.0/CW_DSP_SAMPLING_RATE) / (pCwDSP->cfg.iSidetoneRiseTime_ms);
                 // Example: CW_DSP_SAMPLING_RATE = 12000 [Hz],
                 //    iSidetoneRiseTime_ms = 4 [ms] -> 12 kHz * 4 ms = 48 samples
                 //    during which pCwDSP->iSidetoneRamp must rise from 0 to 1.0
                 // -> fltSidetoneRampInc = (1000 / 48000) / 4 = circa 0.005 .
               }
              else // no ramp but "hard keying" (clickety-click, who cares about bandwidth..)
               { fltSidetoneRampInc = 32767;
               }
              for( iSample=0; iSample<nSamples; ++iSample )
               {
                 // "Ramp up" or "Ramp down" ?
                 if( fKeyDown ) // "ramp up" (until reaching the sidetone's full amplitude)
                  { pCwDSP->fltSidetoneRamp += fltSidetoneRampInc;
                    if( pCwDSP->fltSidetoneRamp > 1.0 )
                     {  pCwDSP->fltSidetoneRamp = 1.0;
                     }
                  }
                 else // "ramp down" (until arriving at The Sound of Silence)
                  { pCwDSP->fltSidetoneRamp -= fltSidetoneRampInc;
                    if( pCwDSP->fltSidetoneRamp < 0.0 )
                     {  pCwDSP->fltSidetoneRamp = 0.0;
                     }
                  }
                 fltSample = fltFactor * pCwDSP->fltSidetoneRamp * CwDSP_fltCosTable[ pCwDSP->iSidetonePhaseAccu ];
                 if( fltSample > 32767.0 )
                  {  fltSample = 32767.0;
                  }
                 if( fltSample < -32767.0 )
                  {  fltSample = -32767.0;
                  }
                 i16Samples[iSample] = (short)fltSample;
                 pCwDSP->iSidetonePhaseAccu += iSidetonePhaseInc;
                 pCwDSP->iSidetonePhaseAccu &= (CW_DSP_COS_TABLE_SIZE-1);
               }
              if( ! DSW_WriteBlock( pCwDSP->pDSW, (char *)i16Samples, nSamples*sizeof(short) ) )
               { // Don't STRING-COPY the error message, just copy a POINTER here:
                 pCwDSP->pszLastError = pCwDSP->pDSW->sz255LastOutputError;
               } // end if < DSW_WriteBlock() failed >
            }   // end if < enough space to 'generate' a few milliseconds of CW sidetone >
         }     // end if < Mr. DirectSound-wrapper successully queried the OUTPUT BUFFER SPACE >
#      if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
        CwDSP_LeaveCriticalSection_AudioIO( pCwDSP ); // allow others to call DirectSound again
#      endif        
      }      // end if < iOutputState == DSP_OUTPUT_STATE_GENERATE_SIDETONE > ?
   }        // end if < got a valid pointer to a DirectSound wrapper instance, *AND* "open for output" > ?
}          // end CwDSP_UpdateSidetone()


//---------------------------------------------------------------------------
static int CwDSP_FrequencyToPhaseIncrement( int iAudioFrequency_Hz )
  // Computes the integer 'phase increment' for our cosine lookup table,
  // CwDSP_fltCosTable[CW_DSP_COS_TABLE_SIZE] .
  // Again: This module was originally written with a microcontroller firmware
  //        in mind where we can't afford to call floating point functions
  //        like sin(), cos(), etc - especially not from a fast-running
  //        timer interrupt that directly calculates audio samples for a DAC.
{
  // At typical "audio frequencies" (like 650 Hz for the CW sidetone),
  //    we get along with the following "rounding integer calculation":
  //    (the developer -WB- still had the implementation on a low-power
  //     microcontroller in mind, possibly a PIC24, dsPIC, Cortex-M0,
  //     or similar WITHOUT hardware floating point)
  return (iAudioFrequency_Hz * CW_DSP_COS_TABLE_SIZE + (CW_DSP_SAMPLING_RATE/2) )
                                                      / CW_DSP_SAMPLING_RATE;
  // Examples :
  //  iAudioFrequency_Hz = 0 (off) -> iSidetonePhaseInc = 0
  //  iAudioFrequency_Hz = 650 Hz  -> iSidetonePhaseInc = (650 * 4096 + 6000) / 12000 = 222
  //  iAudioFrequency_Hz = 640 Hz  -> iSidetonePhaseInc = (640 * 4096 + 6000) / 12000 = 218
  // ,----------------------------------------------------|________|
  // '--> Easily fits inside a 32-bit int, but wouldn't fit in a 16-bit integer.
  //      Old (12- or 16-bit) 'PIC' microcontroller firmware developers beware.
} // end CwDSP_FrequencyToPhaseIncrement()


//---------------------------------------------------------------------------
void CwDSP_ShortToFloat( short *pi16In, float *pfltOut, int nSamples, float fltFactor )
  // [in] fltFactor: usually 1.0 / 32767.0 to "normalize" the 16-bit input
  //                 to -1.0 .. +1.0 in the floating point output input .
{
  while( (nSamples--) > 0 )
   { *pfltOut++ = (float)(*pi16In++) * fltFactor;
   }
} // end CwDsp_ShortToFloat()

//---------------------------------------------------------------------------
void CwDSP_FloatToShort( float *pfltIn, short *pi16Out, int nSamples, float fltFactor )
  // [in] fltFactor: usually 32767.0 to "de-normalize" the floating-point
  //                 (ranging from -1.0 to +1.0) into 16-bit signed integer ouput.
{
  float flt;
  while( (nSamples--) > 0 )
   { flt = (*pfltIn++) * fltFactor;
     if( flt > 32767.0f )  // limit (clip) BEFORE converting to integer
      {  flt = 32767.0f;
      }
     if( flt < -32767.0f ) // limit (clip) BEFORE converting to integer
      {  flt = -32767.0f;  // Note that we deliberately omit -32768 here.
         // That pattern (0x8000 as 'short int') may be used for framing, etc.
      }
     *pi16Out++ = (short)flt;
   }
} // end CwDsp_FloatToShort()

//---------------------------------------------------------------------------
int CwDSP_GetNumSamplesInFifoForTailIndex( T_CwDSP_Fifo *pFifo, int iTail )
  // Takes A MOMENTARY SNAPSHOT of the number of samples in any of the DSP's FIFOs,
  // suitable for MULTIPLE READERS (where each of those readers has its own
  // 'tail index', while there is always only ONE WRITER / one HEAD index) .
{ int nSamples = pFifo->iHead - iTail;
  if( nSamples < 0 )  // circular index wrap between "head" and "tail" ->
   {  nSamples += CW_DSP_PROCESS_FIFO_NSAMPLES;
   }
  return nSamples;
} // end CwDSP_GetNumSamplesInFifo()

//---------------------------------------------------------------------------
int CwDSP_GetNumSamplesInFifo( T_CwDSP_Fifo *pFifo ) // .. using the primary head- and tail index
  // Takes A MOMENTARY SNAPSHOT of the number of samples in any of the DSP's FIFOs,
{ return CwDSP_GetNumSamplesInFifoForTailIndex( pFifo, pFifo->iTail );
} // end CwDSP_GetNumSamplesInFifo()

//---------------------------------------------------------------------------
int CwDSP_GetFreeSpaceInFifo( T_CwDSP_Fifo *pFifo )
  // Takes A MOMENTARY SNAPSHOT of the number of samples that may be WRITTEN
  // into this FIFO before it overflows.
{ return (CW_DSP_PROCESS_FIFO_NSAMPLES-1) - CwDSP_GetNumSamplesInFifo( pFifo );
  // ,---------------------------------'
  // '--> This "classic, lock-free, circular FIFO" can store ONE SAMPLE LESS
  //      than the fixed queue size (in sample points), because :
  //      head index == tail index                 ->    FIFO is empty
  //      (head index + 1), wrapped == tail index  -> FIFO is COMPLETELY full
} // end CwDSP_GetFreeSpaceInFifo()

//---------------------------------------------------------------------------
void CwDSP_ReadFromFifo( T_CwDSP_Fifo *pFifo, int *piTail, float *pfltDestBuffer, int nSamples )
  // Note: The caller has already checked the number of samples in the source FIFO !
{ int i, iTail = *piTail & (CW_DSP_PROCESS_FIFO_NSAMPLES-1);
  for( i=0; i<nSamples; ++i )
   { pfltDestBuffer[i] = pFifo->fltQueue[ iTail++ ];
     if( iTail >= CW_DSP_PROCESS_FIFO_NSAMPLES/*2048?*/ )
      {  iTail = 0; // <-- good place for a breakpoint ..
      }
   }
  *piTail = iTail; // put the FIFO TAIL index back where it belongs
} // end CwDSP_ReadFromFifo()


//---------------------------------------------------------------------------
void CwDSP_WriteToFifo( T_CwDSP_Fifo *pFifo, int *piHead, float *pfltSourceBuffer, int nSamples )
  // Note: The caller has already checked the remainin capacity in the destination FIFO,
  //       or simply ignores if the READER (that drains the FIFO) can keep up the pace.
{ int i, iHead = *piHead & (CW_DSP_PROCESS_FIFO_NSAMPLES-1);
  for( i=0; i<nSamples; ++i )
   { pFifo->fltQueue[ iHead++ ] = pfltSourceBuffer[i];
     if( iHead >= CW_DSP_PROCESS_FIFO_NSAMPLES/*2048?*/ )
      {  iHead = 0; // <-- good place for a breakpoint ..
      }
   }
  *piHead = iHead; // put the FIFO HEAD index" back where it belongs .
  // This make the new samples in pCwDSP->fltInputFifo[] available for other threads,
  // for example when THE KEYER THREAD reads another bunch
  // in CwKeyer_CollectDataForTimingScope(), every TWO MILLISECONDS .
} // end CwDSP_WriteToFifo()

//---------------------------------------------------------------------------
void CwDSP_ProcessSampleFromReceivedAudioStream( T_CwDSP *pCwDSP,
       float fltSample ) // [in] single audio sample received with f_sample = 8 kHz
       // from the NETWORK THREAD (CwNet.c), already normalized to -1 .. +1 .
{
  if( pCwDSP != NULL )
   {
     if( (pCwDSP->iOutputState != DSP_OUTPUT_STATE_GENERATE_SIDETONE )
       &&(pCwDSP->iOutputState != DSP_OUTPUT_STATE_NETWORK_AUDIO ) )
      { // ok; switch to OUTPUT NETWORK AUDIO :
        CwDSP_SwitchOutputState( pCwDSP, DSP_OUTPUT_STATE_NETWORK_AUDIO );
        // ,-----------------------------'
        // '--> Controls the 'routing' of audio in DspThread() .
      }
     CwDSP_WriteToFifo( &pCwDSP->sNetworkRxFifo, &pCwDSP->sNetworkRxFifo.iHead, &fltSample, 1/*nSamples*/ );
     // Depending on pCwDSP->iOutputState, the DSP worker thread
     // will either pass this received audio stream on to the RADIO (voice-TX?)
     // or to the CLIENT OPERATOR's loudspeaker, headphone, or whatever .
     // Of course that doesn't happen sample-by-sample but block-by-block.
   }
} // end CwDSP_ProcessSampleFromReceivedAudioStream()



//---------------------------------------------------------------------------
DWORD WINAPI DspThread( LPVOID lpParam )
  // 'DSP thread' : Reads input from the audio device (if configured),
  //                sends output to another audio device (if configured),
  //                proceses audio from / for streaming via network,
  //                and who-knows-what ...
{
  T_CwDSP *pCwDSP = (T_CwDSP*)lpParam;
  // T_CwNet *pCwNet = CwKeyer_Config.pCwNet; // optional instance for audio "via CW network"

  T_TIM_Stopwatch sw_BeginOfThreadLoop, sw_SpeedTest;
  long nMicrosecondsSinceLastCall;
  int  t_us, t_ms, iResult, iFifoHead, iFifoTail;
  int  iThreadExitCode = 1;
  long nBytesFilled, nBytesEmpty, nBytesToReadOrWrite;
  int  i, nSamplesIn, nSamplesOut, nDecimatedSamples/*at 8 kS/s*/;
  int  nLoopsWithoutOutput = 0;
# define SAMPLES_PER_AUDIO_IO_BUFFER 2048 // <- must be large enough for
      // approximately 20 ms (worst-case-estimate for the thread loop time)
      //          * 48 kHz = 960 samples ~~ 1024 (next larger power of two)
  short i16AudioInOutBuffer[SAMPLES_PER_AUDIO_IO_BUFFER];
  float fltAudioInputBuffer[SAMPLES_PER_AUDIO_IO_BUFFER];
  float fltAudioOutputBuffer[SAMPLES_PER_AUDIO_IO_BUFFER];
  float fltFactor;
  BOOL  ok;
  T_CwDSP_Fifo *pSourceFifo;


  pCwDSP->iThreadStatus = DSP_THREAD_STATUS_RUNNING;
  pCwDSP->dwThreadLoops = pCwDSP->dwThreadErrors = 0;


  TIM_StartStopwatch( &sw_BeginOfThreadLoop ); // don't leave this un-initialized..

  // Almost endless DSP thread loop begins HERE.............................
  while( (pCwDSP->iThreadStatus == DSP_THREAD_STATUS_RUNNING)
      && (pCwDSP->dwInitMagic == DSP_INIT_MAGIC) ) // bail out when the magic smoke escaped
   {
     HERE_I_AM__DSP();  // -> DSP_iLastSourceLine (when compiled for 'hardcore debugging')
     TIM_StartStopwatch( &pCwDSP->sw_ThreadWatchdog ); // <- allow the GUI to check for "crashed threads" :
         // (If iThreadStatus == DSP_THREAD_STATUS_RUNNING but
         //  but TIM_ReadStopwatch( &pCwDSP->sw_ThreadWatchdog ) > 500,
         //  the GUI may report problems in the error history)
     ++pCwDSP->dwThreadLoops;
     pCwDSP->dw8ThreadIntervals_us[pCwDSP->dwThreadLoops&7]
       = TIM_ReadAndRestartStopwatch_us( &sw_BeginOfThreadLoop );

     Sleep(5); // <- Not very reliable because windows may let this thread "sleep"
     // for MUCH LONGER than 5 milliseconds (often seen 6 ms, but expect more) .
     HERE_I_AM__DSP(); // didn't arrive HERE ? Most likely, crash in ANOTHER thread.

     // -------------- AUDIO INPUT (from selected audio device) -------------
#   if( SWI_USE_DSOUND ) // use Windows "DirectSound" / dsound_wrapper.c ?
     if( pCwDSP->pDSW != NULL )
      { HERE_I_AM__DSP();
        TIM_StartStopwatch( &sw_SpeedTest );
#      if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
        CwDSP_EnterCriticalSection_AudioIO( pCwDSP ); // prevent other threads...
        // from interfering between DSW_QueryInputFilled() and DSW_ReadBlock().
#      endif
        if( DSW_QueryInputFilled( pCwDSP->pDSW, &nBytesFilled ) ) // ready for AUDIO INPUT ?
         { // Note: From Mr. DirectSound, bytes arrived in funny block sizes
           //       of e.g. nBytesFilled = 3000 bytes, sometimes even 4024 bytes,
           //       so don't assume they are nice powers of two !
           nSamplesIn = nBytesFilled / sizeof(short); // -> 16 ... SAMPLES_PER_AUDIO_IO_BUFFER
           if( nSamplesIn > SAMPLES_PER_AUDIO_IO_BUFFER ) // repect the capacity of i16AudioInOutBuffer[SAMPLES_PER_AUDIO_IO_BUFFER]
            {  nSamplesIn = SAMPLES_PER_AUDIO_IO_BUFFER;
            }
           nBytesFilled = nSamplesIn * sizeof(short); // back from NUMBER OF SAMPLE POINTS to a 'number of bytes' ... sigh
           if( nSamplesIn < 16 ) // don't waste time reading less than 16 samples from the input
            {
#            if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
              CwDSP_LeaveCriticalSection_AudioIO( pCwDSP ); // immediately allow others to call Mr. DirectSound again
#            endif
            }
           else // got ENOUGH samples for processing (don't waste time processing one or two samples..) ->
            { HERE_I_AM__DSP();  // <- this confirmed crashing in DSW_ReadBlock() with the '0xFEEEFEEE' exception !
              ok = DSW_ReadBlock( pCwDSP->pDSW, (char*)i16AudioInOutBuffer, nBytesFilled/*! .. not nSamplesIn !*/ );
              HERE_I_AM__DSP(); // <- survived another call of DSW_ReadBlock() ...
#            if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
              CwDSP_LeaveCriticalSection_AudioIO( pCwDSP ); // after DSP_ReadBlock(), allow others to call Mr. DirectSound again
              HERE_I_AM__DSP(); // <- survived leaving the critter supposed to protect Mr. DirectSound ...
#            endif
              if( ok )
               { // DirectSound (input) delivered i16AudioInOutBuffer[0..nSamplesIn-1] .
                 // Typically seen here: nSamplesIn = 476 .   476 / 48kHz = 9.9 ms,
                 //  obviously caused by the "10 millisecond DirectSound I/O update rate".
                 CwDSP_UpdateSpeedTestResult( pCwDSP, DSP_SPEEDTEST_READ_AUDIO_INPUT, TIM_ReadStopwatch_us( &sw_SpeedTest) );
                 // ,---------------------------------|____________________________|
                 // '--> here: time spent in DSW_QueryInputFilled() plus DSW_ReadBlock().
                 TIM_StartStopwatch( &sw_SpeedTest ); // now measure the time spent for converting/downsampling the INPUT
                 // Convert the 16-bit signed integers from Mr. "DirectSound" (or whatever)
                 // into more DSP-friendly, but on a PC equally fast floating point samples.
                 // So we don't have to care about integer overflow (or clipping)
                 // until converting the processed result back into integer,
                 // with e.g. automatic gain control, or whatever :
                 fltFactor = CwDSP_DecibelToVoltageGainFactor( pCwDSP->cfg.iAudioInGain_dB ) / 32767.0;
                 CwDSP_ShortToFloat( i16AudioInOutBuffer, fltAudioInputBuffer, nSamplesIn, fltFactor );
                 //  '--> not only converts the data type, but may also SCALE
                 //       (normalize) the sample value ranges to -1.0 ... +1.0 .
                 // Next processing step : DOWNSAMPLE to 8 kSamples/second :
                 nDecimatedSamples = SRConv_TwoStageDecimator( pCwDSP->Decimators, fltAudioInputBuffer,
                                        nSamplesIn, fltAudioOutputBuffer );
                 // We're at fsample = 8 kHz here now .
                 //       From nSamples = 476, nDecimatedSamples = 80 poured out
                 //       of the decimator (decimating by SIX in this case).
                 //                       476 / 6 = 79.3333 -> ok .
                 // Store the decimated result in the DSP's CIRCULAR INPUT BUFFER:
                 CwDSP_WriteToFifo( &pCwDSP->sInputFifo, &pCwDSP->sInputFifo.iHead, fltAudioOutputBuffer, nDecimatedSamples );
                 CwDSP_UpdateSpeedTestResult( pCwDSP, DSP_SPEEDTEST_DOWNSAMLE_INPUT, TIM_ReadStopwatch_us( &sw_SpeedTest) );
               }
            } // end while < enough samples to fill another audio buffer >
         } // end if < DSW_QueryInputFilled() successful ? >
        // No-No: CwDSP_LeaveCriticalSection_AudioIO( pCwDSP ); (already called above, e.g. shortly after DSW_ReadBlock() )
        HERE_I_AM__DSP();
      } // end if( pCwDSP->pDSW != NULL )
#   endif // SWI_USE_DSOUND ?


     // At THIS point, the decimated pCwDSP->fltInputFifo[] is ready for processing.
     // In many cases ("CW network" enabled), the next steps is
     //    AUDIO COMPRESSION (into 8 bits per sample) and merging of that stream
     //    into the outbound TCP/IP traffic (in module CwNet.c) .
     // Only if THERE IS NO CLIENT/SERVER COMMUNICATION INVOLVED,
     //    the "processing" merely consists of copying
     //    as many samples from pCwDSP->fltInputFifo[] into pCwDSP->fltOutputFifo[],
     //    e.g. to pass on the audio from the local receiver
     //         to the operator's "shack PC" or whatever.
     pSourceFifo = &pCwDSP->sInputFifo;  // default "source" for the audio output
     if( pCwDSP->iOutputState == DSP_OUTPUT_STATE_NETWORK_AUDIO ) // audio output currently "from the network" (e.g. remote server) ?
      { pSourceFifo = &pCwDSP->sNetworkRxFifo;
      }


     nSamplesOut = CwDSP_GetFreeSpaceInFifo( &pCwDSP->sOutputFifo );
        // '--> number of samples at fs=8 kHz that CAN be appended to the circular output FIFO.
        //      Note that the maximum CAPACITY of this type of FIFO is the array-length MINUS ONE.
     nSamplesIn = CwDSP_GetNumSamplesInFifo( pSourceFifo );
        // '--> number of samples at fs=8 kHz we MAY read from the circular input FIFO.
     if( nSamplesOut < nSamplesIn )  // for COPYING samples, take the lower of the two..
      { nDecimatedSamples = i;    // here: number of samples limited by INPUT-FIFO-usage
      }
     if( nSamplesIn > 0 )  // ok, copy as much as we can from in- to output FIFO:
      { iFifoTail = pSourceFifo->iTail;
        iFifoHead = pCwDSP->sOutputFifo.iHead;
        for( i=0; i<nSamplesIn; ++i )
         {
#         if( 1 ) // (1)=normal compilation, (0)=test with a pure sinewave for the DECIMATED OUTPUT
           pCwDSP->sOutputFifo.fltQueue[ iFifoHead++ ] = pSourceFifo->fltQueue[ iFifoTail++ ];
#         else
           pCwDSP->sOutputFifo.fltQueue[ iFifoHead ] = CwDSP_fltCosTable[
                                 (iFifoHead * (32*CW_DSP_COS_TABLE_SIZE/CW_DSP_PROCESS_FIFO_NSAMPLES) ) % CW_DSP_COS_TABLE_SIZE];
                // sinewave periods per FIFO___|  |_______ 4096 ______| |_________ 2048 ___________|
           iFifoHead++;
           iFifoTail++;  // TEST: throw away the input
#         endif // normal compilation or "output interpolator test" with a clean sinewave ?
           iFifoHead &= (CW_DSP_PROCESS_FIFO_NSAMPLES-1); // circular index wrap ..
           iFifoTail &= (CW_DSP_PROCESS_FIFO_NSAMPLES-1);
         }
        pSourceFifo->iTail = iFifoTail;  // drained the "source" (sInputFifo or sNetworkRxFifo)
        pCwDSP->sOutputFifo.iHead = iFifoHead;  // we've FILLED the output FIFO..
      } // end if < copy any (decimated) samples from pCwDSP->sInputFifo to pCwDSP->sOutputFifo ?

     // -------------- AUDIO OUTUT (via selected audio device) --------------
     if( pCwDSP->iOutputState != DSP_OUTPUT_STATE_GENERATE_SIDETONE ) // audio output currently NOT occupied by the SIDETONE ?
      {
#      if( SWI_USE_DSOUND ) // use Windows "DirectSound" / dsound_wrapper.c ?
        if( pCwDSP->pDSW != NULL ) // ok, there is a DirectSound-wrapper ...
         { HERE_I_AM__DSP();

           // Determine the number of samples we COULD read from the OUTPUT FIFO
           // (contains DECIMATED samples, i.e. at f_sample = 8 kHz) :
           nDecimatedSamples = CwDSP_GetNumSamplesInFifo( &pCwDSP->sOutputFifo );
           nSamplesIn = nDecimatedSamples * 6; // we're upsampling by SIX..
           // nSamplesIn is the number of samples we COULD 'interpolate' (upsample)
           // to f_sample = 48 kHz at the moment, if Mr. DirectSound-OUT permits.
           // But avoid to interpolate and write to the output device if only
           // a few samples can be written at all:
           if( nSamplesIn >= 16 ) // "enough samples available to write to the output" ?
            { // How many samples can Mr. DirectSound's OUTPUT consume at the moment ?
#            if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
              CwDSP_EnterCriticalSection_AudioIO( pCwDSP ); // prevent other threads
              // from interfering between DSW_QueryOutputSpace() and DSW_WriteBlock().
              // Interpolating a few samples for output only takes a few microseconds,
              // so this is ok, too.
#            endif
              // Determine the number of interpolated (upsampled) samples
              // we COULD write to the audio output device at the moment:
              if( DSW_QueryOutputSpace( pCwDSP->pDSW, &nBytesEmpty ) ) // ready for AUDIO OUTPUT ?
               { // As usual from DirectSound, exect to see funny "empty sizes"
                 // like nBytesEmpty = 3776, when the thread WANTS TO WRITE .
                 nSamplesOut = nBytesEmpty / sizeof(short);
                 if( nSamplesOut >= 64 ) // also "worth the effort" now ?
                  { HERE_I_AM__DSP();
                    if( nSamplesOut > nSamplesIn )
                     {  nSamplesOut = nSamplesIn;
                     }
                    if( nSamplesOut > SAMPLES_PER_AUDIO_IO_BUFFER )
                     { // Limit this to the capacity of fltAudioInputBuffer[] :
                       nSamplesOut = SAMPLES_PER_AUDIO_IO_BUFFER; // e.g. 1024 (enough for 20 ms, if the thread loop is "slow"..)
                     }
                    nDecimatedSamples = nSamplesOut / 6;  // <- this is an EXACT number..
                    // '--> ALWAYS fits inside fltAudioInputBuffer[SAMPLES_PER_AUDIO_IO_BUFFER],
                    //      e.g. nDecimatedSamples = 1024 / 6 = 170; enough for 20 ms @ fsample=8 kHz.
                    // Copy as many DECIMATED sample as we can from the DSP's circular FIFO.
                    // SRConv_TwoStageUpsampler() needs them in a simple, non-circular ARRAY.
                    CwDSP_ReadFromFifo( &pCwDSP->sOutputFifo, &pCwDSP->sOutputFifo.iTail, fltAudioInputBuffer, nDecimatedSamples );
                    nSamplesOut = SRConv_TwoStageUpsampler( pCwDSP->Interpolators,
                                             fltAudioInputBuffer/*pfltInput*/,
                                              nDecimatedSamples/*nSamplesIn*/,
                                           fltAudioOutputBuffer/*pfltOutput*/);
                    if(  (nSamplesOut > 0 )  // the "two-stage upsampler" seems to be working..
                      && (nSamplesOut <= SAMPLES_PER_AUDIO_IO_BUFFER) ) // and we won't exceed our audio I/O buffer...
                     { // .. so convert the 32-bit floating point values from our DSP
                       //    into 16-bit signed integers for Mr. DirectSound :
                       fltFactor = 32767.0 * CwDSP_DecibelToVoltageGainFactor( pCwDSP->cfg.iAudioOutGain_dB );
                       CwDSP_FloatToShort( fltAudioOutputBuffer, i16AudioInOutBuffer, nSamplesOut, fltFactor );
                       // For example, with nDecimatedSamples = 170 ( ~~ 20ms @ fsample=8 kHz),
                       //     there are nSamplesOut = 170 * 6 = 1020(!) samples to write now:
                       if( ! DSW_WriteBlock( pCwDSP->pDSW, (char *)i16AudioInOutBuffer, nSamplesOut*sizeof(short) ) )
                        { // Don't STRING-COPY the error message, just copy a POINTER here:
                          pCwDSP->pszLastError = pCwDSP->pDSW->sz255LastOutputError;
                        } // end if < DSW_WriteBlock() failed >
                       else // successfully wrote samples to the AUDIO OUTPUT DEVICE ->
                        { nLoopsWithoutOutput = 0;
                        }
                     } // end if < SRConv_TwoStageUpsampler() working > ?
                  } // end if( nBytesEmpty >= (int)sizeof(i16AudioInOutBuffer) )
               }   // end if < DSW_QueryOutputSpace() successful > ?
#            if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
              CwDSP_LeaveCriticalSection_AudioIO( pCwDSP ); // allow others to call DirectSound again
#            endif
            }    // end if < "enough samples available to write to the output" > ?
           else  // less than 16(?) samples available to write to the output ->
            { // This often happened when e.g. a network connection was unstable.
              // The DirectSound output repeated the same samples in a short loop,
              // causing an unpleasant buzzing sound. So try to "silence" the output..
              if( (nLoopsWithoutOutput++) > 10 ) // .. after 10 thread loops
               {
#               if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
                 CwDSP_EnterCriticalSection_AudioIO( pCwDSP ); // prevent other threads from interfering..
#               endif
                 DSW_ZeroEmptySpaceInOutputBuffer( pCwDSP->pDSW );
#               if( SWI_USE_CRITICAL_SECTIONS_FOR_AUDIO_IO )
                 CwDSP_LeaveCriticalSection_AudioIO( pCwDSP ); // allow others to call DirectSound again
#               endif
               }
            }
         }      // end if( pCwDSP->pDSW != NULL )
#      endif   // SWI_USE_DSOUND ?
      }       // end if( pCwDSP->iOutputState != DSP_OUTPUT_STATE_GENERATE_SIDETONE )

   } // end while < DSP thread loop >

  // If the program ever gets here, the thread has "politely" terminated itself,
  //                    or someone has pulled the emergency brake .
  pCwDSP->iThreadStatus = DSP_THREAD_STATUS_TERMINATED;

  ExitThread( iThreadExitCode ); // exit code for this thread
  return iThreadExitCode; // will this ever be reached after "ExitThread" ?
} // end DspThread()
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void CwDSP_UpdateSpeedTestResult( T_CwDSP *pCwDSP, // .. for development only ..
             int iTestItem,     // [in] e.g. KEYER_SPEEDTEST_POLL_KEYBOARD
             int nMicroseconds ) // [in] new test result in microseconds
{
  if( nMicroseconds >= pCwDSP->iSpeedTestPeaks_us[iTestItem] )
   {  pCwDSP->iSpeedTestPeaks_us[iTestItem] = nMicroseconds;
      // Note: A new "peak- and average detection" begins
      //       when clicking "Report Test Results (on the 'debug' tab)" in the GUI
      //  -> CwDSP_ResetSpeedTestResults()
   }
  pCwDSP->i32SpeedTestSums_us[iTestItem] += nMicroseconds;
  ++pCwDSP->i32SpeedTestCounts_us[iTestItem];

} // end CwDSP_UpdateSpeedTestResult()

//---------------------------------------------------------------------------
int  CwDSP_GetSpeedTestAverage_us( T_CwDSP *pCwDSP,
              int iTestItem ) // [in] e.g. KEYER_SPEEDTEST_POLL_KEYBOARD
{ long i32Divisor = pCwDSP->i32SpeedTestCounts_us[ iTestItem ];
  if( i32Divisor > 0 )
   { return (int)( pCwDSP->i32SpeedTestSums_us[iTestItem] / i32Divisor );
   }
  else // no valid data for this 'item' -> say the average time spent for it was ZERO
   { return 0;
   }
} // end CwDSP_GetSpeedTestAverage_us()

//---------------------------------------------------------------------------
void CwDSP_ResetSpeedTestResults(T_CwDSP *pCwDSP) // begins a new peak- and average detection
{
  memset( (void*)pCwDSP->iSpeedTestPeaks_us,   0, sizeof(pCwDSP->iSpeedTestPeaks_us) );
  memset( (void*)pCwDSP->i32SpeedTestSums_us,  0, sizeof(pCwDSP->i32SpeedTestSums_us) );
  memset( (void*)pCwDSP->i32SpeedTestCounts_us,0, sizeof(pCwDSP->i32SpeedTestCounts_us));
} // end CwDSP_ResetSpeedTestResults()



/* EOF < CwDSP.c > */


