// File:  C:\cbproj\Remote_CW_Keyer\StraightKeyDecoder.c
// Date:  2024-05-09
// Author:  Wolfgang Buescher (DL4YHF)
// Purpose: Morse code decoder for a 'straight keying' signal.
//          First used in the author's "Remote CW Keyer" project 
//          to check the CW keying signal on the SERVER SIDE
//          (which drives the radio with a "straight keying" signal,
//           not via separate "dash" and "dot" signals).
//          The API (callable "C" functions) is similar as the
//          "Elbug" module (Remote_CW_Keyer/Elbug.c), which can 
//          also emit finished characters (at the end of an inter-
//          character-space) in the same 'CW_CHR_...' format as specified
//          in Remote_CW_Keyer/Elbug.h .
//

#include <string.h> // no string functions used here, but memset() 

#include "Elbug.h" // header file for Elbug.c; also used for the "Straight Keying Decoder" !
     //    '--> stuff like Elbug_WordsPerMinuteToDotTimeInMilliseconds(),
     //                    Elbug_DotTimeInMillisecondsToWordsPerMinute(),
     //                    Elbug_MorseCodePatternToASCII().
     // Some of the above 'elbug' functions are also used by the
     // 'straight keying' decoder, and we don't want to reinvent the wheel.

#include "StraightKeyDecoder.h" // header file especially for StraightKeyDecoder.c

//---------------------------------------------------------------------------
void StraightKeyDecoder_Init( T_StraightKeyDecoder *pDecoder, int iDefaultDotTime_ms )
{ memset( pDecoder, 0, sizeof(T_StraightKeyDecoder) );
  pDecoder->iDotTime_ms = iDefaultDotTime_ms;
}

//---------------------------------------------------------------------------
int StraightKeyDecoder( T_StraightKeyDecoder *pDecoder,
       BOOL fKeyDown, // [in] input from straight key: TRUE = "key down", FALSE = "key up"
       long nMicrosecondsSinceLastCall ) // [in] ideally 2000 us, if windoze permits !
  // Values returned by StraightKeyDecoder() : One of the CW_CHR-codes (More code
  // pattern emitted at the end of a single character or space),
  // or one of the ELBUG_RESULT_.. values defined in Elbug.h, for example
  //               ELBUG_RESULT_NOTHING_DECODED  ( == 0 ) .
  // Called periodically from KeyerThread() .
{
  int iResult = ELBUG_RESULT_NOTHING_DECODED;

  pDecoder->i32TimeAccu_us += nMicrosecondsSinceLastCall;
  if( pDecoder->i32TimeAccu_us > 10000000L )  // avoid 32-bit integer overflow (limit to 10 seconds)
   {  pDecoder->i32TimeAccu_us = 10000000L;
   }
  if( pDecoder->fPrevKeyDown != fKeyDown )
   {  pDecoder->fPrevKeyDown =  fKeyDown;
      pDecoder->i32DebounceTime_us = 0;
   }
  else if( pDecoder->i32DebounceTime_us < 10000000L )
   {  pDecoder->i32DebounceTime_us += nMicrosecondsSinceLastCall;
   }
  if( pDecoder->i32DebounceTime_us >= 5000L ) // ignore any input-state-change shorter than X milliseconds
   {
     switch( pDecoder->iState )
      {
        case DS_WAITING:   // waiting for anything to happen (key was "up" so far)
           if( fKeyDown )
            { // Prepare to decode the next character:
              pDecoder->i32TimeAccu_us = 0;
              pDecoder->iState = DS_RCV_DOT;
              pDecoder->wShiftReg = 1; // place a new "start bit" in the shift register
              // (the position of the startbit encodes the number of dots and dashes
              //   in the LOWER BITS of the shift register later)

              // Let the caller of StraightKeyDecoder() know we just STARTED a new character:
              iResult |= ELBUG_RESULT_BEGIN_NEW_CHAR; // indicate "BEGIN of a new character"
                                                      // in the 'Timing Scope'.
            }
           break;
        case DS_RCV_DOT:    // currently receiving a dot (which MAY turn out a DASH soon)
           if( ! fKeyDown ) // key JUST RELEASED (debounced) -> time for action ..
            { pDecoder->wShiftReg = (pDecoder->wShiftReg << 1) | ( pDecoder->i32TimeAccu_us >= (pDecoder->iDotTime_ms * 2000) );
              // ,-----------------------------------------------'
              // '--> Bit values in the dash/dot-matrix : 0=dot 1=dash .
              //    > Bit 0 always contains the LAST TRANSMITTED dash/dot.
              //  Anything shorter than TWO NOMINAL DOT-TIMES is considered a dot.
              //  Anything else can only be a DASH. To test the decoder's tolerance,
              //  use e.g. 20 WPM on the CLIENT SIDE (sending from memory),
              //       and 35 WPM on the SERVER SIDE (decoding the 'straight key'
              //           signal from the timestamped keying commands).
              //    In that case, a single dot time becomes 35/20 = 1.75 dot times
              //    from the receiver's point of view, and if all works as planned
              //    will still be decoded properly.
              pDecoder->iState = DS_RCV_GAP;
              pDecoder->i32TimeAccu_us = 0;
            }
           break; // end case DS_RCV_DOT
        case DS_RCV_GAP:  // Receiving a GAP WITHIN A CHARACTER (which may turn out a SPACE later)
           if( fKeyDown ) // key just PRESSED AGAIN within two dot times -> next element in the same character
            { pDecoder->iState = DS_RCV_DOT;
              pDecoder->i32TimeAccu_us = 0;
              // NOT here: pDecoder->wShiftReg = 1;
            }
           else // key still UP -> is the "gap" the end of a character ?
            { // 3 dots is the ideal spacing of characters within a word,
              // 1 dot is the ideal spacing between dots and dashes within a character,
              // thus the only "threshold" can be a length of TWO dots:
              if( pDecoder->i32TimeAccu_us >= (pDecoder->iDotTime_ms * 2000) )
               { // two-dot gap -> reached the end of a CHARACTER, so "emit" it:
                 iResult |= pDecoder->wShiftReg;
                 pDecoder->wShiftReg = 1;          // prepare reception of the next character
                 // No-No: pDecoder->i32TimeAccu_us = 0; // do NOT reset the time before detecting e.g. a SPACE CHARACTER in DS_RCV_SPACE !
                 pDecoder->iState = DS_RCV_SPACE;  // finished receiving a CHARACTER, but now check for a SPACE after the character
               }
            } // end else < key still UP in state DS_RCV_GAP ? >
           break; // end case DS_RCV_GAP

        case DS_RCV_SPACE: // finished receiving a CHARACTER but didn't emit a trailing SPACE yet
           if( fKeyDown ) // key just PRESSED AGAIN before the WORD-space expired -> new character but NO SPACE (between words)
            { iResult = ELBUG_RESULT_BEGIN_NEW_CHAR;
              pDecoder->wShiftReg = 1;       // prepare reception of the next character
              pDecoder->i32TimeAccu_us = 0;
              pDecoder->iState = DS_RCV_DOT; // receivr the next CHARACTER, without a space in between
            }
           else // key still UP -> is it a "long gap" = SPACE after a WORD ?
            { // 3 dots is the ideal spacing of characters within a word,
              // 5 dots or more (ideally 6) means emit a SPACE character:
              if( pDecoder->i32TimeAccu_us >= (pDecoder->iDotTime_ms * 5000) )
               { // time to emit a SPACE :
                 iResult |= CW_CHR_SPACE;
                 pDecoder->iState = DS_WAITING;
                 pDecoder->i32TimeAccu_us = 0;
               }
            } // end else < key still UP in state DS_RCV_GAP ? >
           break; // end case DS_RCV_SPACE

        default:  // oops.. obviously a bug in the state machine
           break;
      } // end switch( pDecoder->iState )
   } // end if < not a "bouncing contact" >

  return iResult;

} // end StraightKeyDecoder()

/* EOF < StraightKeyDecoder.c > */

