//---------------------------------------------------------------------------
// File: C:\cbproj\Remote_CW_Keyer\RigControl.c
// Authors: Wolfgang Buescher (DL4YHF)
//       (the stripped-down variant for DL4YHF's "Remote CW Keyer"         
//        does NOT support the ugly "RS-BA1 protocol" and thus is MUCH     
//        easier to port to a simple microcontroller firmware) .           
// Date:   2024-02-18
// Purpose:
//  Initially just a parser for Icon's CI-V protocol, used for CAT control
//  over serial ports (including 'virtual COM ports' for USB).
//  Unlike the mimosa known as "Hamlib" (used in WSJT-X), RigControl.c
//  should work with ANY setting in the IC-7300's Setup / CI-V menu :
//        "Setup" .. "Connectors" .. "CI-V" :
//          "CI-V USB Port" : "Unlink from [REMOTE]" or "Link to [REMOTE]"
//          "CI-V USB Echo Back" : Either ON or OFF (WSJT-X needed "ON").
//
// Literature:
//  [IC7300FM]: Icom IC-7300 "Full Manual", Section 19 : "Control Command"
//              (contains the full set of CI-V commands in this radio)
//               C:\datasheets\Rig_Manuals\IC-7300\IC-7300_Full_Manual.pdf
//  [IC7610CIV]: Icom IC-7610 "CI-V Reference Guide", #A7380-7EX (2017)
//  [IC-R8600CIV]: IC-R8600 (RX) "CI-V Reference..",  #A7375-2EX-3a (2018)
//  [G3NRW_CIV]: IC-7300 TechNote: CI-V Controls  The Big Picture, G3NRW
//  [DF4OR_CIV]: http://www.plicht.de/ekki/civ/civ-p33.html .
//
// Revision History:
//   2018-12 : Created for Spectrum Lab by DL4YHF . Original location:
//              C:\cbproj\SpecLab\RigControl\RigControl.c .
//   2024-02 : Made a stripped-down variant for the "Remote CW Keyer":
//              C:\cbproj\Remote_CW_Keyer\RigControl.c ( + *.h ) .
//              Threw out the support for Icom's undocumented/proprietary
//              "RS-BA1" protocol (that's the thing with those THREE
//              UDP PORTS used when connected to the radio via LAN (IC-9700)
//              or WLAN (IC-705). If you need all these whistles and bells,
//              try to let WFVIEW run side-by-side with the Remote CW Keyer.
//            
//---------------------------------------------------------------------------

#include "switches.h" // project specific 'compilation switches'

#include <string.h>     // memset() isn't really a STRING function, but..
#include <stdio.h>      // snprintf() isn't "I/O", but declared in string.h
#include <math.h>       // fmodl() ..
#include "Timers.h"     // high-resolution stopwatches, etc
#include "Utilities.h"  // UTL_FormatDateAndTime(), UTL_GetCurrentUnixDateAndTime(), etc
#include "Inet_Tools.h" // INET_GetU16FromLE(), etc
#include "StringLib.h"  // DL4YHF's string library (designed for MICROCONTROLLERS)
#include "RigControl.h" // header file for THIS module

// Internal defines (only used within THIS module, thus not in the header file)
#define RIGCTRL_CLIENT_ACTION_NONE          0 // for iClientAction : "do nothing"
#define RIGCTRL_CLIENT_ACTION_RETRANSMIT    1 // immediately retransmit command from client to radio
#define RIGCTRL_CLIENT_ACTION_ENTER_FIFO    2 // store command in FIFO, retransmit or execute later
#define RIGCTRL_CLIENT_ACTION_FAKE_RESPONSE 3 // send a "fake response" without forwarding anything to the radio

#define RIGCTRL_MAX_PINGS_LOGGED 5
#define RIGCTRL_MAX_IDLES_LOGGED 5

// Internal forward references (because C compilers are still stupid..)
void RigCtrl_ForgetScopeFrequencyRanges( T_RigCtrlInstance *pRC );
BOOL RigCtrl_SetVFOFrequency_Internal( T_RigCtrlInstance *pRC, double dblFreqHz );
BOOL RigCtrl_SetOperatingMode_Internal( T_RigCtrlInstance *pRC, int iOpMode );
void RigCtrl_SendReadCommandFromFIFO( T_RigCtrlInstance *pRC );
BOOL RigCtrl_RegisterCIVReponseForClient(T_RigCtrlInstance *pRC, int iRigCtrlClientPort, T_RigCtrl_MsgFilter *pFilter);
int  RigCtrl_GetCombinedCmdAndSubcodeForCIV( BYTE *pbMessage, int iMsgLength );
T_RigCtrl_ParamInfo * RigCtrl_GetInfoForUnifiedParameterNumber(int iUnifiedPN);
T_RigCtrl_RadioInfo * RigCtrl_GetRadioInfoByDefaultAddress( int iRadioCtrlProtocol, int iDefaultAddress );
void RigCtrl_IntToBCD_CIV( long i32Value, BYTE *pbDest, int nDigits );

BOOL IcomPkt_SendToken(T_RigCtrlInstance *pRC, BYTE magic, char *pszComment);
BOOL IcomPkt_SendConnInfo( T_RigCtrlInstance *pRC, int iPortType );
void IcomPkt_PeriodicHandler( T_RigCtrlInstance *pRC, int iPortType );
int _stdcall RigCtrl_TxFunctionForCivViaUDP( T_RigCtrlInstance *pRC, int iRigCtrlPort, BYTE *source, int n_bytes,
                              int fWaitForCompletion, int *piErrorCode, int *piNumMsWaited);


//----------------------------------------------------------------------------
// Global variables for hardcore debugging / post-mortem crash analysis.
//        Useful when the debugger's call stack shows nothing but garbage,
//        e.g. after an exception at the infamous address 0xFEEEFEEE .
//        ( C++Builder's debugger could still inspect SIMPLE GLOBAL
//          VARIABLES after most kinds of exceptions, but of course nothing
//          stack-based because the CPU- or task-stack was usually trashed.)
//----------------------------------------------------------------------------
#if(SWI_HARDCORE_DEBUGGING) // (1) = hardcore-debugging, (0)=normal compilation
 int RigCtrl_iLastSourceLine = 0; // WATCH THIS after crashing with e.g. "0xFEEEFEEE"  ...
# define HERE_I_AM__RIGCTRL()  RigCtrl_iLastSourceLine=__LINE__
# if( 0 )
    If certain worker threads of the application don't seem to terminate
    "politely", watch the following (complete list not HERE but in Keyer_Main.cpp):
     * GUI_iLastSourceLine
     * DSW_iLastSourceLine
     * DSP_iLastSourceLine
     * Keyer_iLastSourceLine
     * CwNet_iLastSourceLine
     * HttpSrv_iLastSourceLine
     * RigCtrl_iLastSourceLine
#  endif // (0) .. kludge to inspect the above variables in Borland via mouse-over ...
  CPROT void CheckSystemHealth(const char *pszModuleName, int iSourceLine); // <- periodically called via macro CHECK_SYSTEM_HEALTH()
  static  const char *s_ModuleName = "RigCtrl";
# define CHECK_SYSTEM_HEALTH() CheckSystemHealth(s_ModuleName,__LINE__)
  BYTE *RigCtrl_pbSource = NULL;
  int   RigCtrl_iScopeBin= 0;
  float *RigCtrl_pfltDest= NULL;
#else
# define HERE_I_AM__RIGCTRL()  /* nuffink   */
# define CHECK_SYSTEM_HEALTH() /* hamwanich */
#endif // SWI_HARDCORE_DEBUGGING ?


//----------------------------------------------------------------------------
// Global variables for troubleshooting CERTAIN RIGS ..
//----------------------------------------------------------------------------

BOOL RigControl_fUpdateTrafficLog = FALSE; // flag signalling "call TDebugForm::UpdateTrafficLog() a.s.a.p" (or whoever DISPLAYS the traffic log on-screen, and CLEARS that flag when 'done')
long RigControl_i32MessageNrForTrafficLog = 0;
int  RigControl_iTrafficMonitorDisplayOptions = RIGCTRL_TRAFFIC_DISPLAY_OPTION_DISABLED;  // RigCtrl_AddToTrafficLog() disabled unless the application sets at least RIGCTRL_TRAFFIC_DISPLAY_OPTION_ENABLE
int  RigControl_iTrafficMonitorExtraColumns = RIGCTRL_TRAFFIC_EXTRA_COLUMN_TIME
                                            | RIGCTRL_TRAFFIC_EXTRA_COLUMN_TX_RX
                                            | RIGCTRL_TRAFFIC_EXTRA_COLUMN_PAYLOAD_LENGTH
                                            | RIGCTRL_TRAFFIC_EXTRA_COLUMN_COMMENTS;
     // The two variables above will be set to UConfig.iTrafficMonitorDisplayOptions and UConfig.iTrafficMonitorExtraColumns; somewhere in SpecMain.cpp.
     // But the above INITIALISATION VALUES will be used as DEFAULTS in Config.cpp : CFG_Load() !
BOOL RigControl_fTrafficMonitorPaused = FALSE; // FALSE=not "automatically paused on a trigger event", TRUE="something happened, PLEASE TAKE A LOOK.."
     // '--> Reason for "pausing" / "stopping" : e.g. RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_FULL_FIFO / .. STOP_ON_ERROR / ...


// Implementation of functions


//--------------------------------------------------------------------------
void RigControl_Init( T_RigCtrlInstance *pRC,
        int iRadioControlProtocol, // [in] e.g. RIGCTRL_PROTOCOL_ICOM_CI_V
        T_CFIFO *pRadioPortRxFifo, // [in] thread-safe circular FIFO to RECEIVE data from the radio
        T_CFIFO *pRadioPortTxFifo, // [in] thread-safe circular FIFO to TRANSMIT data to the radio
        int nClientPorts,       // [in] 0..RIGCTRL_MAX_CLIENT_PORTS
        int *piClientOptions )  // [in] array, one per client (*)
  // Initializes everything for a "single instance" .
  //
  // (*) These "clients" were initially called "CI-V port duplicators".
  //     Connected to extra "COM ports" (in most cases, "com0com" virtual NULL MODEM CABLES).
  //     Later renamed into "Client #1" and "Client #2", at least for the
  //     configuration dialog in Spectrum Lab. Each of those clients can have
  //     it's individual configuration, thus using an ARRAY here !
{
  int i, iClient, iRigCtrlPort, iClientOptions;
  T_RigCtrl_ParamInfo *pPI;

  if( nClientPorts > RIGCTRL_MAX_CLIENT_PORTS )
   {  nClientPorts = RIGCTRL_MAX_CLIENT_PORTS; // oops.. more than expected !?
   }

  memset( pRC, 0, sizeof( T_RigCtrlInstance ) );
  pRC->initState = initState_ColdStart;     // initState_ColdStart /  _Opened / _Closed ?
  pRC->iRadioCtrlProtocol = iRadioControlProtocol;
  pRC->pRadioPortTxFifo = pRadioPortTxFifo; // thread-safe circular FIFO to TRANSMIT data to the radio
  pRC->pRadioPortRxFifo = pRadioPortRxFifo; // thread-safe circular FIFO to RECEIVE data from the radio

  pRC->u32RxAudioSamplesPerSecond = pRC->u32TxAudioSamplesPerSecond = 48000;
  pRC->dblUnixTimeOfFirstSpectrum = 0.0;

  for(iClient=0; iClient<nClientPorts; ++iClient)
   { iRigCtrlPort = iClient + RIGCTRL_PORT_CLIENT_1;
     iClientOptions = piClientOptions[iClient]; // individual settings for each client !
     pRC->PortInstance[iRigCtrlPort].iClientOptions = iClientOptions;
     pRC->PortInstance[iRigCtrlPort].fEmulateCIVEcho = (iClientOptions & RIGCTRL_CLIENT_OPTION_EMULATE_CIV_ECHO) != 0;
   }
  pRC->iRigControlFlags = RIGCTRL_FLAGS_NONE; // don't SEND but READ setttings from the remote rig
  pRC->iCapabilities = RIGCTRL_CAPS_ALL; // be optimistic about FUTURE rigs .. IC-9700 anyone ?
  pRC->iMaxResponseDelay_ms = 1500; // maximum delay for getting a response, in milliseconds
  pRC->dblVfoFrequency = 0.0;
  pRC->iOpMode = RIGCTRL_OPMODE_UNKNOWN; // later: USB, LSB, CW, ...
  pRC->iParameterPollingState = RIGCTRL_POLLSTATE_PASSIVE; // don't do anything until the application calls RigControl_StartReading()
  pRC->iResponseCountdown_ms = 0;
  pRC->iScopeSpeed = RIGCTRL_SCOPE_SPEED_FAST;

  // Set everything that we MAY(!) read from the radio via CI-V to "unknown" / "not read yet".
  // This way, the "report" in SL's debug-window / CAT traffic monitor
  //      can spit out THOSE PARAMETERS THAT HAVE REALLY BEEN ACCESSED,
  //      and omits unused entries (in RigCtrl_ParameterInfo[]) .
  pPI = (T_RigCtrl_ParamInfo*)RigCtrl_ParameterInfo;
  while( pPI->iUnifiedPN != 0 )
   { if( ( pPI->iByteOffsetIntoRigCtrlInstance > 0 )
       &&( pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { switch ( pPI->iRigCtrlDataType )
         { case RIGCTRL_DT_INT  :
              *(int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = RIGCTL_NOVALUE_INT;
              break;
           case RIGCTRL_DT_DOUBLE:
              *(double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = RIGCTL_NOVALUE_DOUBLE;
              break;
           default:
              break;
         } // end switch ( pPI->iRigCtrlDataType )
      }
     ++pPI;
   } // end while < all "unified parameters" >

  RigCtrl_ForgetScopeFrequencyRanges( pRC );
  pRC->initState = initState_Opened; // initState_ColdStart /  _Opened / _Closed ? here: 'Opened' when returning from RigControl_Init()

} // end RigControl_Init()


//-------------------------------------------------------------------- API --
void RigControl_Close( T_RigCtrlInstance *pRC ) // closes handles, sockets, frees resources
  // Called internally when necessary, or via SL's main menu: "Start/Stop" .. "Rig Control" .. "Start" .
{
  // Won't stop the threads running in IOaccess.cpp, but will stop functions called from there:
  //    (closing those threads, and closing "COM" ports isn't RigControl's part
  //     because RigControl neither opened those ports or created those threads) :
  if( pRC->initState == initState_Opened )
   {  pRC->initState = initState_Closed; // here: 'Closed' in
     // Give any RigControl-subroutines and "handlers" the chance to return,
     // before closing sockets, freeing memory, and making pointers invalid:
     Sleep( 50 ); // should be long enough to return from e.g RigControl_Handler() to IO_RadioCommThdFunc() .
   }

} // end RigControl_Close()


//-------------------------------------------------------------------- API --
void RigControl_ReConnect( T_RigCtrlInstance *pRC )
  // Politely closes an existing connection (if any),
  // clears the 'counters' that limit certain messages from the 'traffic log',
  // and finally tries to re-connect the rig.
  // Keeps all previous settings from RigControl_SetLANParams() INTACT !
  //
  // Called for example when clicking the "Apply"-button on SL's 'Radio Control' panel,
  //        or via SL's main menu: "Start/Stop" .. "Rig Control" .. "Start" .
{ char sz80Msg[84], *cp;
  int iMsgTypeForTrafficLog;

  RigControl_Close( pRC );
  if( pRC->initState == initState_Closed )
   {  pRC->initState =  initState_Opened; // allow e.g. RigControl_ProcessIcomUDP(), RigControl_Handler(), and RigControl_ProcessRxData() to "work"
   }

  strcpy( sz80Msg, "RigControl: Re-connect" );  // show what's going on ..
  iMsgTypeForTrafficLog = RIGCTRL_MSGTYPE_ERROR;
  RigCtrl_AddToTrafficLog( pRC, 0/*iRigCtrlPort*/, NULL, 0, iMsgTypeForTrafficLog, sz80Msg );
  RigControl_StartReading( pRC ); // -> RigControl_OpenSocket() -> RigControl_LaunchUDPThread() ...

} // end RigControl_ReConnect()


//-------------------------------------------------------------------- API --
void RigControl_SetClientOptions( T_RigCtrlInstance *pRC, int iRigCtrlPort,
         int iClientOptions ) // [in] bitwise combineable options like ..
         //   RIGCTRL_CLIENT_OPTION_REJECT_UNSOLICITED_MSGS ,
         //   RIGCTRL_CLIENT_OPTION_EMULATE_CIV_ECHO, ..(?)
{
  if( (iRigCtrlPort>=RIGCTRL_PORT_RADIO) && (iRigCtrlPort<=/*!*/RIGCTRL_MAX_CLIENT_PORTS) )
   { pRC->PortInstance[iRigCtrlPort].iClientOptions = iClientOptions;
     pRC->PortInstance[iRigCtrlPort].fEmulateCIVEcho = (iClientOptions & RIGCTRL_CLIENT_OPTION_EMULATE_CIV_ECHO) != 0;
   }
} // end RigControl_SetClientOptions()


//--------------------------------------------------------------------------
void RigCtrl_ForgetScopeFrequencyRanges( T_RigCtrlInstance *pRC )
{  int iRange, iEdge;
   pRC->iScopeNumFreqRanges = pRC->iScopeNumEdgesPerFreqRange = 0;
   for( iRange=0; iRange<RIGCTRL_MAX_FIXED_EDGE_SCOPE_FREQ_RANGES; ++iRange)
    { for( iEdge=0; iEdge<RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE; ++iEdge)
       { pRC->ScopeFreqRange[iRange][iEdge].dblFmin_Hz = 0.0;
         pRC->ScopeFreqRange[iRange][iEdge].dblFmin_Hz = 0.0;
       }
    }
} // end RigCtrl_ForgetScopeFrequencyRanges()



//--------------------------------------------------------------------------
void RigControl_StartReading( T_RigCtrlInstance *pRC )
  // Prepares reading "all we need to know" about the remotely controlled rig,
  //  but doesn't really 'communicate' via serial port, LAN or WLAN yet.
  //  For Icom radios connected via LAN or WLAN, also tries to open the
  //  required ports (in this case UDP sockets), and prepares to log-in .
  // Called from the application (e.g. TConfigDlg::ApplyMyControls() ),
  //        or (if a simple SERIAL PORT is used for CI-V) IOaccess.cpp : IO_OpenComPorts(),
  //        or (if UDP is used for CI-V) called from "RigControl.c" itself,
  //   to begin the first PARAMETER POLLING CYCLE via CI-V .
  // Communication parameters like pRC->iRadioDeviceAddr, iRadioCtrlProtocol
  //        and  [removed] iInetProtocol ( INET_PROTOCOL_OFF/HTTP/TCP/UDP/RAW/COM_PORT)
  //   must have been set by the application prior to this.
  // The rest (communication) happens in RigControl_Handler(); details there.
{

  // 'Forget' various counters and values read during a previous session:
  pRC->iRadioPortTxFifoHead = pRC->iRadioPortTxFifoTail = 0;
  pRC->nCompleteSpectraReceived   = 0;
  pRC->dblUnixTimeOfFirstSpectrum = 0.0;
   
  pRC->iParameterPollingState    = RIGCTRL_POLLSTATE_START_RD;
  pRC->iParameterPollingSubState = pRC->iParameterPollingArrayIndex = 0; // don't leave anything to fate !
  pRC->iResponseCountdown_ms = 0;

} // end RigControl_StartReading()

//--------------------------------------------------------------------------
BOOL RigCtrl_SendAndLogMessage( T_RigCtrlInstance *pRC,
        int iRigCtrlPort, // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 ?
        BYTE *pbMessage, int iMsgLength, // [in] message, including pre- and postamble but not the UDP-CIV-HEADER (!)
        BOOL fWaitForCompletion, // [in] TRUE=yes (blocking call), FALSE=no .
        int iMsgType,     // [in] message type category and 'flags', only for the "traffic log"
        char *pszComment) // [in] human readable info, short enough for the "traffic log"
  // Sends any CI-V message, and (if successful) counts and optional logs it.
{
  int iResult, iErrorCode, iNumMsWaited; // optional results from IO_SendToCOMPort()
  char sz255Msg[256];

  if( pRC->iRadioCtrlProtocol == RIGCTRL_PROTOCOL_ICOM_CI_V )
   { iMsgType |= RIGCTRL_MSGTYPE_FLAG_CIV; // <- for the TRAFFIC LOG: pbMessage points to pure CI-V payload, without overhead from any "transport layer" (and no "UDP-CIV-Header")
     pRC->pszCommentForTxFunction = pszComment; // <- added 2023-07 to avoid yet another parameter for the (UDP-) transmit function called from RigCtrl_SendAndLogMessage()
   }
  pRC->iMsgTypeForTxFunction = iMsgType;   // <- added 2023-07 so RigCtrl_TxFunctionForCivViaUDP() can pass 'iMsgType' on to RigCtrl_AddToTrafficLog(),
                                           //    especially for flags like RIGCTRL_MSGTYPE_FLAG_SUGGEST_NEW_HEADLINE_FOR_TRAFFIC_LOG .

  if( (iRigCtrlPort==RIGCTRL_PORT_RADIO) && (pRC->pRadioPortTxFifo!=NULL) )
   { iResult = CFIFO_Write( pRC->pRadioPortTxFifo, pbMessage, iMsgLength );
     if( iResult >= 0 ) // only a NEGATIVE result indicates a 'real' error. Zero = FALSE = "not finished yet"
      { // Modified 2023-08-15 : If the port is a COM port, RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD
        // is meaningless . Break up the too complex 'if' condition further below.
        BOOL fAllowTrafficLog = ( RigControl_iTrafficMonitorDisplayOptions
                              & RIGCTRL_TRAFFIC_DISPLAY_OPTION_ENABLE) != 0;
        if( pRC->dwRejectMessages4Log & (1<<iMsgType) )
         { fAllowTrafficLog = FALSE;
         }
        if( RigControl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD )
         { //  With RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD SET, RigCtrl_TxFunctionForCivViaUDP() will have
           //        already called RigCtrl_AddToTrafficLog(), for reasons explained THERE.
           // Hmmm... but DO WE USE UDP AT ALL ?
         } // end if .. RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD ?
        if( fAllowTrafficLog )
         {  RigCtrl_AddToTrafficLog( pRC, iRigCtrlPort, pbMessage, iMsgLength, iMsgType | RIGCTRL_MSGTYPE_FLAG_TX, pszComment );
         }
        ++pRC->PortInstance[iRigCtrlPort].dwNumMessagesSent;
        pRC->PortInstance[iRigCtrlPort].fLastSendFailed = FALSE;
        return TRUE;
      }
     else // CFIFO_Write( pRC->pRadioPortTxFifo ) failed .. "tried to send more than possible" ?
      { if( !pRC->PortInstance[iRigCtrlPort].fLastSendFailed )
         {   pRC->PortInstance[iRigCtrlPort].fLastSendFailed = TRUE;
           snprintf( sz255Msg,250, "Failed to send to %s: code %d",
                RigCtrl_RigControlPortNrToString(iRigCtrlPort), iErrorCode );
           RigCtrl_AddToTrafficLog( pRC, iRigCtrlPort, NULL, 0, RIGCTRL_MSGTYPE_ERROR, sz255Msg );
         }
      }
   } // end if < something 'sendable' > ?
  return FALSE;

} // end RigCtrl_SendAndLogMessage()


//--------------------------------------------------------------------------
BOOL RigCtrl_SendShortCIVMessage( T_RigCtrlInstance *pRC,
        int iRigCtrlPort, // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 ?
        int iMsgType,     // [in] message type category and 'flags', only for the "traffic log"
        char *pszComment, // [in] human readable info, short enough for the "traffic log"
        int iCmd,         // [in] CI-V main command like 0x27 for "Scope waveform data"
        int iSubCmd,      // [in] optional sub-command.        -1 to omit.
        int iDataByte1,   // [in] optional 1st parameter byte. -1 to omit.
        int iDataByte2 )  // [in] optional 2nd parameter byte. -1 to omit.
  // Sends a 'short' CI-V message, usually a 'request' for something.
  // To send WITHOUT a "sub-command" (byte), use iSubCmd = -1.
  // To send WITHOUT a "data byte", use iDataByte1 = -1,  etc.
  // Returns TRUE when successfully sent, otherwise FALSE. KISS.
  // In Spectrum Lab, sent messages MAY be displayed by [v] Show CAT traffic
  //    with a YELLOW background under 'View/Windows' / 'CAT traffic monitor'.

{
  BYTE b16TxBuffer[16];
  BYTE *pbBufPtr = b16TxBuffer;
  int  iMsgLength;
  *pbBufPtr++ = 0xFE;  // fixed preamble code for CI-V protocol (term used by Icom .. NOT "Begin of Message" ! )
  *pbBufPtr++ = 0xFE;  // fixed preamble code for CI-V protocol
  *pbBufPtr++ = (BYTE)pRC->iRadioDeviceAddr;  // "to" transceiver's address
  *pbBufPtr++ = (BYTE)pRC->iRadioMasterAddr;  // "from" controller's address
  if( iCmd >= 0 )
   { *pbBufPtr++ = (BYTE)iCmd;  // e.g. 0x03 = CI-V command "Read operating frequency"
   }
  if( iSubCmd >= 0 )
   { *pbBufPtr++ = (BYTE)iSubCmd;
   }
  if( iDataByte1 >= 0 )
   { *pbBufPtr++ = (BYTE)iDataByte1;
   }
  if( iDataByte2 >= 0 )
   { *pbBufPtr++ = (BYTE)iDataByte2;
   }
  *pbBufPtr++ = 0xFD;  // CI-V code for "End Of Message" (term used by Icom .. NOT "postamble", even though 0xFE is called "preamble" ! )

  iMsgLength = pbBufPtr - b16TxBuffer;
  return RigCtrl_SendAndLogMessage( pRC, iRigCtrlPort, b16TxBuffer, iMsgLength,
          FALSE/*fWaitForCompletion*/, RIGCTRL_MSGTYPE_FLAG_CIV | iMsgType, pszComment );
} // end RigCtrl_SendShortCIVMessage()


//--------------------------------------------------------------------------
void RigCtrl_AppendByteToRadioPortTxFifo( BYTE *pbBuffer, int *piHeadIndex, BYTE b )
{
  *piHeadIndex &= (RIGCTL_RADIO_PORT_TX_FIFO_SIZE-1); // safety first..
  pbBuffer[*piHeadIndex] = b;
  *piHeadIndex = (*piHeadIndex + 1) & (RIGCTL_RADIO_PORT_TX_FIFO_SIZE-1);
}

//--------------------------------------------------------------------------
BYTE RigCtrl_ReadByteFromRadioPortTxFifo( BYTE *pbBuffer, int *piTailIndex )
{
  BYTE b;
  *piTailIndex &= (RIGCTL_RADIO_PORT_TX_FIFO_SIZE-1); // safety first..
  b = pbBuffer[*piTailIndex];
  *piTailIndex = (*piTailIndex + 1) & (RIGCTL_RADIO_PORT_TX_FIFO_SIZE-1);
  return b;
}


//--------------------------------------------------------------------------
BOOL RigCtrl_AppendMessageToRadioPortTxFifo( T_RigCtrlInstance *pRC,
        int iRigCtrlClientPort, // [in] RIGCTRL_PORT_CLIENT_1 / 2 ?
        BYTE *pbMessage, int iMsgLength ) // [in] message including pre- and postamble
  // Called from RigCtrl_ParseCIV() after receiving certain messages
  // from any CLIENT that need to be forwarded to the RADIO PORT,
  // but cannot be sent immediately because the RADIO is already busy
  // from a previously transmitted command (with response is still pending).
  //
{
  int nBytesOccupied, nBytesFree;
  int iHeadIndex = pRC->iRadioPortTxFifoHead;
  int iTailIndex = pRC->iRadioPortTxFifoTail;
  BYTE *pbBuffer = pRC->bRadioPortTxFifo;

  if( iMsgLength <= 0 )  // don't trash the TX-FIFO with negative lengths
   { return FALSE;
   }

  nBytesOccupied = iHeadIndex - iTailIndex;
  if( nBytesOccupied < 0 ) // circular wrap, e.g. head=0, tail=2047 ->
   {  nBytesOccupied += RIGCTL_RADIO_PORT_TX_FIFO_SIZE;
   }
  // Remember, a classic circular 2048-byte-FIFO has a maximum USEABLE capacity
  //   of 2047(!) bytes, because HeadIndex==TailIndex means buffer EMPTY (not FULL).
  //   The buffer is completely full when (HeadIndex+1)==TailIndex. Thus:
  nBytesFree = RIGCTL_RADIO_PORT_TX_FIFO_SIZE - 1 - nBytesOccupied;

  // TWO extra bytes (before the message itself) are used to store the LENGTH.
  // ONE additional byte is used to store the CLIENT PORT NUMBER on which
  //     the message was originally received .
  // This way, we can store anything in the FIFO without knowing the structure.
  // It doesn't have to be CI-V !
  if( nBytesFree < (iMsgLength+3) ) // bad luck, running out of TX-FIFO-space !
   { return FALSE;
   }

  // Enter the MESSAGE LENGTH and CLIENT PORT in the FIFO (low byte first as on an Intel CPU)
  RigCtrl_AppendByteToRadioPortTxFifo( pbBuffer, &iHeadIndex, (BYTE)iMsgLength );
  RigCtrl_AppendByteToRadioPortTxFifo( pbBuffer, &iHeadIndex, (BYTE)(iMsgLength >> 8) );
  RigCtrl_AppendByteToRadioPortTxFifo( pbBuffer, &iHeadIndex, (BYTE)iRigCtrlClientPort );

  while( (iMsgLength--) > 0 )
   { RigCtrl_AppendByteToRadioPortTxFifo( pbBuffer, &iHeadIndex, *(pbMessage++) );
   }
  pRC->iRadioPortTxFifoHead = iHeadIndex; // another fragment is now "occupied"

  return TRUE;  // successfully appended the message to the TX-FIFO

} // end RigCtrl_AppendMessageToRadioPortTxFifo()

//--------------------------------------------------------------------------
BOOL RigCtrl_SendMessageFromRadioPortTxFifo( T_RigCtrlInstance *pRC,
        T_RigCtrl_MsgFilter *pFilter) // [out] command, subcommand, and possibly accepted message type
  // Transmits the next message (if any) that has previously been
  //           queued up in the *RADIO PORT's* transmit FIFO, TO the radio
  //   - see details in RigCtrl_AppendMessageToRadioPortTxFifo() .
  // To register the expected RESPONSE (from radio, forwarded to the client),
  //    RigCtrl_SendMessageFromRadioPortTxFifo() also prepares a 'filter'
  //    which the caller will pass on to RigCtrl_RegisterCIVReponseForClient(),
  //    just as if the message HAD BEEN immediately forwarded
  //    from client to radio in RigCtrl_ParseCIV() or similar.
{
  int nBytesOccupied, iMsgLength, iClientPort, i;
  int iHeadIndex = pRC->iRadioPortTxFifoHead;
  int iTailIndex = pRC->iRadioPortTxFifoTail;
  BYTE *pbBuffer = pRC->bRadioPortTxFifo;
  T_RigCtrl_PortInstance *pPort = &pRC->PortInstance[ RIGCTRL_PORT_RADIO ];
  BYTE *pbMessage = pPort->bTxMsg; // temporary 'linear' buffer for transmission
  T_RigCtrl_MsgFilter msg_filter;
  int  iCombinedCmdAndSubcode;

  nBytesOccupied = iHeadIndex - iTailIndex;
  if( nBytesOccupied < 0 ) // circular wrap, e.g. head=0, tail=2047 ->
   {  nBytesOccupied += RIGCTL_RADIO_PORT_TX_FIFO_SIZE;
   }
  if( nBytesOccupied >= 3 ) // need AT LEAST three bytes : 16 bit length + 8 bit client port number
   { // Read the MESSAGE LENGTH and CLIENT PORT in the FIFO (again, low byte first)
     iMsgLength  = RigCtrl_ReadByteFromRadioPortTxFifo( pbBuffer, &iTailIndex );
     iMsgLength |= ((int)RigCtrl_ReadByteFromRadioPortTxFifo( pbBuffer, &iHeadIndex) << 8);
     iClientPort = RigCtrl_ReadByteFromRadioPortTxFifo( pbBuffer, &iTailIndex );
     if( nBytesOccupied >= (3+iMsgLength) )  // still looks good .. queued-up message is COMPLETE
      {
        if( iMsgLength <= RIGCTRL_MAX_MESSAGE_LENGTH ) // must fit inside the "non-circular" TX buffer..
         { for(i=0; i<iMsgLength; ++i)
            { pbMessage[i] = RigCtrl_ReadByteFromRadioPortTxFifo( pbBuffer, &iTailIndex );
            }
         }
        pRC->iRadioPortTxFifoTail = // release transmitted message from the circular FIFO
          ( pRC->iRadioPortTxFifoTail + iMsgLength + 3 ) & (RIGCTL_RADIO_PORT_TX_FIFO_SIZE-1);

        // The following is similar to the code after
        //      'case RIGCTRL_CLIENT_ACTION_RETRANSMIT' in RigCtrl_ParseCIV(),
        //      incl. registering a "filter" for this message to forward its RESPONSE
        //      to the client (later, when a matching response arrived from the radio):
        msg_filter.iCmd = pbMessage[4];
        // any further bytes in any CI-V message are IGNORED for filtering:
        msg_filter.iSubCmd = msg_filter.iDataB1 = msg_filter.iDataB2 = -1;
        // Also let the filter (for the client) pass reponse "OK" and "NOT OK":
        msg_filter.iMsgTypes = RIGCTRL_MSGTYPE_OK | RIGCTRL_MSGTYPE_NOT_OK;
        iCombinedCmdAndSubcode = RigCtrl_GetCombinedCmdAndSubcodeForCIV( pbMessage, iMsgLength );
        if( pRC->pRadioPortTxFifo != NULL )
         { RigCtrl_RegisterCIVReponseForClient( pRC, iClientPort, &msg_filter );
           // Mark the "radio port" as occupied, until the response
           //      for the forwarded command (sent below) arrives :
           pRC->iExpectedResponseOnRadioPort = iCombinedCmdAndSubcode;
           return RigCtrl_SendAndLogMessage( pRC, RIGCTRL_PORT_RADIO, pbMessage, iMsgLength,
                    TRUE/*wait for completion*/, RIGCTRL_MSGTYPE_OTHER, "fwd from FIFO" );
         } // end if < pRadioPortTxFifo != NULL > ?
      } // end if( nBytesOccupied >= (3+iMsgLength) )
   }
  // Arrived here ? The TX-FIFO is either completely empty, or the next message
  //        has not been completely assembled by the "writer" in another thread,
  //        or the address of the "transmit function" for this port is invalid.
  return TRUE;

} // end RigCtrl_SendMessageFromRadioPortTxFifo()




//--------------------------------------------------------------------------
BOOL RigCtrl_RegisterCIVReponseForClient( T_RigCtrlInstance *pRC,
        int iRigCtrlClientPort,       // [in] RIGCTRL_PORT_CLIENT_1 / 2 ?
        T_RigCtrl_MsgFilter *pFilter) // [in] command, subcommand, and possibly accepted message type
  // Called from RigCtrl_ParseCIV() when certain messages (commands)
  //        arrive on one of the *CLIENT PORTS* (iRigCtrlClientPort).
  // Because the client "thinks" he's directly talking to the radio,
  //         anything that the radio sends IN RESPONSE must also be
  //         forwarded (not echoed) to the client's port (iRigCtrlClientPort).
  // For CI-V, the response may be an unspecific "Ok" or "Error".
  //         To allow forwarding such responses to the client, too,
  //         the caller may also specify a CATEGORY of message types, for example
  //         RIGCTRL_MSGTYPE_OK | RIGCTRL_MSGTYPE_NOT_OK (bitwise combineable).
{
  T_RigCtrl_PortInstance *pPortInstance;
  T_RigCtrl_MsgFilter    *pMsgFilter;

  int i;
  // Example: Call after receiving "Power On"-command, 1st frame from "RS-BA1":
  //    pbMessage[0  1  2  3   4   5   6  7 ...
  //         Len Pre.  Addr   cmd sub  d1 d2
  // >    r1 007 FE FE 94 E0   18  01        client to server) -> pFilter->iCmd = 0x18
  // >    TX 007 FE FE 94 E0   18  01        server lab to IC-7300
  // >    RX 007 FE FE 94 E0   18  01        echo directly from IC-7300
  // >    RX 006 FE FE E0 94   FB            IC-7300 to server: "OK", must pass the new filter because it's the "expected response"
  // >    t1 006 FE FE E0 94   FB            response relayed from server to client

  if( (iRigCtrlClientPort<0) || (iRigCtrlClientPort>/*!*/RIGCTRL_MAX_CLIENT_PORTS) )
   { return FALSE;  // not a valid 'port identifier'
   }

  pPortInstance = &pRC->PortInstance[iRigCtrlClientPort];

  // First check if this message (command, subcommand, sub-sub-command) is already registered:
  for( i=0; i<RIGCTRL_MAX_REGISTERED_MESSAGES_PER_CLIENT; ++i )
   { pMsgFilter = &pPortInstance->RegisteredMessages[i];
     if( ( pMsgFilter->iCmd    == pFilter->iCmd )
      && ( pMsgFilter->iSubCmd == pFilter->iSubCmd )
      && ( pMsgFilter->iDataB1 == pFilter->iDataB1 )
      && ( pMsgFilter->iDataB2 == pFilter->iDataB2 ) )
      {
        pMsgFilter->iMsgTypes |= pFilter->iMsgTypes; // e.g. "OK" + "NOT OK"
        return TRUE;
      }
   }

  // Not registered yet, so use the FIRST unused entry :
  for( i=0; i<RIGCTRL_MAX_REGISTERED_MESSAGES_PER_CLIENT; ++i )
   { pMsgFilter = &pPortInstance->RegisteredMessages[i];
     if( (pMsgFilter->iCmd <= 0 ) && (pMsgFilter->iMsgTypes <= 0 ) )
      { *pMsgFilter = *pFilter;
        return TRUE;
      }
   }

  // Running out of space to "register" another message for this client..
  return FALSE;


} // end RigCtrl_RegisterCIVReponseForClient()


//--------------------------------------------------------------------------
void RigCtrl_CheckAndForwardResponseToClients( T_RigCtrlInstance *pRC,
        BYTE *pbMessage, int iMsgLength, // [in] message including pre- and postamble
        T_RigCtrl_MsgFilter *pFilter,    // [in] command, subcommand, and possibly accepted message type
        int iReceivedMessageType,        // [in] class of the received message for the log
        BOOL fIsUnsolicitedMsg,          // [in] TRUE for unsolicited message, e.g. VFO frequency + Waveform data
        char * pszComment )              // [in] optional comment for the log
{
  int iRigCtrlClientPort, iFilter;
  T_RigCtrl_PortInstance *pPortInstance;
  T_RigCtrl_MsgFilter *pMsgFilter;
  BOOL fMatch;

  for( iRigCtrlClientPort = RIGCTRL_PORT_CLIENT_1;
       iRigCtrlClientPort < (RIGCTRL_PORT_CLIENT_1+RIGCTRL_MAX_CLIENT_PORTS);
       iRigCtrlClientPort++ )
   { pPortInstance = &pRC->PortInstance[iRigCtrlClientPort];

     // Does THIS client tolerate 'unsolicited' messages (e.g. new VFO-frequecy, Scope data) ?
     // A well-written CI-V parser must ignore messages he doesn't understand,
     //    because CI-V has once been designed as a multi-master bus system.
     //    Not so with the mimosa known as "Hamlib" (used per default in WSJT-X).
     if( fIsUnsolicitedMsg && (pPortInstance->iClientOptions & RIGCTRL_CLIENT_OPTION_REJECT_UNSOLICITED_MSGS) )
      { // Don't forward this unsolitited message to the client.
        // This stupid Extrawurst was added to prevent WSJT-X to bail out
        // with the oh-so-familar "Ding-Dong ! Rig Control Error" :
        continue;  // continue with the next client . Do NOT forward to the mimosa .
      }

     // Does the received message match any filter registered by THIS client ?
     fMatch = FALSE;
     for( iFilter=0; iFilter<RIGCTRL_MAX_REGISTERED_MESSAGES_PER_CLIENT; ++iFilter )
      { pMsgFilter = &pPortInstance->RegisteredMessages[iFilter];
        if( ( ( pMsgFilter->iCmd    == pFilter->iCmd    ) || ( pMsgFilter->iCmd < 0) )
         && ( ( pMsgFilter->iSubCmd == pFilter->iSubCmd ) || ( pMsgFilter->iSubCmd<0) )
         && ( ( pMsgFilter->iDataB1 == pFilter->iDataB1 ) || ( pMsgFilter->iDataB1<0) )
         && ( ( pMsgFilter->iDataB2 == pFilter->iDataB2 ) || ( pMsgFilter->iDataB2<0) ) )
         { fMatch = TRUE;
           break;
         }
        if( pMsgFilter->iMsgTypes == pFilter->iMsgTypes )
         { fMatch = TRUE;
           break;
         }
        // If the message received on the "RADIO"-port was "OK" or "NOT OK",
        //    pMsgFilter->iCmd will not match (e.g. client was asking for 0x18 = "POWER ON"
        //     but radio answered with Cmd=0xFB = "ok"). To pass such "expected" messages,
        //     when expecting "OK" or "NOT OK", ANY of these will match:
        if(  ((pMsgFilter->iMsgTypes & (RIGCTRL_MSGTYPE_OK|RIGCTRL_MSGTYPE_NOT_OK)) != 0)
          && ((pFilter->iMsgTypes    & (RIGCTRL_MSGTYPE_OK|RIGCTRL_MSGTYPE_NOT_OK)) != 0) )
         { fMatch = TRUE;  // "OK" matches "NOT OK", and "NOT OK" also matches "OK" !
           break;
         }
      }

     if( fMatch )
      { RigCtrl_SendAndLogMessage( pRC, iRigCtrlClientPort, pbMessage, iMsgLength,
           FALSE/*don't wait for completion*/, iReceivedMessageType, pszComment );
      }
   } // end for < iRigCtrlClientPort >
} // end RigCtrl_CheckAndForwardResponseToClients()


//--------------------------------------------------------------------------
BOOL RigCtrl_AskForRadioID( T_RigCtrlInstance *pRC )
{
  BOOL fResult = FALSE;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                       RIGCTRL_MSGTYPE_RADIO_ID | RIGCTRL_MSGTYPE_FLAG_SUGGEST_NEW_HEADLINE_FOR_TRAFFIC_LOG, // iMsgType for the "traffic log" ...
                       "get radio ID",     // comment for the "traffic log"
                       0x19,  // iCmd: CI-V command "Read transceiver ID" (and a few others)
                       0x00,  // iSubCmd: "Read transceiver ID code"
                       -1, -1 );  // additional data bytes : None !
         // > Returns always the default CI-V adress, regardless of the actual
         // > setting of the CI-V adress. This function can be used by software
         // > to determine which radio is connected to the computer.
         // > Despite the functions name this command works also on receivers,
         // > not only transceivers.
         // If all works as planned, the transceiver also responds with cmd 0x1900 :
         pRC->iExpectedResponseOnRadioPort = 0x1900;  // only one known subcode, but.. !
         // An IC-7300 required almost a SECOND to send an answer ?!
         pRC->iResponseCountdown_ms =  // if radio doesn't respond after this time, continue..
          pRC->iMaxResponseDelay_ms;
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )


  return fResult;

} // end RigCtrl_AskForRadioID()

//--------------------------------------------------------------------------
BOOL RigCtrl_AskForVFOFrequency( T_RigCtrlInstance *pRC )
{
  BOOL fResult = FALSE;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                       RIGCTRL_MSGTYPE_FREQUENCY, // iMsgType (for traffic log)
                       "read VFO freq", // short comment for the "traffic log"
                       0x03,  // iCmd: CI-V command "Read operating frequency"
                       -1, -1, -1 );  // iSubCmd and iDataBytes : Omit !
         // If all works as planned, the transceiver responds with 0x03 :
         pRC->iExpectedResponseOnRadioPort = 0x03; // CI-V response "Read operating frequency"
         // [IC7610CIV] doesn't explain the exact format of this (and many other)
         // 'simple' queries. But http://www.plicht.de/ekki/civ/civ-p41.html says:
         // > This command is understood by practically all Icom radios.
         // That's why we use THIS simple query, not one of the fany new
         // multi-byte commands like 0x1A (with its awfully long list of sub-commands).
         pRC->iResponseCountdown_ms =  // if radio doesn't respond after this time, continue..
          pRC->iMaxResponseDelay_ms;
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  return fResult;

} // end RigCtrl_AskForVFOFrequency()

//--------------------------------------------------------------------------
BOOL RigCtrl_AskForOperatingMode( T_RigCtrlInstance *pRC )
{
  BOOL fResult = FALSE;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                     RIGCTRL_MSGTYPE_OP_MODE, // iMsgType (for traffic log)
                       "read op. mode", // short comment for traffic log
                       0x04,  // iCmd: CI-V command "Read operating mode"
                       -1, -1, -1 );  // iSubCmd and iDataBytes : Omit !
         // If all works as planned, the transceiver responds with 0x04 :
         pRC->iExpectedResponseOnRadioPort = 0x04; // CI-V response with "operating mode"
         pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms;
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  return fResult;

} // end RigCtrl_AskForOperatingMode()



//--------------------------------------------------------------------------
BOOL RigCtrl_StartStopSpectrumData( T_RigCtrlInstance *pRC, BOOL fStart )
  // Tries to kick the periodic, unsolicited transmission of an IC-7300's
  // (or IC-7610's, or maybe IC-R8600's) broadband spectrum alive.
  // Icom decided to call this 'Scope waveform data' (command 0x27)
  // but this name is too ambiguous (since a "waveform" sounds like TIME DOMAIN),
  // in fact what we're trying to read here are the data which the LOCAL DISPLAY
  // (TFT screen inside the rig) uses for the spectrum graph + waterfall display.
  //      Actually, this feature was the motivation for the author
  //      to roll his own 'Rig Control' unit for Spectrum Lab,
  //      instead of using some 3rd party library like 'HamLib' or 'OmniRig' .
  // Called from the incredible state machine in RigControl_Handler(),
  //      somewhere in state RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM .
{
  BOOL fResult = FALSE;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                     RIGCTRL_MSGTYPE_C_SPECTRUM, // iMsgType (for traffic log)
                     fStart ? "start spectrum" : "stop spectrum", // comment for traffic log
                       0x27,  // iCmd   : "Scope waveform data"
                       0x11,  // iSubCmd: "Send/read the Scope wave data output setting" [IC7300FM page 19-7]
                      fStart, // 1st data byte : 1 = ON,  0 = off
                       -1 );  // 2nd data byte : omit
         // If all works as guessed, the transceiver agrees as shown below:
         // > 12:01:35.5 TX 008 FE FE 94 E0 27 11 01 FD           (scope)  ; start spectrum
         // > 12:01:35.5 RX 006 FE FE E0 94 FB FD                 (OK)
         // (other transceivers may simply start transmitting SPECTRUM DATA,
         //  don't rely on the "OK"-response for the above command)
         pRC->iExpectedResponseOnRadioPort = 0xFB; // at least an IC-7300 responded with "OK"
         pRC->iResponseCountdown_ms =  // if the radio doesn't respond after this time, continue..
          pRC->iMaxResponseDelay_ms;
         // A few milliseconds later, an IC-7300 began sending "Scope waveform data" periodically,
         //  message format shown in RigCtrl_ParseCIV(), case 0x27 .
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  return fResult;
} // end RigCtrl_StartStopSpectrumData()


//--------------------------------------------------------------------------
BOOL RigCtrl_SetScopeSpeed( T_RigCtrlInstance *pRC, int iScopeSpeed )
{
  BOOL fResult = FALSE;
  BYTE bData;
  char *pszComment;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         switch( iScopeSpeed )
          { case RIGCTRL_SCOPE_SPEED_FAST:
            default:
               bData = 0x00;
               pszComment = "scope:FAST";
               break;
            case RIGCTRL_SCOPE_SPEED_MID:
               bData = 0x01;
               pszComment = "scope:MID";
               break;
            case RIGCTRL_SCOPE_SPEED_SLOW:
               bData = 0x02;
               pszComment = "scope:SLOW";
               break;
          }
         fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                     RIGCTRL_MSGTYPE_C_SPECTRUM, // iMsgType (for traffic log)
                       pszComment, // comment for traffic log
                       0x27,  // iCmd   : "Scope waveform data"
                       0x1A,  // iSubCmd: "Send/read the Scope Sweep speed settings"
                       0x00,  // 1st data byte : 0x00 = MAIN scope, 0x01 = SUB scope
                     bData);  // 2nd data byte : "speed setting"
         pRC->iExpectedResponseOnRadioPort = 0xFB; // at least an IC-7300 responded with "OK"
         pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms;
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  return fResult;
} // end RigCtrl_SetScopeSpeed()

//--------------------------------------------------------------------------
void RigControl_EnterState_Done( T_RigCtrlInstance *pRC )
  // Called from e.g. RigControl_UDPThreadFunc() -> RigControl_Handler(),
  //
{
  char sz255Msg[255];

  pRC->iParameterPollingState = RIGCTRL_POLLSTATE_DONE;

  if( RigControl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_ENABLE )
   { snprintf( sz255Msg, 255, "RigControl: %s, %.1lf kHz, %s",
      RigCtrl_DefaultAddressToString(pRC->iRadioCtrlProtocol,pRC->iDefaultAddress),
      pRC->dblVfoFrequency*1e-3, RigCtrl_OperatingModeToString( pRC->iOpMode ) );
     RigCtrl_AddToTrafficLog( pRC, RIGCTRL_PORT_RADIO, NULL/*pbMessage*/, 0/*iMsgLength*/,
                              RIGCTRL_MSGTYPE_OK, sz255Msg );
     // Emit a single line with 'info' in the error history, with a summary about the remote rig:
     ShowError( ERROR_CLASS_INFO | SHOW_ERROR_TIMESTAMP, sz255Msg );
   }
} // end RigControl_EnterState_Done()


//--------------------------------------------------------------------------
void RigControl_Handler( T_RigCtrlInstance *pRC )
  // Periodically called every few ten milliseconds (roughly C_RIGCTRL_HANDLER_CALLING_INTERVAL_MS).
  //
  // Callers (in Spectrum Lab)
  //  * If a SERIAL PORT is used : IOaccess.cpp  : IO_RadioCommThdFunc() .
  //  * If an INTERNET SOCKET (e.g. UDP) is used : RigControl_UDPThreadFunc().
  //
  // Received data may be passed in ANYTIME between two calls of this handler,
  //               via RigControl_ProcessRxData() [for 'COM' ports, UDP, etc].
  // Potential multithreading issues must be avoided by the CALLER(s).
  // Implemented like a state machine for portability
  //    (with a single-thread microcontroller firmware in mind) .
  //
  // BUGLIST (endless, only a few 'recent cases' shown here):
  //
  //  2023-07-01 : When connecting an IC-705 via USB (and without knowing
  //               which of the TWO freshly appeared 'COM ports' to use),
  //               RigControl_Handler() wasn't called from anywhere,
  //               and all the 'CAT Traffic Log' had to say was :
  //               "RigControl: Re-connect" . Not very user-friendly !
  //
{
  char szComment[ RIGCTRL_MAX_COMMENT_LENGTH+4 ], *pszComment = NULL;
  // ex: T_RigCtrl_IcomUdpPortInstance *pUdpPortInst = NULL;

#if(0)  // removed: variables to read the 'Fixed Edge' frequencies for Icom's waterfall/spectrum display
  int  iRange, iEdge;
  BYTE b1, b2;
#endif

  if( pRC == NULL ) // oops.. instance deleted but some thread is still calling us !
   { return;
   }
  if( pRC->initState != initState_Opened ) // oops.. didn't call RigControl_Init() yet, or called RigControl_Close()
   { return;
   }

  RigControl_ProcessRxData( pRC, RIGCTRL_PORT_RADIO );


  // Arrived here ? The 'connect-to-radio' phase seems to have been completed,
  //   or we are 'connected' via a simple 'COM port' (caller=IO_RadioCommThdFunc),
  //         so the normal 'CI-V parameter-polling' state machine jumps in
  //         to exchange whatever we need to with the remote radio.
  //   Below is the ALMOST UNMODIFIED code from the days when there were only
  //   'serial interfaces' (CI-V on a single wire or USB with virtual COM port).

  if( pRC->iResponseCountdown_ms > 0 ) // "do not send anything at the moment"..
   { pRC->iResponseCountdown_ms -= C_RIGCTRL_HANDLER_CALLING_INTERVAL_MS; // <- e.g. 50 milliseconds
     if( pRC->iResponseCountdown_ms <= 0 )  // just timed out ...
      { pRC->iResponseCountdown_ms = 0;
        if( (pRC->iExpectedResponseOnRadioPort >= 0 )
         && (pRC->iParameterPollingState != RIGCTRL_POLLSTATE_PASSIVE )
         && (pRC->iParameterPollingState != RIGCTRL_POLLSTATE_DONE ) )
         { sprintf( pRC->sz255LastError, "timed out waiting for Cmd 0x%02X, state=%d",
             (int)pRC->iExpectedResponseOnRadioPort, (int)pRC->iParameterPollingState );
           if( RigControl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_ERROR )
            { if( ! RigControl_fTrafficMonitorPaused )
               { RigCtrl_AddToTrafficLog( pRC, RIGCTRL_PORT_RADIO, NULL/*pbMessage*/, 0/*iMsgLength*/,
                   RIGCTRL_MSGTYPE_ERROR, "Traffic Log paused after RESPONSE TIMEOUT (as configured)" );
               }
              RigControl_fTrafficMonitorPaused = TRUE;
            } // end if < RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_ERROR .. >
         }   // end if < EXPECTED a response, AND one of the 'active' parameter-polling-states >
      }     // end if < iResponseCountdown_ms just timed out >
   }
  else  // enough time has passed to send the next command..
   { switch( pRC->iParameterPollingState )
      { case RIGCTRL_POLLSTATE_START_RD:  // RigControl_StartReading() has been called..
           // If the protocol is CI-V, start by reading the TRANSCEIVER ID.
           // We'll know if a radio supports a certain 'advanced' feature then,
           // much better than relying on the radio's "default ID" !
           RigCtrl_AskForRadioID( pRC );
           pRC->iParameterPollingState = RIGCTRL_POLLSTATE_FREQUENCY;
           break;
        case RIGCTRL_POLLSTATE_FREQUENCY: // (try to) read the current VFO frequency
           RigCtrl_AskForVFOFrequency( pRC );
           pRC->iParameterPollingState = RIGCTRL_POLLSTATE_OPMODE;
           break;
        case RIGCTRL_POLLSTATE_OPMODE  :  // (try to) read the operating mode (USB,LSB,CW,..)
           RigCtrl_AskForOperatingMode( pRC );
           // ex: if( pRC->iCapabilities  & RIGCTRL_CAPS_SPECTRUM_VIA_CAT )
           // Forget it. Don't rely on hard-coded (built-in) table with
           //            rig-capabilities. It's THE USER'S choice to try, so:
           if( pRC->iRigControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA )
            {  pRC->iParameterPollingState = RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM;
               pRC->iParameterPollingSubState = 0;
            }
           else // THE USER doesn't WANT spectrum data via CI-V .
            {   // If our 'database' says the rig supports spectrum data,
                // let the rig know we don't want those data right now:
              if( pRC->iCapabilities & RIGCTRL_CAPS_SPECTRUM_VIA_CAT )
               { pRC->iParameterPollingState = RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM; // .. to turn spectrum data OFF
                 pRC->iParameterPollingSubState = 0;
               }
              else // the radio most likely will not "understand" CI-V 0x27, so:
               { // ex: pRC->iParameterPollingState = RIGCTRL_POLLSTATE_DONE;
                 RigControl_EnterState_Done( pRC );
               }
            }
           break;
        case RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM: // (try to) activate transmission of "spectrum scope waveform data"
           if( pRC->iRigControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA )
            { if( pRC->iRadioCtrlProtocol == RIGCTRL_PROTOCOL_ICOM_CI_V ) // only for "CI-V"..
               {
                 switch( pRC->iParameterPollingSubState++ )
                  { case 0  : // find out if the spectrum scope is in CENTER- or FIXED mode:
                       RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                            RIGCTRL_MSGTYPE_C_SPECTRUM,
                            "center/fixed ?", // comment for traffic log
                            0x27,  // iCmd   : "Scope waveform data"
                            0x14,  // iSubCmd: "Send/read the Scope Center mode or Fixed mode setting"
                            0x00,  // 1st data byte : 0x00 "Fixed" (what a misnomer.. not the FIXED MODE)
                            -1 );  // 2nd data byte : omit for the REQUEST (only used in the RESPONSE)
                       pRC->iExpectedResponseOnRadioPort = 0x2714; // at least an IC-7300 responded with 0x27 0x14 ..
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms;
                       break;
                    case 1  : // find out if the spectrum scope's "span" setting
                       RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                            RIGCTRL_MSGTYPE_C_SPECTRUM,
                            "scope span ?", // comment for traffic log
                            0x27,  // iCmd   : "Scope waveform data"
                            0x15,  // iSubCmd: "read the span setting in the Center mode Scope"
                            0x00,  // 1st data byte : 0x00 "Fixed" (what a misnomer.. not the FIXED MODE)
                            -1 );  // 2nd data byte : omit for the REQUEST (only used in the RESPONSE)
                       pRC->iExpectedResponseOnRadioPort = 0x2715; // at least an IC-7300 responded with 0x27 0x15 ..
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms;
                       break;
                    case 2  : // find out the "Edge number setting in the Fixed mode Scope"
                       RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                            RIGCTRL_MSGTYPE_C_SPECTRUM,
                            "edge nr ?", // comment for traffic log
                            0x27,  // iCmd   : "Scope waveform data"
                            0x16,  // iSubCmd: "Edge number setting[s] in the Fixed mode Scope"
                            0x00,  // 1st data byte : 0x00 "Fixed" (what a misnomer.. not the FIXED MODE)
                            -1 );  // 2nd data byte : omit for the REQUEST (only used in the RESPONSE)
                       pRC->iExpectedResponseOnRadioPort = 0x2716;
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms;
                       break;
                    // don't need the "Scope hold function status" yet (0x27 0x17) !
                    case 3  : // find out the "Send/read the Scope Attenuator setting" (IC-7851, n/a on IC-7300)
                       RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                            RIGCTRL_MSGTYPE_C_SPECTRUM,
                            "scope atten. ?", // comment for traffic log
                            0x27,  // iCmd   : "Scope waveform data"
                            0x18,  // iSubCmd: "Scope Attenuator setting"
                            0x00,  // 1st data byte : 0 = main scope, 1 = sub scope
                            -1 );  // 2nd data byte : omit for the REQUEST (only used in the RESPONSE)
                       pRC->iExpectedResponseOnRadioPort = 0x2718;
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms;
                       // An IC-7300 (which doesn't have an extra attenuator for the SCOPE)
                       // responded with "Not Good" (NG, cmd 0xFA) .
                       break;
                    case 4  : // find out the "Scope Reference level setting"
                       RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                            RIGCTRL_MSGTYPE_C_SPECTRUM,
                            "ref lvl ?", // comment for traffic log
                            0x27,  // iCmd   : "Scope waveform data"
                            0x19,  // iSubCmd: "Scope Reference level settings"
                            0x00,  // 1st data byte : 0 = main scope, 1 = sub scope
                            -1 );  // 2nd data byte : omit for the REQUEST (only used in the RESPONSE)
                       pRC->iExpectedResponseOnRadioPort = 0x2719;
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms;
                       pRC->iParameterPollingArrayIndex = 0; // prepare reading "Fixed edge frequency settings" ...
                       pRC->iScopeNumFreqRanges        = RIGCTRL_MAX_FIXED_EDGE_SCOPE_FREQ_RANGES;
                       pRC->iScopeNumEdgesPerFreqRange = RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE;
                       break;
#                 if(0)
                    case 5 :  // read all available "Fixed edge frequency settings"
                       if( pRC->iParameterPollingArrayIndex <
                             (  RIGCTRL_MAX_FIXED_EDGE_SCOPE_FREQ_RANGES
                              * RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE ) )
                        { --pRC->iParameterPollingSubState; // undo pRC->iParameterPollingSubState++
                          iRange = 1 + (pRC->iParameterPollingArrayIndex / RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE );
                          iEdge  = 1 + (pRC->iParameterPollingArrayIndex % RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE );
                          if( (iRange<=pRC->iScopeNumFreqRanges) && (iEdge<=pRC->iScopeNumEdgesPerFreqRange) )
                           { // For example, the IC-R8600 only has ONE "range" and ONE "edge" in that range.
                             // This 'if' will limit the number of "questions" we send to the radio.
                             sprintf(szComment, "edge[%d][%d] ?", iRange, iEdge );
                             RigCtrl_IntToBCD_CIV( iRange, &b1, 2/*digits*/ );
                             RigCtrl_IntToBCD_CIV( iEdge,  &b2, 2/*digits*/ );
                             ++pRC->iParameterPollingArrayIndex; // next "Fixed edge frequency setting"
                             RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                               RIGCTRL_MSGTYPE_C_SPECTRUM,
                               szComment, // comment for traffic log
                               0x27,  // iCmd   : "Scope waveform data"
                               0x1E,  // iSubCmd: "read the Fixed edge frequency settings"
                                 b1,  // 1st data byte : "frequency range" (01..13, guess this is BCD, not hex)
                                 b2); // 2nd data byte : "edge number) (01..03, guess this is also meant as BCD)
                             pRC->iExpectedResponseOnRadioPort = 0x271E;
                             pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms;
                             // When deliberately exceeding the number of "scope frequency ranges"
                             //  ( 13 for an IC-7300 ), the radio responded with "Not Good" :
                             // > 23:23:00.0 TX 009 FE FE 94 E0 27 1E 14 01 FD (scope)    ; edge[14][1] ?
                             // > 23:23:00.1 RX 006 FE FE E0 94 FA FD          (Not OK)
                             // If this happens, RigCtrl_ParseCIV(case 0x271E) limits
                             //    pRC->iScopeNumFreqRanges and pRC->iScopeNumEdgesPerFreqRange
                             //  to what the radio *really* supports.
                             // This seems to be the only method to avoid hard-coding the number,
                             // because other transceivers (like the IC-7851) have LESS frequency ranges,
                             // and future devices (like HF/VHF/UHF/SHF radios) may have MORE .
                           } // end if < valid frequency-RANGE and -EDGE > ?
                          else
                           { // Will waste about 50 ms until the next attempt. So what ?
                           }
                        }
                       break;
#                 else  // skip reading all "Fixed edge frequency settings" :
                    case 5 :
                       break;
#                 endif // (0,1)
                    case 6  : // try to make the scope FASTER VIA SERIAL PORT
                        // (even when "Waterfall" set to FAST, an IC-7300
                        //  only delivered 4 new spectra per second)
                       RigCtrl_SetScopeSpeed( pRC, pRC->iScopeSpeed );
                       break;
                    case 7  : // start or stop periodic transmission of spectrum data (which Icom calls "waveform")
                       RigCtrl_StartStopSpectrumData( pRC, TRUE );
                       break;
                    default :
                       // ex: pRC->iParameterPollingState = RIGCTRL_POLLSTATE_DONE;
                       RigControl_EnterState_Done( pRC );
                       pRC->iParameterPollingSubState = 0;
                       break;
                  } // end switch( pRC->iParameterPollingSubState )
               }
              else // not CI-V ?
               { // reading spectrum data not supported here !
                 // ex: pRC->iParameterPollingState = RIGCTRL_POLLSTATE_DONE;
                 RigControl_EnterState_Done( pRC );
               }
            }
           else  // ! RIGCTRL_FLAG_WANT_SPECTRUM_DATA ...  so turn it off
            { RigCtrl_StartStopSpectrumData( pRC, FALSE );
              // ex: pRC->iParameterPollingState = RIGCTRL_POLLSTATE_DONE;
              RigControl_EnterState_Done( pRC );
             }
           break; // end case RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM
        case RIGCTRL_POLLSTATE_PASSIVE : // nothing else to do.. or not open for business
        case RIGCTRL_POLLSTATE_DONE    : // finished polling "all we need to know" about the radio
           if( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_FREQUENCY) )
            { RigCtrl_SetVFOFrequency_Internal( pRC, pRC->dblPostponedFrequency );
              pRC->dblVfoFrequency = pRC->dblPostponedFrequency;
              pRC->iPostponedSetMessages &= ~(1<<RIGCTRL_MSGTYPE_FREQUENCY); // "done"
            }
           else if( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_C_SPECTRUM) )
            { RigCtrl_StartStopSpectrumData( pRC,
                 (pRC->iRigControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA/*2*/) != 0 );
              pRC->iPostponedSetMessages &= ~(1<<RIGCTRL_MSGTYPE_C_SPECTRUM); // "done"
            }
           else if( pRC->iReadPNFifoHead != pRC->iReadPNFifoTail )  // other 'read'-command pending ?
            { RigCtrl_SendReadCommandFromFIFO( pRC );
            }
           else // nothing else to do ...
           if( pRC->iParameterPollingState == RIGCTRL_POLLSTATE_DONE )
            { pRC->iScopeDataCountdown_ms -= 50; // countdown for 'Scope'-data ...
              if( pRC->iScopeDataCountdown_ms < 0 )
               { // Reception of scope waveform data has timed out -> kick it alive again.
                 // Since we're in state RIGCTRL_POLLSTATE_DONE here, there's
                 // nothing else to do, so :
                 if( pRC->iRigControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA )
                  { // Mr Scope NOT alive but he SHOULD be active ...
                    strcpy( szComment, "Remote spectrum timeout, trying to kick it alive again" );
                    strcpy( pRC->sz255LastError, szComment );
                    RigCtrl_AddToTrafficLog( pRC, RIGCTRL_PORT_RADIO, NULL/*pbMessage*/, 0/*iMsgLength*/, RIGCTRL_MSGTYPE_ERROR, szComment );
                    if( RigControl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_ERROR )
                     { if( ! RigControl_fTrafficMonitorPaused )
                        { RigCtrl_AddToTrafficLog( pRC, RIGCTRL_PORT_RADIO, NULL/*pbMessage*/, 0/*iMsgLength*/,
                             RIGCTRL_MSGTYPE_ERROR, "Traffic Log paused after SPECTRUM TIMEOUT (as configured)" );
                        }
                       RigControl_fTrafficMonitorPaused = TRUE;
                     } // end if < RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_ERROR .. >

                    // The trouble started when WSJT-X switched an IC-7300 to TRANSMIT,
                    // causing it to stop the periodic 'Scope Waveform' transmissions:
                    // >  RX 008 FE FE E0 94 1C 00 01 FD (other) ; TRXStatus:1 [WSJT-X to SpecLab]
                    // >  t2 008 FE FE E0 94 1C 00 01 FD (other) ; TRXStatus:1 [SpecLab to IC7300]
                    //   ... other messages, here irrelevant. A second later:
                    // >  Remote spectrum timeout, trying to kick it alive again .
                    // Of course a hopeless case during TX .. and we don't want
                    //    to reconfigure the IC-7300 to have it sending scope data
                    //    during TRANSMISSION (possible, but meaningless, because
                    //    the waterfall would still stop in the preferred FIXED mode).
                    // Cure: The CI-V parser must keep track of all messages
                    //       that may cause the radio to TRANSMIT .
                    //       While the radio TRANSMITS, the scope-data-timeout
                    //       must be IGNORED (no attempt to "kick it alive" again).

                    pRC->iParameterPollingState = RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM;
                    // Got here, for example, when closing Icom's "RS-BA1"
                    //  because the damned thing sends 27 11 00 (hex) to STOP spectrum data.
                    //  See details in RigCtrl_ParseCIV(), case 0x27 .
                  }
                 pRC->iScopeDataCountdown_ms = RIGCTRL_SPECTRUM_TIMEOUT_MS; // stay calm for a short while..
                 // but longer than the slowest "Sweep speed setting", and
                 // also longer than the time Icom's "RS-BA1" temporarily
                 // turns off the scope while reading dozens of parameters.
               } // end if < reception of "waveform data" timed out >
            } // end else < no "postponed" messages to send >
           break;
      } // end switch( pRC->iParameterPollingState )
   }   // end if < enough time passed to send the next command >

  (void)pszComment; // "assigned a value that was never used" .. oh, shut up (it's better than reading a non-initialized variable)



}     // end RigControl_Handler()


//--------------------------------------------------------------------------
BOOL RigCtrl_SetVFOFrequency_Internal( T_RigCtrlInstance *pRC,
                              double dblFreqHz ) // [in] freq in HERTZ
  // As most functions, returns TRUE when successful, otherwise FALSE.
  // Returns BEFORE the remote rig has actually 'digested' the new setting !
  //
  // Often called "by mistake" from Spectrum Lab, despite the option
  //       "READ initial settings from radio" instead of
  //       "SEND initial settings to radio". Call stack:
  //       IO_RadioCommThdFunc()
  //         -> RigControl_Handler( with iPostponedSetMessages containing RIGCTRL_MSGTYPE_FREQUENCY )
  //          -> RigCtrl_SetVFOFrequency_Internal() .
  //       Real culprit : See RigCtrl_SetVFOFrequency() ...

{
  char sz16FreqAsText[20], szComment[ RIGCTRL_MAX_COMMENT_LENGTH+4 ];
  char *pszLowDigit;
  BYTE b80TxBuffer[84];
  BYTE b;
  int  iTxLength = 0;
  int  i,n;
  int  iRigCtrlPort = RIGCTRL_PORT_RADIO;

  // convert frequency in Hertz into a string: 10 digits without fraction,
  // LEADING ZEROES, ok for up to 9 GHz ...
  sprintf(sz16FreqAsText, "%010.0lf", (double)dblFreqHz );
  pszLowDigit = sz16FreqAsText + (strlen(sz16FreqAsText)-1); // now points to least significant digit

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         b80TxBuffer[iTxLength++] = 0xFE;                      // fixed preamble code for CI-V protocol
         b80TxBuffer[iTxLength++] = 0xFE;                      // fixed preamble code for CI-V protocol
         b80TxBuffer[iTxLength++] = (BYTE)pRC->iRadioDeviceAddr;  // "to" transceiver's address (0x48 = IC706)
         b80TxBuffer[iTxLength++] = (BYTE)pRC->iRadioMasterAddr;  // "from" controller's address (often 0xE0)
         b80TxBuffer[iTxLength++] = 0x05;                      // CI-V command "Write operating frequency data"
         // If the frequency is below 100 MHz, may have to use 4 BCD-bytes only,
         // because some old radios like the IC-735 only support 4 bytes = 8 digits.
         // Modern radios (and VHF/UHF rigs) support 5 BCD data bytes = 10 digits.
         // Furthermore, the modern radios (like IC-706) NEED these long numbers,
         // otherwise the 100-MHz-place remains UNCHANGED !
         n = (  pRC->iDefaultAddress==RIGCTRL_DEF_ADDR_IC_725
             || pRC->iDefaultAddress==RIGCTRL_DEF_ADDR_IC_735
             || pRC->iDefaultAddress==RIGCTRL_DEF_ADDR_IC_751A ) ? 4 : 5;
         for(i=0; (i<n) && (pszLowDigit>=sz16FreqAsText); ++i)
          { b =  (BYTE)( (*pszLowDigit--)-'0'); // lowest digit, beginning at 1 Hz
            b |= (BYTE)(((*pszLowDigit--)-'0')<<4);   // next digit, beginning at 10 Hz
            b80TxBuffer[iTxLength++] = b;        // append next two digits to BCD data area
          }
         b80TxBuffer[iTxLength++] = 0xFD;        // CI-V code for "End Of Message"
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  if( iTxLength>0 )
   {
     sprintf( szComment, "%.1lf Hz", (double)dblFreqHz );
     return RigCtrl_SendAndLogMessage( pRC, iRigCtrlPort, b80TxBuffer, iTxLength,
              FALSE/*don't wait for completion*/,
              RIGCTRL_MSGTYPE_FREQUENCY | RIGCTRL_MSGTYPE_FLAG_TX, szComment );
   } // end if < something 'sendable' > ?

  return FALSE;
} // end RigCtrl_SetVFOFrequency_Internal()


//--------------------------------------------------------------------------
BOOL RigCtrl_SetVFOFrequency( T_RigCtrlInstance *pRC, /* API */
                              double dblFreqHz ) // [in] freq in HERTZ
  // As most functions, returns TRUE when successful, otherwise FALSE.
  // Returns BEFORE the remote rig has actually 'digested' the new setting,
  //         thus TRUE only means "the new setting is on its way" !
  //
  // To retrieve the current 'VFO frequency', simply read pRC->dblVfoFrequency .
  // To convert an 'baseband frequency' (in the radio's demodulated output stream)
  //        into a 'radio frequency', see RigCtrl_BasebandToRadioFrequency().

  //
{
  // This API function may be called by the application at any time, but it may
  // have to be postponed (internally) when still busy from an 'uninterruptable'
  //      sequence of commands in RigControl_Handler().
  // In ANY case, immediately copy the new "wanted" VFO frequency :
  pRC->dblVfoFrequency = dblFreqHz;
  if( pRC->iResponseCountdown_ms > 0 ) // "do not send anything at the moment"..
   {  pRC->iPostponedSetMessages |= (1<<RIGCTRL_MSGTYPE_FREQUENCY);
      pRC->dblPostponedFrequency = dblFreqHz;
      return TRUE;  // cannot send AT THE MOMENT, but data are (almost) "on the way" !
   }
  else // no need to postpone the "Set"-command :
   {  pRC->iPostponedSetMessages &= ~(1<<RIGCTRL_MSGTYPE_FREQUENCY);
      return RigCtrl_SetVFOFrequency_Internal( pRC, dblFreqHz );
      // Note: Even though an IC-7300 sent the "OK" (acknowledge) much later,
      //       the VFO frequency could be modified many times per second.
      //       Thus do NOT reload pRC->iResponseCountdown_ms
      //       in this case, even though we expect a response for the CI-V
      //       command that has just been sent. From SL's "CAT traffic" log:
      // > 18:08:39.5: TX 007 FE FE 94 E0 19 00 FD               (radio ID)    ; get transceiver ID
      // > 18:08:40.4: RX 008 FE FE E0 94 19 00 94 FD            (radio ID)    ; rig=IC-7300
      // > 18:08:40.4: TX 006 FE FE 94 E0 03 FD                  (VFO freq)    ; read VFO frequency
      // > 18:08:41.4: RX 00B FE FE E0 94 03 50 00 55 03 00 FD   (VFO freq)    ; 3550050.0 Hz
      // > 18:08:41.5: TX 006 FE FE 94 E0 04 FD                  (op mode)     ; read operating mode
      // > 18:08:42.4: RX 008 FE FE E0 94 04 03 03 FD            (op mode)     ; CW, filter=3
      // > 18:08:59.5: TX 00B FE FE 94 E0 05 60 00 55 03 00 FD   (VFO freq)    ; 3550060.0 Hz
      // > 18:09:00.3: RX 006 FE FE E0 94 FB FD                  (OK)
      // > 18:09:01.7: TX 00B FE FE 94 E0 05 70 00 55 03 00 FD   (VFO freq)    ; 3550070.0 Hz (step #1)
      // > 18:09:01.8: TX 00B FE FE 94 E0 05 80 00 55 03 00 FD   (VFO freq)    ; 3550080.0 Hz (step #2)
      // > 18:09:01.9: TX 00B FE FE 94 E0 05 90 00 55 03 00 FD   (VFO freq)    ; 3550090.0 Hz (step #3)
      // > 18:09:02.4: RX 006 FE FE E0 94 FB FD                  (OK)          (ack for step #1)
      // > 18:09:02.4: RX 006 FE FE E0 94 FB FD                  (OK)          (ack for step #2)
      // > 18:09:02.4: RX 006 FE FE E0 94 FB FD                  (OK)          (ack for step #3)
      // -> At least modern Icom radios seem to have large buffers
      //    for RECEPTION and TRANSMISSION. So for these *simple* commands,
      //     it's unnecessary to wait for a response (which would have wasted
      //     2.4 minus 1.7 = 0.7 seconds in the example shown above) .
   }

} // end RigCtrl_SetVFOFrequency()

//--------------------------------------------------------------------------
BOOL RigCtrl_SendCmdToReadUnifiedPN( T_RigCtrlInstance *pRC, int iUnifiedPN )
  // As the name implies, we don't wait for the "arrival" of the response !
  // The return value only indicates if we could *SEND* the command .
  // The read value will LATER be stored in the equivalent member variable,
  // e.g. iUnifiedPN=RIGCTRL_PN_S_METER_LEVEL_DB  ->  pRC->iSMeterLevel_dB .
{
  // Like many other API functions, the actual transmission may have to be
  // postponed (internally) when still busy from an 'uninterruptable'
  //      sequence of commands in RigControl_Handler().
  // Since there is a FIFO for this anyway, use it :
  int iNewFifoHead = (pRC->iReadPNFifoHead + 1) % RIGCTRL_READ_PN_FIFO_SIZE;
  if( iNewFifoHead != pRC->iReadPNFifoTail )  // FIFO not completely full ?
   { pRC->i5ReadPNFifo[pRC->iReadPNFifoHead] = iUnifiedPN;
     pRC->iReadPNFifoHead = iNewFifoHead;
     return TRUE;  // command to "read something" queued up successfully
     // The rest happens in RigControl_Handler()
     //    -> RigCtrl_SendReadCommandFromFIFO() and RigCtrl_ParseCIV() .
   }
  else
   { return FALSE; // out of FIFO space to queue up another 'read'-command
     // (got here quite often, because RigControl_Handler() was often busy
     //  from receiving spectra from an IC-7300 / IC-9700, and barely had
     //  the chance to ask for new S-meter readings or similar.)
   }
} // end RigCtrl_SendCmdToReadUnifiedPN()

//--------------------------------------------------------------------------
void RigCtrl_SendReadCommandFromFIFO( T_RigCtrlInstance *pRC )
  // [in] pRC->i5ReadPNFifo[pRC->iReadPNFifoTail++]
  //
  // Only called from RigControl_Handler() when another parameter needs to be
  // polled (e.g. RIGCTRL_PN_S_METER_LEVEL_DB), and the current state
  // ALLOWS sending another frame to the remotely controlled rig.
  //
  // The to-be-read "unified parameter" has been requested by the application
  // via RigCtrl_SendCmdToReadUnifiedPN( T_RigCtrlInstance *pRC, int iUnifiedPN );
{ int iUnifiedPN = pRC->i5ReadPNFifo[pRC->iReadPNFifoTail];
  int iCmd=-1, iSubCmd =-1;
  pRC->iReadPNFifoTail = (pRC->iReadPNFifoTail+1) % RIGCTRL_READ_PN_FIFO_SIZE;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         // What's the 'command' and 'subcommand' in CI-V to *READ* this parameter ?
         switch( iUnifiedPN )
          { // Various "meter" readings via CI-V,  command 0x15 --->
            // [IC7300FM] page 19-3 : cmd 0x15 subcode 2 = Read S-meter level, etc
            case RIGCTRL_PN_SQUELCH_STATUS     : iCmd=0x15; iSubCmd=0x01; break;
            case RIGCTRL_PN_S_METER_LEVEL_DB   : iCmd=0x15; iSubCmd=0x02; break;
            case RIGCTRL_PN_POWER_METER_PERCENT: iCmd=0x15; iSubCmd=0x11; break;
            case RIGCTRL_PN_SWR_METER_LEVEL    : iCmd=0x15; iSubCmd=0x12; break;
            case RIGCTRL_PN_ALC_METER_LEVEL    : iCmd=0x15; iSubCmd=0x13; break;
            case RIGCTRL_PN_COMP_METER_LEVEL_DB: iCmd=0x15; iSubCmd=0x14; break;
            case RIGCTRL_PN_SUPPLY_VOLTATE_mV  : iCmd=0x15; iSubCmd=0x15; break;
            case RIGCTRL_PN_DRAIN_CURRENT_mA   : iCmd=0x15; iSubCmd=0x16; break;
            default:  // anything else isn't supported here yet ...
               break; // leave iCmd negative to ignore the read-request
          } // end switch( iUnifiedPN )
         if( iCmd>=0 )
          { if( RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                     RIGCTRL_MSGTYPE_OTHER, // message type (category) for the traffic log
                     "read param", // short comment for traffic log
                     iCmd, iSubCmd, -1, -1 ) )
             { // If all works as planned, the transceiver responds with the same 'cmd' byte:
               pRC->iExpectedResponseOnRadioPort = iCmd; // CI-V response with e.g. 0x15
               pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms;
               // In SL's CAT traffic monitor, the full 'data exchange' looked like this ...
               // > 20:57:48.4 TX 007 FE FE 94 E0 15 02 FD        (other) ; read param
               // > 20:57:48.5 RX 007 FE FE 94 E0 15 02 FD        (echo)
               // > 20:57:48.5 RX 009 FE FE E0 94 15 02 01 23 FD  (other) ; SMeterLevel:123
               // ... and the data arrived a bit later in RigCtrl_ParseCIV(),
               //     which scaled the result ('123' in BCD) into 'dB over S0'
               //     and updated pRC->iSMeterLevel_dB, from where Spectrum Lab
               //     polled it, and distributed it to OpenWebRX-alike clients.
             }
            else // failed to send the "read"-command...
             { // better luck next time !
             }
          }
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

} // end RigCtrl_SendReadCommandFromFIFO()


//--------------------------------------------------------------------------
void RigCtrl_UpdateBasebandFrequencyConversionParams( T_RigCtrlInstance *pRC )  // -> DemodBasebandFreqOffset, ..Factor
  // Called, for example, from Spectrum Lab when drawing the filter passband
  // into the 'Remote Spectrum Display', received from IC-7300 and similar rigs.
  //    This function must know all the details about the relation between
  //    VFO frequency (aka 'dial frequency' seen on the radio's display)
  //    and the audio frequency.
  //  For USB, this is simple : Audio-frequency = Radio-Frequency - VFO-frequency;
  //                            thus [out] pRC->dblDemodBasebandFreqOffset = 0.0
  //                             and [out] pRC->dblDemodBasebandFreqFactor = +1.0
  //  For CW, it's not,
  //    because (example) : (a) Audio-frequency = Radio-Frequency - VFO-frequency + CW-pitch
  //                   or   (b) Audio-frequency = VFO-frequency - Radio-Frequency + CW-pitch
  //    depending on the radio receiving CW in "narrow USB" or "narrow LSB".
  //
  //
{
  double dblCwPitch_Hz;
  if( (pRC->iCWPitch_Hz != 0) && (pRC->iCWPitch_Hz != RIGCTL_NOVALUE_INT ) )
   { dblCwPitch_Hz = pRC->iCWPitch_Hz;
   }
  else // CW pitch (audio center frequency + CW sidetone) hasn't been read yet
   { dblCwPitch_Hz = 650.0; // author's favourite (when adjustable..)
   }

  switch( pRC->iOpMode & RIGCTRL_OPMODE_MASK_TO_STRIP_FLAGS )
   {
     case RIGCTRL_OPMODE_CW : // here, the "demodulator baseband frequency offset"
        if( pRC->iOpMode & RIGCTRL_OPMODE_REVERSE )
         { // In most rigs, "CW-REVERSE" is actually USB (no frequency inversion)
           // With the VFO tuned tuned to 3560 kHz, a signal on 3560.1 Hz
           // will be 100 Hz *above* the configurable 'CW pitch' frequency, thus:
           pRC->dblDemodBasebandFreqOffset = dblCwPitch_Hz;
           pRC->dblDemodBasebandFreqFactor = 1.0;
         }
        else // In most rigs, "CW" is actually LSB (with frequency inversion)
         { pRC->dblDemodBasebandFreqOffset = dblCwPitch_Hz;
           pRC->dblDemodBasebandFreqFactor = -1.0;
         }
        break; // end case RIGCTRL_OPMODE_CW
     default:  // for any other mode, assume "0 Hz baseband" == "VFO frequency",
               // without sideband inversion :
        pRC->dblDemodBasebandFreqOffset = 0.0;
        pRC->dblDemodBasebandFreqFactor = 1.0;
        break;
   } // end switch( pRC->iOpMode & RIGCTRL_OPMODE_MASK_TO_STRIP_FLAGS )

} // end RigCtrl_UpdateBasebandFrequencyConversionParams()


//--------------------------------------------------------------------------
double RigCtrl_BasebandToRadioFrequency( T_RigCtrlInstance *pRC, double dblBasebandFreq)
{
  RigCtrl_UpdateBasebandFrequencyConversionParams( pRC );
  return pRC->dblDemodBasebandFreqOffset // e.g. -650 Hz in "USB CW" with a 650 Hz 'CW pitch' !
       + pRC->dblDemodBasebandFreqFactor * dblBasebandFreq // e.g. +1.0 for "USB", -1.0 for "LSB".
       + pRC->dblVfoFrequency; // plus for example 3.560 MHz "VFO" frequency aka "dialing frequency"

} // end RigCtrl_BasebandToRadioFrequency()


//--------------------------------------------------------------------------
BOOL  RigCtrl_SetOperatingMode_Internal( T_RigCtrlInstance *pRC, int iOpMode )
  // Sends the new operating mode (e.g. RIGCTRL_OPMODE_USB) to the radio.
  // As most functions, returns TRUE when successful, otherwise FALSE.
  // Returns BEFORE the remote rig has actually 'digested' the new setting !
  // (thus returning TRUE only means "successfully SENT" but not necessarily
  //  "new setting was accepted"). Also applies to many other 'set' functions.
{
  int  iMode, iFilter;
  BOOL fResult = FALSE;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
     case RIGCTRL_PROTOCOL_NONE :
     default:
        break;  // unknown protocol, can only send "direct" strings

     case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
        // Use CI-V command 0x06 because we want a response ("OK" or "NoGo"), and ..
        // > This command is understood by practically all Icom radios.
        iFilter = 0x01;  // Icom's filter codes : 1="wide", 2="medium", 3="narrow"
        // From DF4OR [DF4OR_CIV] :
        // > Due to historical reasons the filter settings are somewhat confusing.
        // > The original CI-V protocol initially had only two settings
        // > for filter width: normal and narrow. When newer rigs came out,
        // > supporting three settings (wide, normal, narrow) the codes
        // > for the filter width had to be extended.
        // > This leads to some confusion about which value sets which filter,
        // > further aggravated by erroneous documentation (e.g. IC-R75).
        // > And when looking at the DSP-filter rigs like IC-756Pro ff.,
        // > where the user can define and arrange each of the three filters
        // > individually, the settings of 'wide' or 'narrow' become meaningless.
        // > Here a filter value of 0x01 corresponds to filter 1,
        // > value 0x02 to filter 2 and 0x03 to filter 3,
        // > regardless of whether filter 1 is acutally wider than filter 2 or 3.
        // > Formerly only $01 and $02 were used for wide and narrow,
        // > medium/normal was introduced later and the data rearranged.
        // > So it can happen that 'narrow' on older rigs is programmed using $02,
        // > on other, newer rigs as $03.
        // DL4YHF: In the DEFAULT settings of an IC-7300, "Fil3" is the NARROWEST.
        switch( iOpMode & RIGCTRL_OPMODE_MASK_TO_STRIP_FLAGS )  // translate RigControl's "operating mode" into Icom's
         {
           case RIGCTRL_OPMODE_CW :
               iMode = 0x03;  // CI-V : 3 = "CW". Hopefully on ALL radios.
               if( iOpMode & RIGCTRL_OPMODE_REVERSE )
                { iMode = 0x07; // CI-V : 0x07 = "CW-R". At least on IC-7300.
                }
               if( iOpMode & RIGCTRL_OPMODE_NARROW )
                { iFilter = 0x02;  // some radios: 1="wide", 2="medium", 3="narrow"
                }
               if( iOpMode & RIGCTRL_OPMODE_VERY_NARROW )
                { iFilter = 0x03;  // some radios: 1="wide", 2="medium", 3="narrow"
                }
               break;
           case RIGCTRL_OPMODE_LSB  :
               iMode = 0x00;  // CI-V : 0 = "LSB"
               break;
           case RIGCTRL_OPMODE_USB  :
               iMode = 0x01;  // CI-V : 1 = "USB"
               break;
           case RIGCTRL_OPMODE_AM   :
               iMode = 0x02;  // CI-V : 2 = "AM"
               break;
           case RIGCTRL_OPMODE_FM   :
               iMode = 0x03;  // CI-V : 3 = "FM" (at least on IC706)
               break;
           case RIGCTRL_OPMODE_FM_WIDE :
               iMode = 0x06;  // CI-V : 6 = "FM wide" (at least for IC706)
               break;
           case RIGCTRL_OPMODE_RTTY :
               iMode = 0x04;    // CI-V : 4 = "RTTY". But on IC-910, 9 is FM. Holy shit.
               // (these annoying subtleties are one reason to use HEX VALUES here,
               //  not macro constants which everyone would have to look up permanently..)
               if( iOpMode & RIGCTRL_OPMODE_REVERSE )
                { iMode = 0x08; // CI-V : 0x08 = "RTTY-R". At least on IC-7300..
                }
               break;
           case RIGCTRL_OPMODE_PSK  :
               iMode = 0x12;    // CI-V : 0x12 = "PSK". At least on IC-7300..
               if( iOpMode & RIGCTRL_OPMODE_REVERSE )
                { iMode = 0x13; // CI-V : 0x12 = "PSK-R". At least on IC-7300
                }
               break;
           default:
               iMode = 0x03;  // CI-V : "CW"
               break;
          } // end switch( pRC->iOpMode )

         if( iMode >= 0 )  // successfully translated at least the MODE..
          { fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                    RIGCTRL_MSGTYPE_OP_MODE,  // message type for traffic log
                       "set mode and filter", // comment for traffic log
                       0x06,  // iCmd: CI-V command "Set mode and filter"
                         -1,  // iSubCmd and iDataByte1 : Omit !
                       iMode, // iDataByte1 : here "mode".
                    iFilter); // iDataByte2 : here "filter" (optionally)
            pRC->iExpectedResponseOnRadioPort = 0x06; // CI-V response "Set operating mode"
          }
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  return fResult;

} // end RigCtrl_SetOperatingMode_Internal()

//--------------------------------------------------------------------------
BOOL  RigCtrl_SetOperatingMode( T_RigCtrlInstance *pRC, int iOpMode ) /* API */
  // Sends the new operating mode (e.g. RIGCTRL_OPMODE_USB) to the radio.
  //  [in] iOpMode : RIGCTRL_OPMODE_CW, RIGCTRL_OPMODE_LSB, RIGCTRL_OPMODE_USB,
  //                 RIGCTRL_OPMODE_AM, RIGCTRL_OPMODE_FM,  RIGCTRL_OPMODE_FM_WIDE
  //                   ("basic modes" below RIGCTRL_OPMODE_MASK_TO_STRIP_FLAGS),
  //                    OPTIONALLY bitwise combined with flags like
  //                 RIGCTRL_OPMODE_REVERSE (for "CW Reverse"),
  //                 RIGCTRL_OPMODE_NARROW, RIGCTRL_OPMODE_VERY_NARROW
  //                 (the latter for compatibility with older rigs,
  //                  which don't support reading the "filter bandwidth").
  //
  // As most functions, returns TRUE when successful, otherwise FALSE.
  // Returns BEFORE the remote rig has actually 'digested' the new setting !
  // (thus returning TRUE only means "successfully SENT" but not necessarily
  //  "new setting was accepted"). Also applies to many other 'set' functions.
{
  // This API function may be called by the application at any time.
  // May have to postpone this when still busy from an 'uninterruptable'
  // sequence of commands in RigControl_Handler().
  // But anyway, immediately copy the new frequency :
  pRC->iOpMode = iOpMode;
  if( pRC->iResponseCountdown_ms > 0 ) // "do not send anything at the moment"..
   {  pRC->iPostponedSetMessages |= (1<<RIGCTRL_MSGTYPE_OP_MODE);
      pRC->iPostponedOpMode = iOpMode;
      return FALSE;  // cannot send AT THE MOMENT !
   }
  else // no need to postpone the "Set"-command :
   {  pRC->iPostponedSetMessages &= ~(1<<RIGCTRL_MSGTYPE_OP_MODE);
      return RigCtrl_SetOperatingMode_Internal( pRC, iOpMode );
   }

} // end RigCtrl_SetOperatingMode()

//--------------------------------------------------------------------------
char* RigCtrl_OperatingModeToString(int iOpMode)
  // First used in Spectrum Lab for interpreter function "rct.mode_string" .
{ static char sz10Result[16];
  char *pszBasicMode;

  // If it's one of the common modes, don't use the evil static buffer:
  switch( iOpMode & 15 )
   {
     case RIGCTRL_OPMODE_CW : pszBasicMode = "CW";
          if( iOpMode & RIGCTRL_OPMODE_NARROW )
           { pszBasicMode = "CWN";
           }
          if( iOpMode & RIGCTRL_OPMODE_VERY_NARROW )
           { pszBasicMode = "CWNN";
           }
          break;
     case RIGCTRL_OPMODE_LSB     : pszBasicMode = "LSB"; break;
     case RIGCTRL_OPMODE_USB     : pszBasicMode = "USB"; break;
     case RIGCTRL_OPMODE_AM      : pszBasicMode = "AM";  break;
     case RIGCTRL_OPMODE_FM      : pszBasicMode = "FM";  break;
     case RIGCTRL_OPMODE_FM_WIDE : pszBasicMode = "FMW"; break;
     case RIGCTRL_OPMODE_RTTY    : pszBasicMode = "RTTY"; break;
     case RIGCTRL_OPMODE_PSK     : pszBasicMode = "PSK"; break;
     default: pszBasicMode = "???"; break;
   } // end switch( pRC->iOpMode )
  if( iOpMode & RIGCTRL_OPMODE_REVERSE )
   { strcpy( sz10Result, pszBasicMode );
     strcat( sz10Result, "R" );
     return  sz10Result;
   }
  else
   { return pszBasicMode;
   }

} // end RigCtrl_OperatingModeToString()

//--------------------------------------------------------------------------
void RigCtrl_IntToBCD_CIV( long i32Value, BYTE *pbDest, int nDigits )
  // More or less inverse to RigCtrl_ParseBCD_CIV_LSDigitFirst(). Details there.
  // Hopefully the number of digits is always EVEN...
{
  int i, iDigit;
  for(i=0; i<nDigits; ++i) // begin with the LEAST SIGNIFICANT digit at i=0
   { iDigit = i32Value % 10;
     i32Value /= 10;
     if( !(i & 1) )       // 1st, 3rd, 5th digit ? calculate LOW nibble
      { *pbDest = (BYTE)iDigit;
      }
     else                 // 2nd, 4th, 6th digit ? emit the next byte
      { *pbDest |= ((BYTE)iDigit << 4);
        ++pbDest;
      }
   }
} // end RigCtrl_IntToBCD_CIV()

//--------------------------------------------------------------------------
long RigCtrl_ParseBCD_CIV_LSDigitFirst( BYTE *pbData, int nDigits )
  // Parses an BCD-coded integer as frequently used in Icom's CI-V protocol.
  // Most may actually be 10-digit fields, but padded with zeros.
  // NOT SUITABLE FOR PARSING THE 4-DIGIT BCD PARAMETER NUMBERS
  //              after command 0x17 0x5 - see RigCtrl_ParseBCD_MSDigitFirst()
  // Example:  cmd 0x28, sub-cmd 0x15,  "Scope span settings" :
  //         pbData[0]  [1]     [2]     [4]     [5]
  //         |<----->|
  //          _______________________________________
  //         |   :   |   :   |   :   |   :   |   :   |
  //         | 0 : 0 | X : X | X : X | 0 : 0 | 0 : 0 |
  //         |___:___|___:___|___:___|___:___|___:___|
  //                  /|\ /|\ /|\ /|\ /|\ /|\ /|\ /|\   .
  //                   |   |   |   |   |   |   |   |
  //         (10   1)  1  100 100  10  10MHz ? 1 GHz ?
  //            Hz    kHz  Hz  kHz kHz     1 MHz ? 100 MHz ?
{
  long sum = 0;
  long factor = 1;
  int  i = 0;
  BYTE b;

  while( (nDigits--) > 0 )
    { b = pbData[i++]; // get next byte (with two digits, BCD)
      sum += factor * (long)(b & 0x0F);
      factor *= 10;
      if( (nDigits--) > 0 )
       {
         sum += factor * (long)((b>>4) & 0x0F);
         factor *= 10;
       }
    }
  return sum;

} // end RigCtrl_ParseBCD_CIV_LSDigitFirst()

//--------------------------------------------------------------------------
long RigCtrl_ParseBCD_CIV_MSDigitFirst( BYTE *pbData, int nDigits )
  // Parses an BCD-coded integer as SOMETIMES used in Icom's CI-V protocol,
  //   for example, the 4-DIGIT BCD PARAMETER NUMBERS after command 0x17 0x05 .
  // Example:  cmd 0x17, sub-cmd 0x05,  parameter "1234"
  //               0x17 0x05 0x12 0x34                                   .
  //                         /|\                                         .
  //                          '---  two MOST SIGNIFICANT DIGITS FIRST !  .
{
  long sum = 0;
  int  i = 0;
  BYTE b;

  while( (nDigits--) > 0 )
    { b = pbData[i++]; // get next byte (with two digits, BCD)
      sum = (10*sum) + (long)((b>>4) & 0x0F);
      if( (nDigits--) > 0 )  // 2nd digit from the same source-byte ?
       { sum = (10*sum) + (long)(b & 0x0F);
       }
    }
  return sum;

} // end RigCtrl_ParseBCD_CIV_MSDigitFirst()


//---------------------------------------------------------------------------
void RigCtrl_ParseSpectrumData_CIV( T_RigCtrlInstance *pRC,
               BYTE *pbPayload, int iPayloadLength,
               char *cpComment ) // [out,optional] comment for the "traffic log"
  // Parses the fragment of a spectrum ("Scope waveform data"),
  //     as sent by IC-7300, IC-7610, IC-7851, IC-9700, IC-705, IC-R8600, etc .
  //  [in] pbPayload[0..iPayloadLength-1] : begins with the
  //                  main scope / sub scope indicator byte,
  //                  and ends at the byte before the CI-V postable.
  //
  //  [out] pRC->SpectrumFifo[pRC->iScopeFifoHeadIndex++]
  //               |_.nBinsUsed, dblBinWidth_Hz, dblFmin_Hz, fltMagnitudes_dB[]
  //
  // NOTE: When the spectrum scope on an IC-7300 showed "Scope Out of Range"
  //       because the VFO frequency was too far away from the fixed edge
  //       frequencies, the radio kept sending unsolicited 0x27 0x00 messages,
  //       but WITHOUT waveform data !
{
  int iMainOrSubScope = pbPayload[0];       // IC-7300 : always 0x00 = MAIN scope
  int iFragment = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbPayload+1, 2/*nDigits*/ );  // IC-7300 : 1..11 (decimal)
  int nFragments= RigCtrl_ParseBCD_CIV_LSDigitFirst( pbPayload+2, 2/*nDigits*/ );  // IC-7300 : 11 fragments
  int nBinsInFragment;
  long i32Fcenter_Hz, i32HalfSpan_Hz, i32RangeStart_Hz, i32RangeEnd_Hz;
  BYTE *pbSrcBins;
  T_RigCtrl_Spectrum  *pSpectrum;


  // [IC7300FM] page 19-12 :
  // > The 1st data sends only the wave information ( (1) .. (6) )
  // >     without the waveform data (7).
  // > The 2nd or later data sends the minimum wave information ( (1) .. (3) )
  //       with waveform data (7).
  //
  //  "Command 27 00 : Outputs the waveform data to the controller" .
  //   Example from SL's 'CAT traffic monitor', from an IC-7300 :
  //
  //                                 Cmd   pbPayload ...
  //                 Len Pre.  Addr  |<->| [0  1  2  3  4  5  6  7  8  9 10 11 12 13 14]
  // >             0x016 FE FE E0 94 27 00 00 01 11 00 56 34 12 70 00 00 25 00 00 00 00 FD ("waveform info" in CENTER mode)
  //                    /|\          |  |  |           |070.123456MHz|<--Span=2500Hz->|  |
  //                     |<------------------------- Message, here: 22 bytes ----------->|
  //   pbMessage --------'           |  |  |<------- Payload, here: 15 bytes -------->|  |
  //                                 |  |  |  |  |  |  |  |  |                           |
  //           CI-V main command ----'  |  |  |  |  |  |  |  '- (7)...                   CI-V
  //           CI-V sub command --------'  |  |  |  |  |  '---- (6)                      Postamble
  // (1) 00=main scope, 01=sub scope ------'  |  |  |                                    |
  // (2) Division number (Current, aka NOW) --'  |  | <-  here: iFragment                |
  // (3)       "Division number (Maximum *) -----'  | <-  here: nFragments               |
  // (4)        "Center or Fixed mode data" --------' <-  here: iFixedMode               |
  //                 Len             |  |  |  |  |  | |<----from---->|<---- to ------>|  |
  // >             0x016 FE FE E0 94 27 00 00 01 11 01 00 00 00 70 00 00 00 50 70 00 00 FD ("waveform info" in FIXED mode)
  //                                                   0070.000000MHz  000070.500000MHz
  //
  // Fragments #2 to #10 (full length on IC-7300) and last fragment (#11, shorter on IC-7300):
  //     "minimum wave information":       |<----->|
  //     "waveform data"           :       |       |<-------------------------------------------------------------------------------------------------------------------------------------------------->|
  //                             pbPayload[0   1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52]
  // >             0x03C FE FE E0 94 27 00 00 02 11 32 34 31 30 2D 0D 10 1D 30 30 1C 1A 1A 1B 31 2C 12 2C 30 1C 11 1D 16 31 34 2D 12 16 11 2B 08 2E 33 2D 1E 12 10 1E 29 07 1D 27 10 1D 1C 29 09 08 19 1C FD (fragments #2..16 of 17)
  // >             0x023 FE FE E0 94 27 00 00 11 11 14 10 0D 08 0C 1A 1E 2F 1B 0A 08 13 15 31 34 1C 1C 0F 07 16 0A 1D 1A 1D 0F FD (LAST fragment, #11 of 11 from IC-7300)                                              |__ CI-V postamble (no payload)
  //                                 |  |  |<------- Payload, last fragment : Only 0x23 - 5 = #30 byte = 30 frequency bins-->|
  //                                 |  |  |  |  |
  //           CI-V main command ----'  |  |  |  |
  //           CI-V sub command --------'  |  |  |
  //                                       |  |  |
  //                                       |  |  |
  // (1) 00=main scope, 01=sub scope ------'  |  |  <- IC-7300: always 0x00 = main scope
  // (2) Division number (Current, aka NOW) --'  |  <-  here: iFragment  = 2..11 (decimal)
  // (3)       "Division number (Maximum *) -----'  <-  here: nFragments = 11 (decimal)
  //     No "Center or Fixed mode data" info in fragments #2..11 (thus "min wave info")
  //
  //  (*) The number of "divisions" (fragments in subsequent messages)
  //      depends on the radio model. This parser shall work on ANY radio.
  // Quoted from IC-7300 "Full Manual" page 19-12:
  //  > When sent through the USB port, the data is
  //  > divided by 11 (*) and sent in sequential order.
  //  > The 1st data sends only the wave information ( (1) ~ (6) )
  //  >                 without the waveform data (7).
  //  > The 2nd or later data sends the minimum wave information
  //  >             ( (1) ~ (3) ) with waveform data (7).
  // (*) "divided by 11" : What they mean is "split into 11 fragments" .
  //     Don't take this number of FRAGMENTS per spectrum for granted:
  //   IC-7300, IC-R8600 : "the data is divided by 11" (over USB).
  //   IC-7610, IC-7851  : "the data is divided by 15" (over USB),
  //         but sometimes "all data is sent together" (over LAN).
  if( iMainOrSubScope != 0 )  // only for the MAIN spectrum scope ...
   { return;  // sub-scope not supported. A donated IC-7851 would change this :)
   }

  // Assemble the received fragment in SpectrumFifo[] ..
  pRC->iScopeFifoHeadIndex %= RIGCTRL_SPECTRUM_FIFO_SIZE; // safety first...
  pSpectrum = &pRC->SpectrumFifo[ pRC->iScopeFifoHeadIndex ];

  if( iFragment == 1 )  // first fragment (Icom: "1st data") ?
   { pRC->iScopeBinCounter = 0; // index into pRC->Spectrum[].
     pSpectrum->dblUnixTime= UTL_GetCurrentUnixDateAndTime(); // capture timestamp a.s.a.p.
     if( pbPayload[3] == 0 ) // 0 = center mode, 1=fixed mode
      { // (this info is only contained in fragment #1)
        pSpectrum->fFixedEdgeMode = FALSE; // from IC-7300/IC-9700 with spectrum scope in CENTER mode
        i32Fcenter_Hz  = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbPayload+4, 10/*nDigits*/ );
        i32HalfSpan_Hz = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbPayload+9, 10/*nDigits*/ );
        pSpectrum->dblCenterFreq_Hz = i32Fcenter_Hz;
        // Unfortunately, before the total number of frequency bins is known,
        // it's impossible to determine pSpectrum->dblBinWidth_Hz. Thus:
        pRC->ScopeFreqRangeFromFirstFragment.dblFmin_Hz = (double)(i32Fcenter_Hz-i32HalfSpan_Hz); // "radio frequency" stored in the 1st bin
        pRC->ScopeFreqRangeFromFirstFragment.dblFmax_Hz = (double)(i32Fcenter_Hz+i32HalfSpan_Hz); // "radio frequency" stored in the last bin
            // ... without knowing how many bins the radio will actually send !
        if( cpComment != NULL )
         { sprintf(cpComment, "%ld+-%ld kHz", (long)i32Fcenter_Hz/1000, (long)i32HalfSpan_Hz/1000 );
         }
      }
     else // not "center" mode but "fixed" mode :
      { pSpectrum->fFixedEdgeMode = TRUE; // from IC-7300/IC-9700 with spectrum scope in FIXED EDGE mode
        i32RangeStart_Hz = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbPayload+4, 10/*nDigits*/ );
        i32RangeEnd_Hz   = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbPayload+9, 10/*nDigits*/ );
        pSpectrum->dblCenterFreq_Hz = 0.5 * ( i32RangeStart_Hz + i32RangeEnd_Hz );
        pRC->ScopeFreqRangeFromFirstFragment.dblFmin_Hz = (double)i32RangeStart_Hz;
        pRC->ScopeFreqRangeFromFirstFragment.dblFmax_Hz = (double)i32RangeEnd_Hz;
        if( cpComment != NULL )
         { sprintf(cpComment, "%ld..%ldkHz", (long)i32RangeStart_Hz/1000, (long)i32RangeEnd_Hz/1000 );
         }
      }
     pRC->dblSpectrumCenterFreq_Hz = pSpectrum->dblCenterFreq_Hz;

     // The "first data" (first of N fragments) often contains "wave information",
     //       and usually no "waveform data" (spectrum frequency bins) .
     // BUT : An IC-7610 (which not only has USB, but also a LAN port)
     //       may send EVERYTHING in a single fragment, when using LAN !
     // To support both (USB + LAN), use the PAYLOAD LENGTH to know
     // if (as quoted for the IC-7300 further above) this FIRST FRAGMENT
     // really only contains "wave information" but no "waveform data" :
     nBinsInFragment = iPayloadLength - 15; // -> IC-7300 : iPayloadLength = 15, nBinsInFragment = 0, i.e. NO FREQUENCY BINS in the first fragment.
     pbSrcBins = pbPayload + 15;

   } // end if( iFragment == 1 )
  else if( iFragment>1 && (iFragment<=nFragments) )
   { // the "2nd or later data" (2nd to Nth fragment) sends the minimum
     // wave information, no info about the frequency range :
     nBinsInFragment = iPayloadLength - 3; // IC-7300 : 53 minus 3 bytes "minimum wave information"
         // -> max 50 bins per fragment (except the last, which only contained
         //    28 minus 3 = 25 frequency bins -> total = 9 * 50 + 25 = 475 bins.
     pbSrcBins = pbPayload + 3;
     if( cpComment != NULL )
      { sprintf(cpComment, "fragment %d/%d", iFragment, nFragments );
      }

   } // end if( iFragment>1 && (iFragment<=nFragments) )


  while( ((nBinsInFragment--) > 0)  // anything to append in THIS fragment ?
     &&  (pRC->iScopeBinCounter<RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM) )
   {
#   if( SWI_HARDCORE_DEBUGGING )
     RigCtrl_pbSource = pbSrcBins;
     RigCtrl_iScopeBin= pRC->iScopeBinCounter;
     RigCtrl_pfltDest = pSpectrum->fltMagnitudes_dB;
#   endif
     pSpectrum->fltMagnitudes_dB[ pRC->iScopeBinCounter++ ] = (float)( *pbSrcBins++ );
     // '-- 2024-04-21 : read from address ZERO ? ?  (only after pause/resume via debugger)
   }

  pSpectrum->iAmplitudeUnit = SCALE_UNIT_dB; // pSpectrum->fltMagnitudes_dB[] still uses Icom's "phantasy" unit.
  pSpectrum->fltAmplOffset_dBm = -146.0; // offset to convert fltMagnitudes_dB into dBm, see below
  // The CI-V manual only mentions "data range 0..160" but doesn't explain
  //   the scaling (one step = one decibel ? Relative to WHAT ? Details below.)
  // Design goal : Value RETURNED from here in dBm, comparable with S-meter units.
  // > IARU Region 1 Technical Recommendation R.1 defines S9 for the HF bands
  // >  to be a receiver input power of -73 dBm. This is a level of 50 microvolts
  // >  at the receiver's antenna input assuming the input impedance of the
  // > receiver is 50 ohms.
  // > For VHF bands the recommendation defines S9 to be a receiver input power
  // >  of -93 dBm. This is the equivalent of 5 microvolts in 50 ohms.[6]
  // > The recommendation defines a difference of one S-unit corresponds
  // >   to a difference of 6 decibels (dB), equivalent to a voltage ratio
  // >   of two, or power ratio of four.
  // An IC-7300, fed by a "-20 dBm" signal showed ~~ "S9+43dB" at 3.560 MHz,
  //             (Rigol tracking generator)   and ~~ "S9+22dB" WITH ATTENUATOR.
  //             Scope Bandwidth = 500 kHz .
  // With the generator attenuated until the S-meter JUST showed "S0", "S1", "S2", ..,
  //          the peak bin (in pbSrcBins) delivered  :
  // S-meter   |  "Scope waveform data" (value in pbSrcBins)
  // ----------+---------------------------------------------------
  //   "S0"    :  20 without attenuator,   0 (flat line) with attenuator
  //                 (should be  -127 dBm  =  -22 dBuV in a 50-Ohm system)
  //
  //   "S1"    :  26 without attenuator,   still flat line with attenuator
  //                 (should be  -121 dBm  =  -14 dBuV )
  //     ...
  //   "S5"    :  48 (should be   -97 dBm  =  +10 dBuV )
  //   "S6"    :  54 (should be   -91 dBm  =  +16 dBuV )
  //   "S7"    :  61 (should be   -85 dBm  =  +22 dBuV )
  //   "S8"    :  66 (should be   -79 dBm  =  +28 dBuV )
  //   "S9"    :  73 without attenuator,  35 with attenuator (-> "below S3").
  //                 (should be   -73 dBm  =  +34 dBuV )
  //   "S9+20" :  110 (!?!) without "  ,  73 with attenuator (-> "S9" ).
  // -> Conclusion: pbSrcBins uses neither dBm nor dBuV but some 'phantasy dB",
  //                possibly  "dB relative to the own noise floor".
  //                DL4YHF initially subtracted 73+73 = 146, if (in SpecLab)
  //                the result shall be displayed in dBm ( dB "over" a milliwatt) .
  //                But with a dummyload on the RX antenna input,
  //                the spectrum read '-146 dBm' which is nonsense for an
  //                equivalent receiver bandwidth of 10.5 Hz (IC-7300 spanning
  //                5 kHz of "scope bandwidth" in 475 frequency bins) .
  //     From https://en.wikipedia.org/wiki/Minimum_detectable_signal :
  //        > To calculate the minimum detectable signal we first need to establish
  //        > the noise floor in the receiver by the following equation:
  //        > Noise_floor_dBm = 10*log10( kb * To * 1000 ) + NF + 10*log10(BW),
  //        >  where  kb = Boltzmann constant, 1.38 * 10^-23 Joules per Kelvin
  //        >         To = Temperature of the receiver system in Kelvin, e.g. 290 K
  //        >         NF = Receiver noise figure in dB, let's assume NF = 1.5 [dB]
  //        >         BW = Detection bandwidth in Hz, let's use 84 Hz (Icom spectrum scope).
  //        > Numerical example (modified for the IC-7300's spectrum scope) :
  //        > Noise_floor_dBm = 10*log10( kb * To * 1000 ) + NF + 10*log10(BW)
  //        >                 = 10*log10(1.38E-23 * 290 * 1000) + 1.5 + 10*log10(10.5)
  //        >                 = -162.3 dBm .
  //     Ouch. So we're about 162 - 146 = 16 dB "too high" near the noise floor.
  // The MAXIMUM ("data range 160")
  //                would then be 160 "dB" - 146 dBm = +14 dBm = "overload".
  //     A broadcaster on 7.215 MHz, indicated as circa "S9+15 dB" by the IC-7300
  //       reached '119' (non-scaled peak in pbSrcBins), indicated as
  //       was peaking at (119-146) = -27 dBm in spectrum lab (from 5 kHz wide spectrum scope).
  //       "S9 + 15db" should be -73 dBM + 15 dB = -58 dBm,
  //     so here, we're about 58 - 27 = 31 dB "too low" near the maximum level.
  //     WB, 2020-06-07 : Decided to give up at this point.
  //                      An exact "S9" signal from a generator gave
  //                      "-73 dBm" in SL's own waterfall (with "FFT output unit"
  //                      set to "dBm" in Spectrum Lab) with the preamp OFF,
  //                      so decided to keep pSpectrum->fltAmplOffset_dBm .
  //          Then tried different preamp/attenuator settings in the IC-7300 :
  //             No preamp, no attenuator : "S9" on the S-meter, -73 dBm in SL.
  //             Preamp "1" : "S9 + 5dB" on the s-meter,  -47 dBm in SL (nuts..),
  //             Preamp "2" : "S9 + 10dB" on the s-meter, -39 dBm in SL (nuts..),
  //             ATTENUATOR : "S3" on the Icom's s-meter, -118 dBm in SL (CRAZY!).
  //
  if( (iFragment == nFragments ) && (pRC->iScopeBinCounter>1) ) // all fragments complete -> "emit" the spectrum
   {
     // With an IC-7300, got here with pRC->iScopeBinCounter = 475 .
     pSpectrum->nBinsUsed = pRC->nSpectrumBinsUsed = pRC->iScopeBinCounter;
     pSpectrum->dblFmin_Hz = pRC->ScopeFreqRangeFromFirstFragment.dblFmin_Hz;
     pSpectrum->dblBinWidth_Hz = (pRC->ScopeFreqRangeFromFirstFragment.dblFmax_Hz
                                - pRC->ScopeFreqRangeFromFirstFragment.dblFmin_Hz)
                      / (double)( pRC->iScopeBinCounter - 1 );
     // Allow reading this entry in another thread now :
     pRC->iScopeFifoHeadIndex = (pRC->iScopeFifoHeadIndex + 1) % RIGCTRL_SPECTRUM_FIFO_SIZE;
     if( pRC->iScopeFifoHeadIndex == 0 )
      {  pRC->fScopeFifoFull = TRUE;
      }
     pRC->iScopeBinCounter = 0; // index into pRC->Spectrum[].
     if( pRC->dblUnixTimeOfFirstSpectrum <= 0.0 )
      {  pRC->dblUnixTimeOfFirstSpectrum = pSpectrum->dblUnixTime;
      }
     ++pRC->nCompleteSpectraReceived;  // total spectrum counter, added 2018-12
     // to find out if the spectrum-update-rate "seen" in Spectrum Lab
     // was really so much lower than it appeared to be on the radio.
     // Test result: see DebugU1.cpp : TDebugForm::Btn_GetReportOnCommsClick().
   }
} // end RigCtrl_ParseSpectrumData_CIV()


//--------------------------------------------------------------------------
T_RigCtrl_Spectrum *RigCtrl_AppendSpectrumToFIFO( T_RigCtrlInstance *pRC,
      double dblBinWidth_Hz, // [in] 'width' of a single FFT bin in Hertz (=resolution
      double dblFmin_Hz,     // [in] "radio frequency" stored in the first frequency bin
      int    nFrequencyBins) // [in] number of 'frequency bins' .
  // Added 2024-03 to use the Spectrum FIFO also for the CLIENT SIDE
  //       in the 'Remote CW Keyer', which receives spectra via TCP/IP
  //       (see cbproj/Remote_CW_Keyer/CwNet.c : )
  // Avoids unnecessary copying. No whistles and bells here, no averaging, etc.
  // Note the fixed limit of RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM !
  //
  // Returns NULL if the RigControl instance hasn't been properly initialized.
  //
{
  T_RigCtrl_Spectrum *pSpectrum;

  if( (pRC==NULL) || (nFrequencyBins<1) || (nFrequencyBins>RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM) )
   { return NULL;
   }

  pRC->iScopeFifoHeadIndex %= RIGCTRL_SPECTRUM_FIFO_SIZE; // safety first...
  pSpectrum = &pRC->SpectrumFifo[ pRC->iScopeFifoHeadIndex++ ];
  pRC->iScopeFifoHeadIndex %= RIGCTRL_SPECTRUM_FIFO_SIZE;


  // Fill out the fixed part of the T_RigCtrl_Spectrum, so the spectrum is
  // at least 'valid' when returning, and the caller only needs to populate the
  // bins ( pSpectrum->fltMagnitudes_dB[0..RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM-1] )
  pSpectrum->nBinsUsed      = nFrequencyBins;
  pSpectrum->dblBinWidth_Hz = dblBinWidth_Hz;
  pSpectrum->dblFmin_Hz     = dblFmin_Hz;
  pSpectrum->dblCenterFreq_Hz= dblFmin_Hz + 0.5 * dblBinWidth_Hz * nFrequencyBins;
  pSpectrum->fFixedEdgeMode = FALSE; // <- just a GUESS here.. the caller may overwrite this when known
  pSpectrum->dblUnixTime = UTL_GetCurrentUnixDateAndTime(); // capture timestamp a.s.a.p.
  pSpectrum->iAmplitudeUnit = SCALE_UNIT_dB; // pSpectrum->fltMagnitudes_dB[] still uses Icom's "phantasy" unit.
  pSpectrum->fltAmplOffset_dBm = -146.0; // offset to convert fltMagnitudes_dB into dBm, see below

  ++pRC->nCompleteSpectraReceived;   // total counter, displayed on the 'Debug' tab   

  return pSpectrum;
} // end RigCtrl_AppendSpectrumToFIFO()


//--------------------------------------------------------------------------
T_RigCtrl_Spectrum *RigCtrl_GetSpectrumFromFIFO( T_RigCtrlInstance *pRC, int *piTailIndex )
  // API used by Spectrum Lab to retrieve the address of another spectrum in the FIFO.
  // Avoids unnecessary copying. No whistles and bells here, no averaging, etc.
  //     Any reader (caller) should initialize HIS "tail index" (*piTailIndex)
  //     with *piTailIndex = RIGCTRL_GET_OLDEST_ENTRY to retrieve the oldest available entry in the FIFO,
  //       or *piTailIndex = RIGCTRL_GET_LATEST_ENTRY to retrieve the newest (most recent) entry.
  //
  // Returns NULL if no (new) spectrum is available .
  //
  // In Spectrum Lab, periodically called from void TSpectrumLab::ProcessSpectraFromRemoteRig(),
  //    which may have to 'automatically adjust' the DISPLAYED FREQUENCY RANGE
  //    to the radio-frequency-range contained in the returned T_RigCtrl_Spectrum.
  //    (what makes this incredibly complex is the fact that the Icom radio's
  //     *VFO FREQUENCY* may be outside the *DISPLAYED FREQUENCY RANGE*,
  //     especially with the "Scope" (spectrum display) in "Fix" mode ("fixed edge
  //     without automatic scolling", which has been implemented in more recent
  //     firmware versions).
  //     Solution / details in SpecMain.cpp : TSpectrumLab::ProcessSpectraFromRemoteRig() !
{
  T_RigCtrl_Spectrum *pSpectrum = NULL;

  if( *piTailIndex == RIGCTRL_GET_LATEST_ENTRY )  // since 2020-06, dummy for the MOST RECENT spectrum
   { if( pRC->fScopeFifoFull ) // ALL FIFO entries (except the one at iScopeFifoHeadIndex) are valid...
      { *piTailIndex = pRC->iScopeFifoHeadIndex - 1; // signed index wrap below..
        if( *piTailIndex < 0 )
         {  *piTailIndex = RIGCTRL_SPECTRUM_FIFO_SIZE-1;
         }
      }
     else // index hasn't wrapped around yet, so the oldest available info is at index zero
      { *piTailIndex = pRC->iScopeFifoHeadIndex - 1; // if negative ..
        if( *piTailIndex < 0 ) // .. then there's no valid spectrum at all
         {  *piTailIndex = 0;
         }
      }
   }
  else if( *piTailIndex == RIGCTRL_GET_OLDEST_ENTRY ) // not the MOST RECENT but the OLDEST entry ...
   { if( pRC->fScopeFifoFull ) // ALL FIFO entries (except the one at iScopeFifoHeadIndex) are valid...
      { *piTailIndex = pRC->iScopeFifoHeadIndex + 1; // circular wrap further below..
      }
     else // index hasn't wrapped around yet, so the oldest available entry is at index zero
      { *piTailIndex = 0;
      }
   }
  *piTailIndex %= RIGCTRL_SPECTRUM_FIFO_SIZE;

  if( *piTailIndex != pRC->iScopeFifoHeadIndex ) // something new available for THIS caller !
   { pSpectrum = &pRC->SpectrumFifo[ *piTailIndex ];
     *piTailIndex = (*piTailIndex + 1) % RIGCTRL_SPECTRUM_FIFO_SIZE;
   }

  return pSpectrum;
} // end RigCtrl_GetSpectrumFromFIFO()

//--------------------------------------------------------------------------
CPROT int RigCtrl_GetNumSpectraAvailableInFIFO( T_RigCtrlInstance *pRC, int iTailIndex )
  // Called from e.g. OpenWebRX_Server.c to check IN ADVANCE (before
  //  calling RigCtrl_GetSpectrumFromFIFO() ) HOW MANY spectra (or "waterfall lines")
  //  are 'still waiting to be sent' to a certain client.
{
  int nEntries = 0;
  switch( iTailIndex )
   { case RIGCTRL_GET_LATEST_ENTRY : // dummy for the "MOST RECENT spectrum"
     case RIGCTRL_GET_OLDEST_ENTRY : // dummy for the OLDEST entry ...
          if( pRC->fScopeFifoFull ) // ALL FIFO entries (except the one at iScopeFifoHeadIndex) are valid...
           { nEntries = 1;
           }
          else // "scope FIFO" not full, thus check the *HEAD INDEX* :
           { if( pRC->iScopeFifoHeadIndex > 0 )
              { nEntries = 1;
              }
           }
          break;
     default:  // caller passed in a FIFO TAIL index, so compare against the HEAD index:
          nEntries = (pRC->iScopeFifoHeadIndex - iTailIndex);
          if( nEntries < 0 )   // circular wrap between head and tail ..
           {  nEntries += RIGCTRL_SPECTRUM_FIFO_SIZE;
           }
          break;
   }
  return nEntries;
} // end RigCtrl_GetNumSpectraAvailableInFIFO()


//--------------------------------------------------------------------------
BOOL RigCtrl_GetDisplayableRadioFrequencyRange( T_RigCtrlInstance *pRC,
        double *pdblRadioFreqMin, double *pdblRadioFreqMax, // [out] : RADIO  frequencies [Hz]
        BOOL   *pfFixedEdgeMode )   // [out,optional: TRUE= "fixed edges", FALSE="centered on VFO"
  // Spectrum Lab shall be aware of Icom's "FIXED" / "CENTER" mode setting !
  //    - in FIXED-EDGE mode, the frequency scale (plus spectrum, etc)
  //      must not(!) scroll along with the VFO frequency, because
  //      the VFO frequency may even be OUTSIDE the scope's displayed
  //      frequency range (that's not a bug, but a feature of Icom's
  //      spectrum scope).
  //    - in CENTER mode, the frequency scale (plus spectrum, etc)
  //      must (or at least should) scroll along with the VFO frequency.
  //
  // Return value : TRUE = ok ("we know the frequency range"),
  //                FALSE= no info received from the radio yet, using defaults.
  //
  // That's why RigCtrl_GetDisplayableRadioFrequencyRange() optionally returns
  //      the spectrum scope's current "FIXED" / "CENTER" setting .
{
  T_RigCtrl_Spectrum *pSpectrum;
  int iTailIndex;
  double dblSpan_Hz;

  // When already running, use the most recent spectrum to find the range :
  iTailIndex = RIGCTRL_GET_LATEST_ENTRY;
  pSpectrum  = RigCtrl_GetSpectrumFromFIFO( pRC, &iTailIndex );
  if( pSpectrum != NULL )
   { if( pfFixedEdgeMode != NULL )
      { *pfFixedEdgeMode =  pSpectrum->fFixedEdgeMode; // TRUE=fixed frequencies, FALSE=spectrum centered on VFO
      }
     // Bugfix 2020-06-19 : Span_Hz = (nBinsUse MINUS ONE) * BinWidth_Hz  !
     //        If there are only TWO bins per spectrum, the total span
     //        (from the center of the first  to  the center of the last bin)
     //        is only ONE frequency bin width !
     // ex: dblSpan_Hz = (double)pSpectrum->nBinsUsed * pSpectrum->dblBinWidth_Hz;
     dblSpan_Hz = (double)(pSpectrum->nBinsUsed-1) * pSpectrum->dblBinWidth_Hz;
     //   After this modification, the "sp" edit field in SL's main window
     //   showed the correct span ( 1000.0 kHz, not 1.0021 MHz, for an IC-7300 ).
     if( dblSpan_Hz > 0.0 )
      { *pdblRadioFreqMin = pSpectrum->dblCenterFreq_Hz - 0.5 * dblSpan_Hz;
        *pdblRadioFreqMax = pSpectrum->dblCenterFreq_Hz + 0.5 * dblSpan_Hz;
        return TRUE;
      }
   } // end if( pSpectrum != NULL )

  // Called "early", or spectrum FIFO still empty for any other reason: 2nd guess !
  // If it's an Icom radio, it may have reported its "Span setting in the Center mode Scope"...
  if( pRC->dblScopeSpan_Hz > 0.0 ) // An IC-7300 with a 5 kHz wide display reported "2500 Hz",
   { // .. but that applies to BOTH SIDES of the "center"- (not always the "VFO"-) frequency !
     *pdblRadioFreqMin = pRC->dblVfoFrequency - pRC->dblScopeSpan_Hz;
     *pdblRadioFreqMax = pRC->dblVfoFrequency + pRC->dblScopeSpan_Hz;
     return TRUE;
   }

  // Arrived here ? The spectrum scope's frequency range is completely unknown,
  //  so return the WIDEST possible setting (*) :
  *pdblRadioFreqMin = pRC->dblVfoFrequency-500e3;  // 1 MHz total span from IC-7300
  *pdblRadioFreqMax = pRC->dblVfoFrequency+500e3;
     // (*) Returning FALSE from here makes Spectrum Lab use the last known
     //     frequency range, stored between two sessions in e.g.
     //     UConfig.SpDisplay[0].Chn[0].dblBasebandFreqMin/Max, added to
     //     UConfig.SDR_dblNCOfreq .
  return FALSE;

} // end RigCtrl_GetDisplayableRadioFrequencyRange()


//--------------------------------------------------------------------------
int RigCtrl_SubcodeFrom1A_to_UnifiedParameterNumber( T_RigCtrlInstance *pRC,
           int iSubcode )
{
  switch( pRC->iDefaultAddress )
   { case RIGCTRL_DEF_ADDR_IC_7300 :
        switch( iSubcode ) // IC-7300-specific !!
         {
         //case 0x06 :  // "DATA mode with filter width setting" (IC-7300)
              // There's a hard-coded parser for this in RigCtrl_ParseCIV() !
           case 0x07 :  // "IP+ Setting"       (IC-7300)
              return RIGCTRL_PN_IP_PLUS;  // in many rigs, "IP+ ON" == "Preamp OFF".
                 // But there's a different command / subcode for the PREAMP,
                 // so at least in THIS radio, "IP+" may be something more...
           default:
              break;
         }
        break;
     case RIGCTRL_DEF_ADDR_IC_7610 :
        switch( iSubcode ) // IC-7610-specific !!
         {
           case 0x07 :  // "NTP server access" (IC-7610)
              return RIGCTRL_PN_NTP_SERVER_ACCESS;
           case 0x08 :  // "NTP access result" (IC-7610)
              return RIGCTRL_PN_NTP_ACCESS_RESULT;
           case 0x09 :  // "AF Mute"           (IC-7610)
              return RIGCTRL_PN_AF_MUTE;
           default:
              break;
         }
        break;
     default:
        break;
   }

  // Arrived here ? Good luck digging out the manual, and finding out
  //  the meaning of the subcode directly after command 0x1A yourself !
  return RIGCTRL_PN_UNKNOWN;


} // end  RigCtrl_SubcodeFrom1A_to_UnifiedParameterNumber()


//--------------------------------------------------------------------------
int RigCtrl_SubcodeFrom1A05_to_UnifiedParameterNumber( T_RigCtrlInstance *pRC,
           int iSubcode )
  // Tries to convert A FEW(!) of those rig-specific 4-digit subcodes
  //       after Icom's command 0x1A 0x05 into a UNIFIED, and CONSTANT
  //       parameter number that is usable for a switch-case list in "C".
  //       It's a shame that not even an IC-7300 and IC-7610 are compatible
  //       in this context . They could have simply added NEW "subcodes"
  //       for the IC-7610 (for parameters that don't exist in an IC-7300), but..
  //  [in] pRC->iDefaultAddress : e.g. RIGCTRL_DEF_ADDR_IC_7300, .._7610, etc.
  //       Has been read via CI-V *from the radio itself*.
  //  [in] iSubcode = rig-specific number, parsed from a 4-digit BCD coded
  //       value after CI-V command 0x1A 0x05 .
  //  Return value : RIGCTRL_PN_CIV_USB_ECHO, etc
  //       (only those parameter that appeared important are supported here)
{
  switch( pRC->iDefaultAddress )
   { case RIGCTRL_DEF_ADDR_IC_7300 :
        switch( iSubcode ) // IC-7300-specific !!
         { case    1 : return RIGCTRL_PN_SSB_RX_HPF_LPF;
           case    2 : return RIGCTRL_PN_SSB_RX_BASS_LEVEL;
           case    3 : return RIGCTRL_PN_SSB_RX_TREBLE_LEVEL;
           // etc etc, the above were just A FEW examples.. now for the important ones, especially those used by "RS-BA1":
           case   52 : return RIGCTRL_PN_SSB_CW_SYNC_TUNING; // "SSB/CW synchronous tuning". In IC-7300, PN #0052. Used by RS-BA1.
           case   57 : return RIGCTRL_PN_CALIBRATION_MARKER; // "Calibration marker". In IC-7300, PN #0057. Used by RS-BA1.
           case   60 : return RIGCTRL_PN_AF_OUTPUT_LEVEL_TO_ACC; // .. or "to ACC or USB". In IC-7300, PN #0060
           case   61 : return RIGCTRL_PN_SQUELCH_FOR_AF_ON_ACC;  // "Squelch function for the AF signal output to ACC/USB"
           case   62 : return RIGCTRL_PN_BEEP_AND_SPEECH_ON_ACC; // "Beep and speech output setting to ACC/USB"
           case   63 : return RIGCTRL_PN_IF_OUTPUT_LEVEL_TO_ACC; // "IF signal output level to ACC/USB" (0..255 for 100%)
           case   64 : return RIGCTRL_PN_MOD_INPUT_LEVEL_FROM_ACC; // "MOD input level from ACC" (0..255 for 100%)
           case   65 : return RIGCTRL_PN_MOD_INPUT_LEVEL_FROM_USB; // "MOD input level from USB" (0..255 for 100%)
           case   66 : return RIGCTRL_PN_MOD_IN_CONN_FOR_DATA_OFF; // "MOD input connector during DATA OFF" (0=MIC, 1=ACC, 2=MIC/ACC, 3=USB)
           case   67 : return RIGCTRL_PN_MOD_IN_CONN_FOR_DATA; // "MOD input connector during DATA" (0=MIC, 1=ACC, 2=MIC/ACC, 3=USB)
           case   71 : return RIGCTRL_PN_CIV_TRANSCEIVE;
           case   72 : return RIGCTRL_PN_CIV_ADDR_USB_TO_REMOTE; // data: hex address
           case   73 : return RIGCTRL_PN_CIV_OUTPUT_FOR_ANT;
           case   74 : return RIGCTRL_PN_CIV_UNLINK_FROM_REMOTE; // data: 0=linked, 1=unlinked
           case   75 : return RIGCTRL_PN_CIV_USB_ECHO;
           default   : break;
         }
        break; // end case RIGCTRL_DEF_ADDR_IC_7300

     case RIGCTRL_DEF_ADDR_IC_7610 :
        switch( iSubcode ) // IC-7610-specific !!
         { case    1 : return RIGCTRL_PN_SSB_RX_HPF_LPF;
           case    2 : return RIGCTRL_PN_SSB_RX_BASS_LEVEL;
           case    3 : return RIGCTRL_PN_SSB_RX_TREBLE_LEVEL;
           // etc etc, the above were just A FEW examples.. now for the important ones:
           case  112 : return RIGCTRL_PN_CIV_TRANSCEIVE;
           case  113 : return RIGCTRL_PN_CIV_ADDR_USB_TO_REMOTE; // here: also "LAN to remote"
           case  114 : return RIGCTRL_PN_CIV_OUTPUT_FOR_ANT;
           case  115 : return RIGCTRL_PN_CIV_UNLINK_FROM_REMOTE; // data: 0=linked, 1=unlinked
           case  116 : return RIGCTRL_PN_CIV_USB_ECHO;
           default   : break;
         }
        break; // end case RIGCTRL_DEF_ADDR_IC_7610
   } // end switch( pRC->iDefaultAddress )

  // Arrived here ? Good luck digging out the manual, and finding out
  //  the meaning of the subcode for command 0x1A 0x05 yourself !
  return RIGCTRL_PN_UNKNOWN;
} // end RigCtrl_SubcodeFrom1A05_to_UnifiedParameterNumber()


//--------------------------------------------------------------------------
T_RigCtrl_ParamInfo * RigCtrl_GetInfoForUnifiedParameterNumber(int iUnifiedPN)
{
  T_RigCtrl_ParamInfo *pPI = (T_RigCtrl_ParamInfo*)RigCtrl_ParameterInfo;

  while( pPI->iUnifiedPN != 0 )
   { if( pPI->iUnifiedPN == iUnifiedPN )
      { return pPI;
      }
     ++pPI;
   }

  return NULL;  // no luck .. the caller MUST check for NULL-pointers !
}

//--------------------------------------------------------------------------
char *RigCtrl_UnifiedParameterNumberToString(int iUnifiedPN)
{
  static char sz11Unknown[12];
  T_RigCtrl_ParamInfo *pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
  if( pPI != NULL )
   { return pPI->pszToken;
   }

  // The following isn't thread-safe but shouldn't be necessary at all:
  sprintf( sz11Unknown, "PN[%d]",(int)iUnifiedPN);
  return   sz11Unknown;

} // end RigCtrl_UnifiedParameterNumberToString()


//---------------------------------------------------------------------------
int * RigCtrl_GetIntValuePtrForUnifiedParameterNumber( T_RigCtrlInstance *pRC, int iUnifiedPN )
{
  T_RigCtrl_ParamInfo *pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
  if( pPI != NULL )
   { if( (pPI->iByteOffsetIntoRigCtrlInstance > 0 )
      && (pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { if( pPI->iRigCtrlDataType == RIGCTRL_DT_INT )
         { return (int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance );
         }
      }
   }
  return NULL;  // no luck .. also here, the caller MUST check for NULL-pointers !
} // end RigCtrl_GetIntValuePtrForUnifiedParameterNumber()

//---------------------------------------------------------------------------
double* RigCtrl_GetDblValuePtrForUnifiedParameterNumber( T_RigCtrlInstance *pRC, int iUnifiedPN )
{
  T_RigCtrl_ParamInfo *pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
  if( pPI != NULL )
   { if( (pPI->iByteOffsetIntoRigCtrlInstance > 0 )
      && (pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { if( pPI->iRigCtrlDataType == RIGCTRL_DT_DOUBLE )
         { return (double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance );
         }
      }
   }
  return NULL;  // no luck .. also here, the caller MUST check for NULL-pointers !
} // end RigCtrl_GetDblValuePtrForUnifiedParameterNumber()


//---------------------------------------------------------------------------
void RigCtrl_SetParamValue_Int( T_RigCtrlInstance *pRC, T_RigCtrl_ParamInfo *pPI, long i32Value )
{
  if( (pRC!=NULL) && (pPI!=NULL) )
   { if( (pPI->iByteOffsetIntoRigCtrlInstance > 0 )
      && (pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { switch( pPI->iRigCtrlDataType )
         { case RIGCTRL_DT_INT :
              *(int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = i32Value;
              break;
           case RIGCTRL_DT_DOUBLE :
              *(double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = (double)i32Value;
              break;
           default:
              break;
         }
      }
   }
} // end RigCtrl_SetParamValue_Int()

//---------------------------------------------------------------------------
void RigCtrl_SetParamValue_Double( T_RigCtrlInstance *pRC, T_RigCtrl_ParamInfo *pPI, double dblValue )
{
  if( (pRC!=NULL) && (pPI!=NULL) )
   { if( (pPI->iByteOffsetIntoRigCtrlInstance > 0 )
      && (pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { switch( pPI->iRigCtrlDataType )
         { case RIGCTRL_DT_INT :
              *(int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = (int)dblValue;
              break;
           case RIGCTRL_DT_DOUBLE :
              *(double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = dblValue;
              break;
           default:
              break;
         }
      }
   }
} // end RigCtrl_SetParamValue_Double()

//---------------------------------------------------------------------------
int RigCtrl_EnumerateUnifiedParameters( int iPreviousUnifiedPN )
  // To retrieve a list of ALL "unified parameter numbers" (which are NOT contiguous),
  // begin with iPreviousUnifiedPN = 0.  Each new call will then provide
  // a new unified PN. When all PNs are through, returns iUnifiedPN = 0 again.
{
  T_RigCtrl_ParamInfo *pPI = (T_RigCtrl_ParamInfo*)RigCtrl_ParameterInfo;
  if( iPreviousUnifiedPN <= 0 )
   {  return pPI->iUnifiedPN;
   }
  while( pPI->iUnifiedPN != 0 )
   { if( pPI->iUnifiedPN == iPreviousUnifiedPN )
      { return pPI[1].iUnifiedPN;
      }
     ++pPI;
   }
  // Arrived here ? Looks like all known "unified parameter numbers" are through.
  return 0;
} // end RigCtrl_EnumerateUnifiedParameters()


//--------------------------------------------------------------------------
int RigCtrl_IcomToRigCtrlOpmode( T_RigCtrlInstance *pRC, BYTE bIcomsOpMode )
{ // RigControl-instance data passed in because we may need it some day.
  // Don't expect the numeric value for "PSK" and "RTTY" to be the same in all rigs !
  switch( bIcomsOpMode )
   { case 0 : return RIGCTRL_OPMODE_LSB;
     case 1 : return RIGCTRL_OPMODE_USB;
     case 2 : return RIGCTRL_OPMODE_AM;
     case 3 : return RIGCTRL_OPMODE_CW;
     case 4 : return RIGCTRL_OPMODE_RTTY;
     case 5 : return RIGCTRL_OPMODE_FM;
     case 6 : return RIGCTRL_OPMODE_FM_WIDE; // used by IC-706 and possibly other VHF-capable radios
     case 7 : return RIGCTRL_OPMODE_CW   | RIGCTRL_OPMODE_REVERSE;
     case 8 : return RIGCTRL_OPMODE_RTTY | RIGCTRL_OPMODE_REVERSE;
     case 12: return RIGCTRL_OPMODE_PSK;
     case 13: return RIGCTRL_OPMODE_PSK  | RIGCTRL_OPMODE_REVERSE;
     default: return RIGCTRL_OPMODE_UNKNOWN;
   }
} // end RigCtrl_IcomToRigCtrlOpmode()



//--------------------------------------------------------------------------
int RigCtrl_GetCombinedCmdAndSubcodeForCIV( BYTE *pbMessage, int iMsgLength )
{
  if( iMsgLength <= 5 ) // preamble(2), addr (2), postamble(1) : that's FIVE BYTES, no command in here
   { return -2;         // will never match anything in pRC->iExpectedResponseOnRadioPort
   }
  if( iMsgLength == 6 ) // preamble(2), addr (2), command byte (1), postamble(1) but no SUBCODE..
   { return pbMessage[4];
     // In this case, it's obvious that the command has no sub-command (subcode),
     //     e.g. cmd = 0x05 = "get operating frequency"
   }
  else // iMsgLength > 6 : This is tricky. The byte "after" the command (pbMessage[4])
   { // may be a sub-command (subcode), but it MAY be a variable ranging from 0 to 255 or even more !
     // To find out if it's a TWO-BYTE command (with "Cmd" and "Sub cmd." in IC-7300 FM page 19-3),
     // branch by the first byte. Not bullet-proof if Icom expand CI-V in future.
     switch( pbMessage[4] ) // check for all known TWO-BYTE commands :
      { case 0x07 : // "Select" something (Subcmd:0=VFO A, 1=VFO B, 0xA0=Equalize A+B, 0xB0=Exchange A+B)
        case 0x0E : // Subcmd: 0=scan stop, 1=scan start, 2=prog scan, 3=F scan, .. 0xB1="set as select channel"
        case 0x10 : // Subcmd: 0=tuning step off, 1="send/read the 100 Hz tuning step" (?!?!)
        case 0x13 : // "Speech" with subcmds 0,1,2
        case 0x14 : // read/set various "Levels" (AF, RF, Squelch)
        case 0x15 : // read S-meter, squelch status, power meter, SWR meter, ...
        case 0x16 : // preamp, AGC, noise blanker, noise reduction, ...
        case 0x19 : // so far, only ONE subcode but Icom say it HAS a subcode
        case 0x1A : // "Set/read" a horrible number of differently addressed items (with sub-sub-commands)
        case 0x1B : // repeater tone frequency, TSQL tone frequency, ...
        case 0x1C : // transceiver status, ATU, XFC, ...
        case 0x1E : // number of TX frequency bands, TX frequency band edge frequencies, ...
        case 0x21 : // RIT frequ, RIT enable, XIT
        case 0x25 : // "selected" and "unselected" VFO freq. // main/sub VFO freq. (IC-7610)
        case 0x26 : // "selected" and "unselected" operating mode, filter
        case 0x27 : // scope waveform data, scope settings, level, .....
           return ( (int)pbMessage[4] << 8) | pbMessage[5];
        case 0x06 : // definitely has NO subcode ("Sub cmd") !
        default   : // Guess it's a single-byte command WITH DATA(!)
           return pbMessage[4];
      }
   }
} // end RigCtrl_GetCombinedCmdAndSubcodeForCIV()

//--------------------------------------------------------------------------
int RigCtrl_ParseCIV( T_RigCtrlInstance *pRC,
          int  iRigCtrlPort, // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 / .. ?
          BYTE *pbMessage, int iMsgLength ) // [in] message data, message length in bytes
  // Called upon COMPLETE reception of any CI-V message, via 'COM' (USB) or LAN/WLAN (UDP),
  //    including those messages received on CLIENT PORTS (see "Callers" further below).
  // ALSO CALLED ON RECEPTION OF CI-V MESSAGES ON ONE OF THE "DUPLICATOR"-PORTS !
  //    Most COMMANDS(!) received on duplicator ports will be answered
  //    immediately (if we know the answer, to avoid flooding the radio with
  //    requests from all over the place). Others (especially those we don't
  //    understand) will be passed on from here to the "radio port",
  //    and the response relayed back to the originating port.
  //
  // Returns RIGCTRL_MSGTYPE_xyz, e.g.   RIGCTRL_MSGTYPE_FREQUENCY when successfull,
  //   otherwise RIGCTRL_MSGTYPE_UNKNOWN (i.e. "not recognized") .
  // In Spectrum Lab, anything that arrives HERE MAY be displayed usually
  //    with a CYAN background under 'View/Windows' / 'CAT traffic monitor'.
  //    Certain messages, like error responses, will have a different colour.
  //
  // Callers:
  //   (a) IO_RadioCommThdFunc( for any kind of "COM" port connected to the RADIO)
  //          -> RigControl_ProcessRxData(with iRigCtrlPort==RIGCTRL_PORT_RADIO) -> RigCtrl_ParseCIV()
  //   (b) IO_SerialPortDuplicatorThread( for "external clients" via "COM" port, e.g. via "com0com" )
  //          -> RigControl_ProcessRxData(with iRigCtrlPort==RIGCTRL_PORT_RADIO) -> RigCtrl_ParseCIV()
{
  int i, iSubcode, iUnifiedPN, iNewSetting, iRange, iEdge;
  int iCombinedCmdAndSubcode; // comparable with pRC->iExpectedResponseOnRadioPort
  T_RigCtrl_ParamInfo *pPI;   // "parameter info" from a lookup table
  double dblConvFactor = 1.0;
  int iMsgType = RIGCTRL_MSGTYPE_UNKNOWN; // for SL's CAT traffic monitor
  int iClientAction = RIGCTRL_CLIENT_ACTION_RETRANSMIT; // RIGCTRL_CLIENT_ACTION_NONE / RIGCTRL_CLIENT_ACTION_ENTER_FIFO ?
  BYTE b;
  long i32Value = 0;
  double d, fact;        // for conversion from BCD to floating point
  char  szComment[ RIGCTRL_MAX_COMMENT_LENGTH+4 ]; // for the human-readable "traffic log" ...
  char  sz80Temp[84];
  char  *cpComment = szComment;
  char  *cpToken = "";
  T_RigCtrl_PortInstance *pPort = &pRC->PortInstance[ iRigCtrlPort/*client-port*/ ];
  T_RigCtrl_RadioInfo *pRadioInfo;
  T_RigCtrl_MsgFilter msg_filter;
  BOOL fShowInLog = (RigControl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_ENABLE) != 0;
  BOOL fIsEcho = FALSE;
  BOOL fIsExpectedResponse = FALSE;   // may be difficult to find out !
  BOOL fIsUnsolicitedMsg = FALSE; // initial guess : it's a REPONSE, not an unsolicited REPORT (new VFO frequ, Spectrum scope data)

  szComment[0] = '\0';

  // The TWO FIRST BYTES (0xFE = "Preamble") have already been checked by the
  //  caller (e.g. IO_HandleRcvdDataFromRadio()), so no need to check again.
  // Within RigCtrl_ParseCIV(), don't care for pbMessage[2] = transceiver's address (e.g. 0x48 = IC706)
  // and  for pbMessage[3] = controller's address (often 0xE0), EXCEPT FOR THIS:
  // If the rig is configured for "CI-V Echo" *ON* (alias "CI-V USB Echo Back" in an IC-7300)
  //  anything we send out (as command) returns as an echo, like this:
  //            pbMessage[0 1  2  3  4  5  6  7  8  9  10]
  // > 01:26:22.2 TX 006 FE FE 94 E0 03 FD                 (VFO freq) ; read VFO freq
  // > 01:26:23.2 RX 006 FE FE 94 E0 03 FD                 (VFO freq) ; echo
  // > 01:26:23.2 RX 00B FE FE E0 94 03 00 70 35 05 00 FD  (VFO freq) ; 5357000.0 Hz
  //   ,-------------'-' |___| |  |  |
  //   |          ,--------'   |  |  |
  //   '--length  '--preamble  |  |  |
  //       "to address" -------'  |  '--- "cmd" / iExpectedResponseOnRadioPort
  //       "from address" --------'
  // Unfortunately, "HamLib" (and thus WSJT-X and FTDX) don't work
  // without this bandwidth-wasting stupidity. HERE (in RigControl.c)
  // don't confuse the ECHO with the EXPECTED RESPONSE :
  if( ( pbMessage[2] == pRC->iRadioDeviceAddr )   // message sent TO THE RADIO ?
    &&( pbMessage[3] == pRC->iRadioMasterAddr ) ) // message sent FROM THE MASTER ?
   { // If it's not from one of the 3rd party "clients", this message is an ECHO:
     if( iRigCtrlPort == RIGCTRL_PORT_RADIO )
      { // That's the same as in our "TX"-message (see CAT traffic dump above), so..
        fIsEcho = TRUE;  // .. so it's an ECHO, not the real response message !
      }
     else // guess a 3rd party application (like RS-BA1) is talking to us.
      {   // Example : First message from RS-BA1 = FE FE 94 E0 18 01 FD (7 bytes)
      }
   }

  // Prepare a new filter that will later be passed to RigCtrl_RegisterCIVReponseForClient():
  msg_filter.iCmd = pbMessage[4];
  // per default, any further bytes in any CI-V message are IGNORED for filtering:
  msg_filter.iSubCmd = msg_filter.iDataB1 = msg_filter.iDataB2 = -1;
  // per default, let the filter (for the client) pass reponse "OK" and "NOT OK"
  //   (exceptions will be made in the monster case-list further below) :
  msg_filter.iMsgTypes = RIGCTRL_MSGTYPE_OK | RIGCTRL_MSGTYPE_NOT_OK;

  iCombinedCmdAndSubcode = RigCtrl_GetCombinedCmdAndSubcodeForCIV( pbMessage, iMsgLength );

  if( ( ! fIsEcho ) && (iRigCtrlPort==RIGCTRL_PORT_RADIO) )
   { if ( iCombinedCmdAndSubcode == pRC->iExpectedResponseOnRadioPort )
      { // This MAY BE (!) what RigControl_Handler() has been "waiting for", but:
        //  Don't confuse the ECHO with the EXPECTED RESPONSE :
        if( ! fIsEcho )  // only if this is NOT an "echo" :
         { fIsExpectedResponse = TRUE; // ok for simple cases without the need for a matching SUB-COMMAND (more later)
         }
        // Note: May have to use 'iCombinedCmdAndSubcode' AFTER the monster-case-list
        //       further below, as the NEW "expected response" when a command
        //       retrieved from any CLIENT PORT is forwarded to the radio !
      }
     else // from the RADIO port, but this doesn't LOOK LIKE the "expected" response.
      { // For example, after sending 0x03 ("read operating frequency")
        // an IC-7300 answered with   0xFB ("OK", but without data).
        // It quickly became obvious that a user-friendly "protocol display"
        // was necessary, with all traffic from CONTROLLER to RADIO and back.
        // In Spectrum Lab, that's the "Comms" tab on the "Debug" window.
      }
   }
  if( ! fIsEcho ) // don't waste time trying to parse CI-V "echoes" !
   { switch( pbMessage[4] )  // [in] what Icom calls 'command'. But it may be a response..
      {
        case 0x05: // Icom CI-V code "frequency control" / "Set operating frequency"
        case 0x00: // Transfer operating freq. data (unsolicited message when turning VFO knob)
        case 0x03: // "Read operating frequency" (response on request)
                   // 0x03 is included in WSJT-X's "polling cycle" .
           fIsUnsolicitedMsg = (pbMessage[4] == 0x00);
           iUnifiedPN = RIGCTRL_PN_FREQUENCY;
           pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
           if( pPI != NULL )
            { cpToken = pPI->pszToken;
            }
           else
            { cpToken = "Split";
            }
           i = 5;
           if( iMsgLength >= 8 )  // 2 bytes preamble + 2 byte address + 1 byte command + at least a few digits + 1 byte end-of-message
            { d = 0.0; fact=1.0;
              while(i<(iMsgLength-1))
               { b = pbMessage[i++];
                 d += fact * (double)(b & 0x0F);
                 fact *= 10.0;
                 d += fact * (double)((b>>4) & 0x0F);
                 fact *= 10.0;
               }
              // Info from an IC-905 user (where the REMOTE WATERFALL DISPLAY didn't work as well) :
              // > I noted in the 905 CI-V programming manual the command 0x00 (Frequency)
              // > now is variable length with 2 extra bytes added when on 10GHz
              // > or higher for the 10's and 100's GHz digit places.
              //    -> Compared (a) C:\datasheets\Rig_Manuals\IC-905\IC-905_CI-V_Reference_Guide.pdf (anno 2023)
              //        against (b) C:\datasheets\Rig_Manuals\IC-705\IC-705_ENG_CI-V_V4b_2021_05.pdf ...
              //   (a) page 16 : 6 bytes with 12 BCD-digits, from 1-Hz to 100-GHz digit (the latter always zero)
              //   (b) page 17 : 5 bytes with 10 BCD-digits, from 1-Hz to 1-GHz digit (the latter always zero)
              sprintf( szComment, "%.1lf Hz", (double)d );
              // Since THIS CODE runs in a worker thread,
              // the rest (updating the GUI) happens somewhere else.
              //
              if( ! ( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_FREQUENCY) ) )
               { // Only accept the radio's reported value if we're not going to SEND a NEW frequency soon:
                 if( d != pRC->dblVfoFrequency )
                  { ++pRC->iFrequencyModifiedByRadio_cnt;
                    // This counter is polled in a GUI-thread somewhere.
                  }
                 pRC->dblVfoFrequency = d;
               } // end if < no postponed SET-command for the frequency pending >
              else // the radio has reported its VFO frequency, but a POSTPONED(!)
               {   // request to set a NEW (different) frequency is still pending.
                 // In this case, do NOT modify pRC->dblVfoFrequency now.
                 // This may cause trouble for clients that will read the VFO
                 // frequency through other means (with CI-V, there are many roads leading to Rome).
               }
            }
           iMsgType = RIGCTRL_MSGTYPE_FREQUENCY; // end case 0x00, 0x03, 0x05
           break;

        case 0x01: // "Send mode data" (transceive)
        case 0x04: // "Read operating mode" (response after RigCtrl_AskForOperatingMode(?) )
                   //   (see also: case 0x26, similar purpose, other syntax)
                   // Cmd 0x04 is included in WSJT-X's "polling cycle",
                   //      but WSJT-X doesn't try to CHANGE the mode.
           iMsgType = RIGCTRL_MSGTYPE_OP_MODE;
           iUnifiedPN = RIGCTRL_PN_OP_MODE;
           pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
           if( pPI != NULL )
            { cpToken = pPI->pszToken;
            }
           else
            { cpToken = "OpMode";
            }
           iNewSetting = pRC->iOpMode;
           // IC-7300, IC-7610 : "Operating Mode" with two data bytes.. IC-7300 used iMsgLength=8
           if( iMsgLength >= 7 )  // 2 bytes preamble + 2 byte address + 1 byte command + at least one 'mode' byte + 1 byte end-of-message
            { iNewSetting = RigCtrl_IcomToRigCtrlOpmode( pRC, pbMessage[5] );
              strcpy( cpComment, RigCtrl_OperatingModeToString(iNewSetting) );
              cpComment += strlen( cpComment );
            }
           if( iMsgLength >= 8 )  // "Operating mode" byte followed by optional "Filter setting" ?
            { switch( pbMessage[6] )
               { case 1 : break;  // "FIL1" (normal filter bandwidth)
                 case 2 : iNewSetting |= RIGCTRL_OPMODE_NARROW;      break; // "FIL2" [IC7610CIV page 10]
                 case 3 : iNewSetting |= RIGCTRL_OPMODE_VERY_NARROW; break; // "FIL3"
                 default: break;
               }
              sprintf( cpComment, ", filter=%d", pbMessage[6] );
              // cpComment += strlen( cpComment );
            } // end if < radio also reported the current FILTER SETTING >
           // Only accept the radio's reported value if we're not going to SEND a NEW operating mode:
           if( ! ( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_OP_MODE) ) )
            {
              if( iNewSetting != pRC->iOpMode )
               { pRC->iOpMode = iNewSetting;
               }
            }
           break; // end case 0x01 = "Send mode data", 0x04 = "Read operating mode"

        case 0x02: // "Read band edge frequencies"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x02 = "Read band edge frequencies"
           break;

        case 0x06: // "Set operating mode"  (possibly "transceived", e.g. relayed)
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x06 = "Set operating mode"
           break;

        case 0x07: // "Select the VFO mode". Seen when "RS-BA1 took over control:
                   // pbMessage[0  1  2  3  4  5  6]
                   // > r1 007 FE FE 94 E0 07 00 FD
                   //                         |____ 00 = VFO A, 01 = VFO B, A0 = Equalize A and B, ...
           if( iMsgLength >= 7 )  // "Select VFO mode" followed by at least one DATA byte ?
            { sprintf( cpComment, "SelVFO:%02X", (int)pbMessage[5] ); // for the log..
            }
           else
            { strcpy( cpComment, "SelVFO ?" );
            }
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           break; // end case 0x07 = "Select VFO mode"

        case 0x08: // "Select the Memory mode"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x08 = "Select memory mode"
           break;

        case 0x09: // "Memory write"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x09 = "Memory write"
           break;

        case 0x0A: // "Memory copy to VFO"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x0A = "Memory copy to VFO"
           break;

        case 0x0B: // "Memory clear"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x0B = "Memory clear"
           break;

        case 0x0E: // "Scan"-related commands. Used quite early when launching "RS-BA1":
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           //      pbMessage[0  1  2  3  4  5  6]
           //      > r1 007 FE FE 94 E0 0E 00 FD (with iMsgLength=7)
           //           len                |
           //                              |_____ sub command (no data)
           //      pbMessage[0  1  2  3  4  5  6  7]
           //      > r1 008 FE FE 94 E0 0E 00 03 FD (with iMsgLength=8)
           //           len             |  |  |__ optional data, if iMsgLength > 7
           //                          cmd sub
           if( iMsgLength >= 7 ) // valid length (with or without data) ?
            { iSubcode = pbMessage[5]; //
              switch( iSubcode )  // "sub-command" after 0x0E ?
               { case 0x00 : cpToken = "Scan stop";       // all for the log..
                   iUnifiedPN = RIGCTRL_PN_SCAN_MODE;
                    break;
                 case 0x01 : cpToken = "P/M scan start"; break;
                 case 0x02 : cpToken = "P scan start";   break;
                 case 0x03 : cpToken = "F scan start";   break;
                 case 0x12 : cpToken = "FP scan start";  break;
                    // etc, etc, a lot has been omitted here
                 default:
                    sprintf( sz80Temp, "Scan[%02X]", (int)iSubcode );
                    cpToken = sz80Temp;
                    break;
               } // end switch < "sub-command" after 0x0E, *WITHOUT* data >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( iMsgLength >= 8 )  // subcode followed by DATA ? Must be a "set"-command..
               { i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+6, 2*(iMsgLength-7) );
                 sprintf( cpComment, "%s:%ld", cpToken, (long)i32Value ); // for the log..
               }
              else // only command and sub-command ..
               { strcpy( cpComment, cpToken );
               }
            } // end if( iMsgLength >= 6 )
           break; // end case 0x0E = "scan" commands

        case 0x0F: // "Read Split setting"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x0F = "Read Split setting"
           // Received from RS-BA1 : > r1 006 FE FE 94 E0 0F FD      ; Split ?
           // Response from IC-7300: > RX 007 FE FE E0 94 0F 00 FD   ; Split:0
           //                        pbMessage[0  1  2  3  4  5  6]
           // Also included in WSJT-X's "polling cycle" .
           if( iMsgLength >= 6 )  // 0x0F with or without data.. (NO subcodes here)
            { iUnifiedPN = RIGCTRL_PN_SPLIT_MODE;
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { cpToken = "Split";
               }
              if( iMsgLength >= 7 )  // 0x0F followed by DATA byte(s)..
               { i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+5,
                                       2*(iMsgLength/*7*/-6)/*nDigits*/ );
                 sprintf( cpComment, "%s:%ld", cpToken, (long)i32Value ); // "set" command with value (0=off, 1=on)
                 RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
               }
              else  // 0x0F followed by subcode but NO DATA ...
               { sprintf( cpComment, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND > ?
           break;  // end case 0x0F

        case 0x10: // "Send/read tuning step"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x10 = "Send/read tuning step"
           if( iMsgLength >= 6 )  // 0x10 with or without data.. (NO subcodes here)
            { cpToken = "TuningStep";
              iUnifiedPN = RIGCTRL_PN_TUNING_STEP;
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( iMsgLength >= 7 )  // 0x10 followed by DATA byte(s)..
               { i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+6,
                                       2*(iMsgLength/*7*/-6)/*nDigits*/ );
                 // 0=off, 1=100 Hz, 2=1kHz, ... 8=25kHz .
                 sprintf( cpComment, "%s:%ld", cpToken, (long)i32Value );
               }
              else  // 0x10 followed by subcode but NO DATA ...
               { sprintf( cpComment, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND > ?
           break; // end case 0x10

        case 0x11: // "Send/read attenuator setting". One of dozens of parameters that "RS-BA1" wanted to know..
           iMsgType   = RIGCTRL_MSGTYPE_RF_GAIN;  // attenuator contributes to "gain" !
           iUnifiedPN = RIGCTRL_PN_ATTENUATOR_DB;
           pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
           if( pPI != NULL )
            { cpToken = pPI->pszToken;
              // From RS-BA1 to SpecLab (on a client port) :
              //           [0  1  2  3  4  5]
              //  > r1 006 FE FE 94 E0 11 FD
              // From an IC-7300 back to SpecLab (on the RADIO port):
              //           [0  1  2  3  4  5  6]
              //  > RX 007 FE FE E0 94 11 00 FD
              //                          \|___ data byte ("current setting")
              if( iMsgLength > 6 )  // 0x11 followed by at least ONE data byte..
               { i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+5,
                                       2*(iMsgLength/*7*/-6)/*nDigits*/ );
                 sprintf( cpComment, "%s:%ld", cpToken, (long)i32Value );
                 RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
               }
              else  // 0x11 followed by subcode but NO DATA ...
               { sprintf( cpComment, "%s ?", cpToken ); // just a query
               }
            } // end if( pPI != NULL )
           break; // end case 0x11 = "Send/read attenuator setting"

        case 0x12: // "Send/read antenna selection"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x12 = "Send/read antenna selection"
           break; // end case 0x12

        case 0x13: // "Speech (speak) with voice synthesizer"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x13 = "Speak" (voice synth)
           break;

        case 0x14: // "Send/read an awful lot of parameters" (AF level, RF gain, squelch, audio pitch, noise reduction, ....)
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x14 = "Send/read a lot of parameters"
           // Seen while RS-BA1 tortured the bus with lots of polling traffic :
           //  pbMessage[0  1  2  3  4  5  6  7  8]
           //  > r1 007 FE FE 94 E0 14 0B FD        (query, iMsgLength=7)
           //  > RX 009 FE FE E0 94 14 0B 01 53 FD  (response with data, iMsgLength=9)
           //                       |  |_sub-cmd
           //                      |__ cmd
           if( iMsgLength >= 7 )  // 0x14 followed by subcode, with or without DATA ...
            { iSubcode = pbMessage[5];
              dblConvFactor = 100.0 / 255.0;  // conversion factor for most "percentages", instead of the strange CI-V internal unit
              switch( iSubcode )  // decode "sub-command" after 0x14 for the log:
               { case 0x01 : iUnifiedPN = RIGCTRL_PN_AUDIO_VOLUME_PERCENT; break;
                 case 0x02 : iUnifiedPN = RIGCTRL_PN_RF_GAIN_PERCENT;      break;
                 case 0x03 : iUnifiedPN = RIGCTRL_PN_SQUELCH_LEVEL_PERCENT; break;
                 case 0x04 : break;
                 case 0x05 : break;
                 case 0x06 : iUnifiedPN = RIGCTRL_PN_NOISE_REDUCTION_PERCENT; break;
                 case 0x07 : iUnifiedPN = RIGCTRL_PN_PASSBAND_TUNING_POS1; break;
                 case 0x08 : iUnifiedPN = RIGCTRL_PN_PASSBAND_TUNING_POS2; break;
                 case 0x09 : iUnifiedPN = RIGCTRL_PN_CW_PITCH_HZ;
                    dblConvFactor = 900.0/*Hz*/ / 255.0; // here: in Hertz, not the strange CI-V internal unit
                    break;
                 case 0x0A : iUnifiedPN = RIGCTRL_PN_RF_POWER_SETTING_PERCENT; break;
                 case 0x0B : iUnifiedPN = RIGCTRL_PN_MIC_GAIN_PERCENT;     break;
                 case 0x0C : iUnifiedPN = RIGCTRL_PN_KEYER_SPEED_PERCENT;  break;
                 case 0x0D : iUnifiedPN = RIGCTRL_PN_NOTCH_POS_PERCENT;    break;
                 case 0x0E : iUnifiedPN = RIGCTRL_PN_COMP_SETTING_PERCENT; break;
                 case 0x0F : iUnifiedPN = RIGCTRL_PN_BREAK_IN_DELAY_PERCENT; break;
                 case 0x12 : iUnifiedPN = RIGCTRL_PN_NBL_SETTING_PERCENT;  break;
                 case 0x15 : iUnifiedPN = RIGCTRL_PN_MONITOR_GAIN_PERCENT; break;
                 case 0x16 : iUnifiedPN = RIGCTRL_PN_VOX_GAIN_PERCENT;     break;
                 case 0x17 : iUnifiedPN = RIGCTRL_PN_ANTI_VOX_GAIN_PERCENT; break;
                 case 0x18 : break;
                 case 0x19 : iUnifiedPN = RIGCTRL_PN_BRIGHTNESS_PERCENT;   break;
                 default   : break;
               } // end switch < "sub-command" after 0x14 >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { sprintf(sz80Temp, "0x14[%02X]", (int)iSubcode );
                 cpToken = sz80Temp;
               }
              if( iMsgLength >= 9 )  // 0x14+subcode followed by DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7  8]
                 // > RX 009 FE FE E0 94 15 12 00 80 FD ; SWRMeterLvl
                 //       |              |  |  |___|-- value in some phantasy unit,
                 //   iMsgLength=9      cmd subcode       usually 4 digit BCD, MSDigit first
                 i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+6,
                                       2*(iMsgLength/*9*/-7)/*nDigits*/ );
                 sprintf( cpComment, "%s:%ld", cpToken, (long)i32Value ); // "set" command with value in some phantasy unit
                 RigCtrl_SetParamValue_Double( pRC, pPI, (double)i32Value * dblConvFactor );
               }
              else  // 0x15 followed by subcode but NO DATA ...
               { sprintf( cpComment, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND + SUB-COMMAND > ?
           break; // end case 0x14

        case 0x15: // "Send/read noise or S-meter squelch status", etc etc
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           //                                     pbMessage[0  1  2  3  4  5  6]
           // Seen when RS-BA1 took over control: > r1 007 FE FE 94 E0 15 01 FD
           //                                                           |  |_sub-cmd
           //                                                           |__ cmd
           // Note: The radio does NOT send any of these messages on its own
           //       - we must ASK FOR them, e.g. the current "S meter" value
           //         as Spectrum Lab periodically does (not very often) by
           //         calling RigCtrl_SendCmdToReadUnifiedPN(RIGCTRL_PN_S_METER_LEVEL_DB)
           //         which, "on the next occasion", would send cmd 0x15 subcmd 0x02.
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           if( iMsgLength >= 7 )  // 0x15 followed by subcode, with or without DATA ...
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // decode "sub-command" after 0x15 for the log:
               { case 0x01 :  // "Read noise or S-meter squelch status"  (IC-7300)
                    iUnifiedPN = RIGCTRL_PN_SQUELCH_STATUS;
                    break;
                 case 0x02 : iUnifiedPN = RIGCTRL_PN_S_METER_LEVEL_DB;
                    dblConvFactor = (9.0/*S-units*/*6.0/*dB*/+60.0/*db*/) / 241.0/* strange value for "S9+60*/;
                    break;
                 case 0x05 : // cpToken = "SquelchStat";
                    break;
                 case 0x11 : iUnifiedPN = RIGCTRL_PN_POWER_METER_PERCENT;
                    dblConvFactor = (100.0/*percent*/ / 213.0/* strange value for "100 %" */ );
                    break;
                 case 0x12 : iUnifiedPN = RIGCTRL_PN_SWR_METER_LEVEL; // here: 0 = SWR1.0 ... 100 = SWR 3.0
                    dblConvFactor = (100.0 / 120.0/* strange value for "SWR = 3.0" */ );
                    break;
                 case 0x13 : iUnifiedPN = RIGCTRL_PN_ALC_METER_LEVEL;
                    dblConvFactor = (100.0 / 120.0/* strange value for "120 = max", whatever that means*/ );
                    break;
                 case 0x14 : iUnifiedPN = RIGCTRL_PN_COMP_METER_LEVEL_DB;
                    // here: convert into decibel, 0=no compression, 30 dB = maximum reading for IC-7300)
                    dblConvFactor = (30.0 / 241.0/* strange value for "30 dB compression" */ );
                    break;
                 case 0x15 : iUnifiedPN = RIGCTRL_PN_SUPPLY_VOLTATE_mV;
                    // here: convert into millivolts, not the strange CI-V internal unit (241 = 16 V)
                    dblConvFactor = (16000.0/*mV*/ / 241.0/* strange value for "16 Volts" */ );
                    break;
                 case 0x16 : iUnifiedPN = RIGCTRL_PN_DRAIN_CURRENT_mA;
                    // here: convert into milliamps,  not the strange CI-V internal unit (241 = 25 A)
                    dblConvFactor = (25000.0/*mA*/ / 241.0/* strange value for "25 Ampere" */ );
                    break;
                 default   : sprintf(sz80Temp, "Meter[%02X]", (int)iSubcode );
                    // cpToken = sz80Temp;
                    break;
               } // end switch < "sub-command" after 0x15 >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { sprintf(sz80Temp, "Meter[%02X]", (int)iSubcode );
                 cpToken = sz80Temp;
               }
              if( iMsgLength >= 9 )  // 0x15+subcode followed by DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7  8]
                 // > RX 009 FE FE E0 94 15 12 00 80 FD ; SWRMeterLvl
                 //       |              |  |  |___|-- value in some phantasy unit,
                 //   iMsgLength=9      cmd subcode       usually 4 digit BCD, MSDigit first
                 i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+6,
                                       2*(iMsgLength/*9*/-7)/*nDigits*/ );
                 sprintf( cpComment, "%s:%ld", cpToken, (long)i32Value ); // "set" command with value in some phantasy unit
                 RigCtrl_SetParamValue_Double( pRC, pPI, (double)i32Value * dblConvFactor );
               }
              else  // 0x15 followed by subcode but NO DATA ...
               { sprintf( cpComment, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND + SUB-COMMAND > ?
           break; // end case 0x15 = "Send/read noise or S-meter squelch status", etc etc

        case 0x16: // Preamp, AGC speed, Noise blanker, Audio peak filter, Noise reduction, auto notch, etc etc
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           // Seen when RS-BA1 took over control: > r1 007 FE FE 94 E0 16 43 FD
           //                           but also: > r1 008 FE FE 94 E0 16 46 00 FD
           if( iMsgLength >= 7 )  // 0x1B followed by subcode, with or without DATA ...
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // decode "sub-command" after 0x15 for the log:
               { case 0x02 : iUnifiedPN = RIGCTRL_PN_PREAMP_SETTING; break;
                 case 0x12 : iUnifiedPN = RIGCTRL_PN_AGC_SPEED;      break;
                 case 0x22 : iUnifiedPN = RIGCTRL_PN_NOISE_BLANKER_ON; break;
                 case 0x40 : iUnifiedPN = RIGCTRL_PN_NOISE_REDUCT_ON; break;
                 case 0x41 : iUnifiedPN = RIGCTRL_PN_AUTO_NOTCH_ON;   break;
                 case 0x42 : iUnifiedPN = RIGCTRL_PN_RPTR_TONE_ON;   break;
                 case 0x43 : iUnifiedPN = RIGCTRL_PN_TONE_SQUELCH_ON; break;
                 case 0x44 : iUnifiedPN = RIGCTRL_PN_COMP_ON;        break;
                 case 0x45 : iUnifiedPN = RIGCTRL_PN_MONITOR_ON;     break;
                 case 0x46 : iUnifiedPN = RIGCTRL_PN_VOX_ON;         break;
                 case 0x47 : iUnifiedPN = RIGCTRL_PN_BK_IN_ON;       break;
                 case 0x48 : iUnifiedPN = RIGCTRL_PN_MANUAL_NOTCH_ON; break;
                 case 0x4F : iUnifiedPN = RIGCTRL_PN_TWIN_PEAK_FILTER_ON; break;
                 case 0x50 : iUnifiedPN = RIGCTRL_PN_DIAL_LOCK_ON;   break;
                 case 0x56 : iUnifiedPN = RIGCTRL_PN_DSP_SHARP_SOFT; break;
                 case 0x57 : iUnifiedPN = RIGCTRL_PN_MANUAL_NOTCH_WIDTH; break;
                 case 0x58 : iUnifiedPN = RIGCTRL_PN_SSB_TX_BANDWIDTH; break;
                 default   : break;
               } // end switch < "sub-command" after 0x16 >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { sprintf(sz80Temp, "0x16[%02X]", (int)iSubcode );
                 cpToken = sz80Temp;
               }
              if( iMsgLength >= 8 )  // 0x16+subcode followed by one or more DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7  8]
                 // > RX 009 FE FE E0 94 16 12 00 80 FD ; SWRMeterLvl
                 //       |              |  |  |___|-- value in some phantasy unit,
                 //   iMsgLength=9      cmd subcode       usually 4 digit BCD, MSDigit first
                 i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+6,
                                       2*(iMsgLength/*9*/-7)/*nDigits*/ );
                 sprintf( cpComment, "%s:%ld", cpToken, (long)i32Value ); // "set" command with value in some phantasy unit
                 RigCtrl_SetParamValue_Int( pRC, pPI, i32Value ); // no conversion factor required here
               }
              else  // 0x16 followed by subcode but NO DATA ...
               { sprintf( cpComment, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND + SUB-COMMAND > ?
           break; // end case 0x16

        case 0x17: // Send CW messages. Also seen when "RS-BA1 took over control:
                   // > r1 007 FE FE 94 E0 17 FF FD
                   //                         |____ "FF stops sending CW messages"
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           sprintf( cpComment, "CWmsg:%02X", (int)pbMessage[5] ); // for the log..
           break; // end case 0x17

        case 0x18: // Turn transceiver on/off (includes "wake-up via USB")
           // This was the first message received from RS-BA1 :
           //  pbMessage[0  1  2  3  4  5  6]
           // >  r1 007 FE FE 94 E0 18 01 FD
           if( iMsgLength >= 7 )
            { iMsgType = RIGCTRL_MSGTYPE_POWER_ON_OFF;
              sprintf( cpComment, "Power:%s", (pbMessage[5]==0) ? "OFF" : "ON" );
            }
           break; // end case 0x18

        case 0x19: // Read the transceiver ID (beware, different meaning depending on length) ? ?
           //   | Example from SL's 'CAT traffic' display, from an IC-7300 :
           //   '------------------------------,
           //   Time           Len Pre.  Adddr |     Postamble
           // > 14:39:59.2: TX 007 FE FE 94 E0 19 00 FD  |      (request the radio ID)
           // > 14:40:00.1: RX 008 FE FE E0 94 19 00 94 FD      (response with radio ID)
           //                                     |___|--- 0x19 followed by 0x00 AND a 'data byte' !!
           //
           //   [IC7610CIV] page 4 : "Cmd 0x19 : Sub Cmd = 0x00  Data = ??? : "Read the transceiver ID".
           if( iMsgLength >= 8 )  // 0x19 followed by TWO BYTES (!) before the postamble:
            { switch( pbMessage[5] )  // "sub-command" after 0x19 ?
               { case 0x00 :  // Guess what follows is the "DEFAULT CI-V address" alias "Transceiver ID code" ...
                    pRC->iDefaultAddress = (int)pbMessage[6];
                    sprintf( cpComment, "rig=%s", RigCtrl_DefaultAddressToString(
                        pRC->iRadioCtrlProtocol, pRC->iDefaultAddress) ); // for the log..
                    iMsgType = RIGCTRL_MSGTYPE_RADIO_ID;
                    pRadioInfo = RigCtrl_GetRadioInfoByDefaultAddress(
                        pRC->iRadioCtrlProtocol, pRC->iDefaultAddress ); // for the 'capabilities'
                    if( pRadioInfo != NULL )
                     {  pRC->iCapabilities = pRadioInfo->iCapabilities;
                     }
                    else // radio-info not available .. but be optimistic, it may be some brand-new model !
                     {  pRC->iCapabilities = RIGCTRL_CAPS_ALL;
                     }
                    break;
                 default   :  // anything else is neither documented in [IC7610CIV] nor [DF4OR_CIV] ...
                    break;
               }
            }
           break; // end case 0x19

        case 0x1A: // send/read memory contents, band stacking registers, memory keyer contents, etc
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           if( iMsgLength >= 7 )  // 0x1A followed by at least ONE byte for the sub-command...
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x1A ?
               { case 0x00 :  // "Send / read memory contents"
                    sprintf( cpComment, "mem[%d]", (int)pbMessage[6] ); // for the log..
                    break;
                 case 0x01 :  // "Send / read band stacking registers"
                    strcpy( cpComment, "band stack" );
                    break;
                 case 0x02 :  // "Send / read memory keyer contents"
                    strcpy( cpComment, "keyer mem" );
                    break;
                 case 0x03 :  // "Send / read selected IF filter"
                    // Surprisingly, Cmd 0x1A sub 0x03 is included in WSJT-X's "polling cycle".
                    // Thus this strange parameter (but converted into Hz) now has its own PN :
                    iUnifiedPN = RIGCTRL_PN_FILTER_BANDWIDTH;
                    pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
                    if( pPI != NULL )
                     { cpToken = pPI->pszToken;
                     }
                    else
                     { cpToken = "FilterBW";
                     }
                    // The strange unit used by Icom / IC-7300 is specified as below:
                    // > Cmd 0x1A subcmd 0x03: Send/read the selected filter width.
                    // >   ( AM: 00=200 Hz to 49=10 kHz;
                    // >     other than AM modes:
                    // >          00=50 Hz to 31/40=2700 Hz/3600 Hz)
                    // Heavens, no. A CONVERSION FORMULA would be asking too much.
                    // We're not interested in AM, where the formula would be
                    //       AM_BW_Hz = ( <Filter-Nr> + 1 ) * 200 Hz .
                    // But, for SSB and CW, if  "31" means 2700 Hz,
                    //                          "40" means 3600 Hz,
                    //      that's (3600-2700) Hz / (40-31) Steps = 100 Hz / step
                    //         near the end of the scale, and
                    //             (2700- 50 ) Hz / (31-0)  Steps = 85 Hz / step.
                    //      The truth must be somewhere in between. Say 88 Hz / step:
                    //   SSB_BW_Hz = ( <Filter-Nr> ) * 85 Hz + 50 Hz ??
                    //     "00"    :       0         * 85 Hz + 50 Hz = 50 Hz
                    //     "01"    :       1         * 85 Hz + 50 Hz = 138 Hz
                    //     "31"    :      31         * 85 Hz + 50 Hz = 2778 Hz
                    //     "40"    :      40         * 85 Hz + 50 Hz = 3570 Hz
                    //
                    // Example with an IC-7300 in USB, "FIL2" (seen in each WSJT-X polling cycle):
                    // pbMessage[0  1  2  3  4  5  6  7]
                    // > TX 007 FE FE 94 E0 1A 03 FD    ; FilterBW ?
                    // > RX 008 FE FE E0 94 1A 03 28 FD ; FilterBW = "28" .
                    if( iMsgLength >= 7 ) // valid SET-command ?
                     { i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+6, 2/*digits*/ );
                       i32Value = 50 + 85 * i32Value; // strange number to bandwidth in Hertz
                       RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
                       sprintf( cpComment, "%s=%d Hz", cpToken, i32Value );
                     }
                    else
                     { sprintf( cpComment, "%s ?", cpToken );
                     }
                    break;
                 case 0x04 :  // "Send / read selected AGC time constant"
                    strcpy( cpComment, "AGC speed" );
                    break;
                 case 0x05 :  // "Send / read an awful lot of RIG-SPECIFIC PARAMETERS"
                    // Typical QUERY (seen after starting RS-BA1 on client port #1):
                    //      pbMessage[0  1  2  3  4  5  6  7  8]
                    //      > r1 009 FE FE 94 E0 1A 05 00 60 FD
                    //           |_|                   |___| '-- postamble
                    //           iMsgLength=9            '------ 4-digit "subcode"
                    if( iMsgLength >= 9 ) // valid QUERY or SET-command ?
                     { iSubcode = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+6, 4/*digits*/ );
                       // Actually, the list of subcodes addressed with this
                       // 4-digit BCD extends over MULTIPLE PAGES in the IC-7610's CI-V manual.
                       // Very unfortunately, the meaning of those 4-digit "subcodes"
                       // is extremely rig specific.
                       // Not even an IC-7300 and IC-7610 are compatible. Thus:
                       iUnifiedPN = RigCtrl_SubcodeFrom1A05_to_UnifiedParameterNumber(pRC, iSubcode);
                       if( iUnifiedPN != RIGCTRL_PN_UNKNOWN )
                        { cpToken = RigCtrl_UnifiedParameterNumberToString(iUnifiedPN);
                        }
                       else // must be one of the "less important" parameters..
                        {   // .. at least RigControl.c doesn't know what it is.
                          sprintf( sz80Temp, "Cfg[%04d]", (int)iSubcode );
                          cpToken = sz80Temp;
                        }
                       if( iMsgLength > 9 ) // long enough for a "SET"-command.
                        { i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+8,
                                       2*(iMsgLength/*11*/-9)/*nDigits*/ );
                          // Typical RESPONSE (from an IC-7300, here: subcode 0060
                          //   -> iUnifiedPN = RIGCTRL_PN_AF_OUTPUT_LEVEL_TO_ACC =
                          // "AF output level to ACC/USB", iMsgLength=11, 2 byte data = 4 BCD digit
                          //      pbMessage[0  1  2  3  4  5  6  7  8  9 10]
                          //      > RX 00B FE FE E0 94 1A 05 00 60 01 30 FD
                          //      _________________________________|___|
                          //     | nDigits = 2 * (11-9) = 4, i.e. BCD-data with
                          //   a 4-digit "value" (here: volume ranging from 000 to 0255 = 100%)
                          sprintf( cpComment, "%s=%d", cpToken, (long)i32Value );
                        }
                       else // message too short for a "set"-command, so guess it's a QUERY ("get"):
                        { sprintf( cpComment, "%s ?", cpToken );
                        }
                       // Seen when launching "RS-BA1", connected via SpecLab to an IC-7300:
                       //    Cfg[0052]=0 ( "Function > Memo Pad Quantity" : 5 channels )
                       //    Cfg[0057]=0 ( "Function > Quick RIT/deltaTX Clear" :  OFF )
                       //    Cfg[0071]=1 ( "Connectors > Phones > Level" ? )
                       //    Cfg[0066]=3 ( IC-7300: "MOD input connector during DATA" : 3=USB.
                       //                  Totally different on IC-7610 :
                       //                  0066 would be "Screen Capture Keyboard".. EEEK !!)
                       switch( iUnifiedPN )
                        { case  RIGCTRL_PN_CIV_USB_ECHO : // 0x1A 0x05 Cfg[0075] : "Echo back setting for CI-V operation from USB"
                             // [G3NRW_CIV]
                             // > Note that the description of the "Set echo back setting"
                             // >  in the Command Table in Section 19 of the
                             // >  IC-7300 Full Manual is misleading. The
                             // >  parameter 00 sets the control to OFF
                             // >  and the parameter 01 sets it to ON .
                             // Here, each client may have prefer its own setting.
                             // Most clients will not know about how to configure
                             // the CI-V "echo back setting" via CI-V itself,
                             // thus in Spectrum Lab, the initial setting of
                             //  T_RigCtrlInstance.PortInstance[x].fEmulateCIVEcho
                             // can be configured via checkmark in the setup dialog.
                             if( iRigCtrlPort != RIGCTRL_PORT_RADIO ) // received on one of the "CLIENT"-ports ?
                              { iClientAction = RIGCTRL_CLIENT_ACTION_FAKE_RESPONSE; // "do NOT forward this to the real radio" !
                              }
                             break;
                          default: // ignore hundreds(!) of other "subcodes" after 0x1A 0x05
                             break;
                        } // end switch( iUnifiedPN )
                     } // end < message long enough for cmd 0x1A 0x05 >
                    break; // end case "Send / read an awful lot of RIG-SPECIFIC PARAMETERS"
                 case 0x06 :  // (IC-7300) : "Data mode with filter width settings"
                    // Example with an IC-7300 in USB, "FIL2" (seen in each WSJT-X polling cycle):
                    // pbMessage[0  1  2  3  4  5  6  7]
                    // > TX 007 FE FE 94 E0 1A 06 FD          ; Data/Filter ?
                    // > RX RX 009 FE FE E0 94 1A 06 00 00 FD ; Data mode OFF, no filter info.
                    //                                |  |__ only when DATA ON: 01=Fil1, 02=Fil2, 03=Fil3
                    //                                0=Data mode off, 1=Data on.
                    iUnifiedPN = RIGCTRL_PN_DATA_MODE;
                    pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
                    if( pPI != NULL )
                     { cpToken = pPI->pszToken;
                     }
                    else
                     { cpToken = "DataMode";
                     }
                    if( iMsgLength >= 8 ) // valid SET-command or GET-response ?
                     { i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+7, 2/*digits*/ );
                       RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
                       sprintf( cpComment, "%s=%d", cpToken, i32Value );
                     }
                    else
                     { sprintf( cpComment, "%s ?", cpToken );
                     }
                    break;
                 case 0x07 :  // "NTP server access" (IC-7610) / "IP+ Setting" (IC-7300)
                 case 0x08 :  // "NTP access result" (IC-7610) / not used in IC-7300
                 case 0x09 :  // "AF Mute"           (IC-7610) / not used in IC-7300
                    iUnifiedPN = RigCtrl_SubcodeFrom1A_to_UnifiedParameterNumber(pRC, iSubcode);
                    pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
                    if( pPI != NULL )  // anything known about this sub-command ? Do we care ?
                     { cpToken = pPI->pszToken;
                       if( iMsgLength > 7 ) // valid SET-command or GET-response ?
                        { // Here just one of the RESPONSES from an IC-7300,
                          // after being asked by RS-BA1 for whatever-it-was :
                          //          [0  1  2  3  4  5  6  7]
                          // > RX 008 FE FE E0 94 1A 07 00 FD
                          i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+6,
                                         2*(iMsgLength/*8*/-7)/*nDigits*/ );
                          RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
                          sprintf( cpComment, "%s=%d", cpToken, i32Value );
                        }
                       else // Just a QUERY .. one of hundreds seen from RS-BA1 :
                        { //          [0  1  2  3  4  5  6]
                          // > r1 007 FE FE 94 E0 1A 07 FD (other)
                          sprintf( cpComment, "%s ?", cpToken );
                        }
                     }
                    else // don't care what it is .. DL4YHF was sick of hard-coding RIG-SPECIFIC details here
                     {
                     }
                    break;
                 default:
                    break;
               } // end switch < "sub-command" after 0x1A >
            } // end if < sufficient message length for this command ? >
           break; // end case 0x1A

        case 0x1B: // repeater tone frequency, TSQL tone frequency
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           if( iMsgLength >= 7 )  // 0x1B followed by subcode, with or without DATA ...
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // decode "sub-command" after 0x1B for the log:
               { case 0x00 : iUnifiedPN = RIGCTRL_PN_REPEATER_TONE_FREQ; break;
                 case 0x01 : iUnifiedPN = RIGCTRL_PN_TONE_SQUELCH_FREQ;  break;
                 default   : break;
               } // end switch < "sub-command" after 0x1B >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { sprintf(sz80Temp, "0x1B[%02X]", (int)iSubcode );
                 cpToken = sz80Temp;
               }
              if( iMsgLength >= 10 )  // 0x1B+subcode followed by DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7  8  9]
                 // > RX 00A FE FE E0 94 1B 01 00 08 85 FD ; TSQL Tone
                 //       |              |  |  |______|-- value in 0.1 Hz units,
                 //   iMsgLength=10     cmd subcode       6 digit BCD, MSDigit first
                 d = 0.1 * (double)RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+6,
                                       2*(iMsgLength/*10*/-7)/*nDigits*/ );
                 sprintf( cpComment, "%s:%.1lf Hz", cpToken, (double)d ); // "set" command with value in Hz
                 RigCtrl_SetParamValue_Double( pRC, pPI, d );
               }
              else  // 0x1B followed by subcode but NO DATA ...
               { // Seen when starting "RS-BA1" : > r1 007 FE FE 94 E0 1B 00 FD
                 sprintf( cpComment, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND + SUB-COMMAND > ?
           break; // end case 0x1B

        case 0x1C: // transceiver status (RX/TX), antenna tuner, XFC, read TX frequency, etc .
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           // Received from RS-BA1 : > r1 007 FE FE 94 E0 1C 04 FD  (with iMsgLen=7, only a QUERY)
           //                                                |__ "CIV_ANT" ?
           if( iMsgLength >= 7 )  // 0x1C followed by at least <??> bytes...
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x1C ?
               { case 0x00 : iUnifiedPN = RIGCTRL_PN_TRANSMITTING;
                    // Used by WSJT-X (Hamlib) for RX/TX switching. Must keep track of this
                    // regardless of who sent the command, because the IC-7300's spectrum scope
                    // will stop sending 'waveform data' while transmitting (in MOST cases).
                    break;
                 case 0x01 : iUnifiedPN = RIGCTRL_PN_ATU_ENABLED; break;
                 case 0x02 : iUnifiedPN = RIGCTRL_PN_XFC_PRESSED; break;
                 case 0x03 : iUnifiedPN = RIGCTRL_PN_TRANSMIT_FREQ; break;
                 case 0x04 : iUnifiedPN = RIGCTRL_PN_CIV_OUTPUT_FOR_ANT; // "CI-V output for ANT"
                    break;
                 default   :
                    break;
               } // end switch < "sub-command" after 0x1C >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { sprintf(sz80Temp, "TrxSt[%02X]", (int)iSubcode );
                 cpToken = sz80Temp;
               }
              if( iMsgLength >= 8 )  // 0x1C+subcode followed by one or more DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7]
                 // > RX 008 FE FE E0 94 1C 04 00 FD
                 //       |              |  |  '-- value, here only two digits,
                 //   iMsgLength=8      cmd subcode       guess this is BCD, MSDigit first
                 i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+6,
                                       2*(iMsgLength/*8*/-7)/*nDigits*/ );
                 sprintf( cpComment, "%s:%ld", cpToken, (long)i32Value ); // "set" command with value
                 RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
               }
              else  // 0x1C followed by subcode but NO DATA ...
               { sprintf( cpComment, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient message length for this command ? >
           break; // end case 0x1C

        case 0x1E: // read number of TX frequency bands, TX band edge frequencies, etc
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           if( iMsgLength >= 7 )  // 0x1E followed by at least <??> bytes...
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x1E ?
               { default   : sprintf(sz80Temp, "TxBands[%02X]", (int)iSubcode );
                    cpToken = sz80Temp;
                    break;
               } // end switch < "sub-command" after 0x1B >
              if( iMsgLength >= 8 )  // 0x1E+subcode followed by one or more DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7]
                 // > RX 008 FE FE E0 94 1C 04 00 FD
                 //       |              |  |  '-- value, here only two digits,
                 //   iMsgLength=8      cmd subcode       guess this is BCD, MSDigit first
                 i32Value = RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+6,
                                       2*(iMsgLength/*8*/-7)/*nDigits*/ );
                 sprintf( cpComment, "%s:%ld", cpToken, (long)i32Value ); // "set" command with value
               }
              else  // 0x1E followed by subcode but NO DATA ...
               { sprintf( cpComment, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient message length for this command ? >
           break; // end case 0x1E

        case 0x21: // RIT frequency setting, delta TX setting (on/off)
           iMsgType = RIGCTRL_MSGTYPE_RIT_XIT;
           break; // end case 0x21

        case 0x25: // "Send/read main or sub-band frequency" (IC-7610)
                   // "Send/read the selected or unselected VFO frequency" (IC-7300)
           iMsgType = RIGCTRL_MSGTYPE_FREQUENCY;
           //                        pbMessage[0  1  2  3  4  5  6  7  8  9 10 11]
           // Received from RS-BA1 : > r1 007 FE FE 94 E0 25 00 FD
           // Response from IC-7300: > RX 00C FE FE E0 94 25 00 00 30 35 05 00 FD
           //                              |              |   | |____________|-- data, BCD, 10 decimal digits, LEAST significant digits first !
           //                          iMsgLength=12     cmd  0=selected VFO,
           //                                                 1=unselected VFO.
           if( iMsgLength >= 7 )  // 0x25 + subcode ("sel/unsel") with or without data..
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x25 ?
               { case 0x00 : cpToken = "SelVFOFreq";   break;
                 case 0x01 : cpToken = "UnselVFOFreq"; break;
                 default   : cpToken = "SomeOtherFreq"; break;
               } // end switch < "sub-command" after 0x1B >
              if( iMsgLength >= 8 )  // 0x25+subcode followed by one or more DATA bytes
               { i32Value = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbMessage+6,
                                       2*(iMsgLength/*12*/-7)/*nDigits*/ );
                 sprintf( cpComment, "%s:%ld", cpToken, (long)i32Value ); // "set" command with value
               }
              else  // 0x16 followed by subcode but NO DATA ...
               { sprintf( cpComment, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient message length for this command ? >
           break; // end case 0x25

        case 0x26: // Send/read operating mode and filter setting
           //                        pbMessage[0  1  2  3  4  5  6  7  8  9]
           // Received from RS-BA1 : > r1 007 FE FE 94 E0 26 01 FD
           // Response from IC-7300: > RX 00A FE FE E0 94 26 01 03 00 02 FD
           //                              |              |   |  |  |  |
           //                              |              |   |  |  |  |__ filter (1..3)
           //                              |              |   |  |  |__ DataMode, 0=off, 1=on
           //                              |              |   |  |__ OpMode, 3=CW
           //                          iMsgLength=10     cmd  0=selected VFO,
           //                                                 1=unselected VFO.
           if( iMsgLength >= 7 )  // 0x26 followed by at least ONE byte before the postamble:
            { iMsgType = RIGCTRL_MSGTYPE_OP_MODE;
              iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x25 ?
               { case 0x00 : cpToken = "SelVFO";   break;
                 case 0x01 : cpToken = "UnselVFO"; break;
                 default   : cpToken = "UnknVFO";  break;
               } // end switch < "sub-command" after 0x1B >
              if( iMsgLength >= 8 ) // 0x26 + VFO-Nr. + at least ONE data byte ?
               { iNewSetting = RigCtrl_IcomToRigCtrlOpmode( pRC, pbMessage[6] );
                 sprintf( cpComment, "%s:%s", cpToken, RigCtrl_OperatingModeToString(iNewSetting) );
                 cpComment += strlen( cpComment );
                 if( iMsgLength >= 9 )  // "Operating mode" byte followed by optional "Filter setting" ?
                  { switch( pbMessage[7] )
                     { case 1 : break;  // "FIL1" (normal filter bandwidth)
                       case 2 : iNewSetting |= RIGCTRL_OPMODE_NARROW;      break; // "FIL2" [IC7610CIV page 10]
                       case 3 : iNewSetting |= RIGCTRL_OPMODE_VERY_NARROW; break; // "FIL3"
                       default: break;
                     }
                    sprintf( cpComment, ",FIL%d", pbMessage[6] );
                    cpComment += strlen( cpComment );
                  } // end if < radio also reported the current FILTER SETTING >
                 if( iMsgLength >= 10 )  // "Filter setting" followed by "Data Mode"-flag ?
                  { sprintf( cpComment, ",Data=%d", (int)pbMessage[7] );
                    // cpComment += strlen( cpComment );
                  } // end if < radio also reported the current FILTER SETTING >
                 // Only accept the radio's reported value if we're not going to SEND a NEW operating mode:
                 if( ! ( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_OP_MODE) ) )
                  {
                    if( iNewSetting != pRC->iOpMode )
                     { pRC->iOpMode = iNewSetting;
                     }
                  }
               }
              else // not a "set"-command but a "get"-query :
               { sprintf( cpComment, "%sMode ?", cpToken );
               }
            } // end if < sufficient message length for this command ? >
           break; // end case 0x26

        case 0x27: // Read "Scope" waveform data (spectrum display):
           if( iMsgLength >= 7 )  // 0x27 followed by at least ONE byte before the postamble:
            { // (RS-BA1 periodically sent > r1 007 FE FE 94 E0 27 1C FD,
              //       i.e. NO DATA FIELD but a "query". So beware.. )
              //
              if( iRigCtrlPort==RIGCTRL_PORT_RADIO )
               { if( pRC->iExpectedResponseOnRadioPort == (0x2700 | pbMessage[5]) )
                  { // RigControl_Handler() has been waiting for THIS in case RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM ...
                    pRC->iResponseCountdown_ms = 0;   // don't wait any longer !
                    pRC->iExpectedResponseOnRadioPort = -1;   // avoid misinterpreting the next unspecific "OK" or "NOT OK".
                    // (only RigControl_Handler() decides what to ask NEXT)
                  } // end if < 0x27.. was the EXPECTED response for something >
               } // end if < message received FROM THE RADIO (not from any of the clients) > ?
              iMsgType = RIGCTRL_MSGTYPE_C_SPECTRUM; // initial guess: spectrum CONTROL
              iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x27 ?
               {  case 0x00 :  // Guess what follows is a block of 'Scope waveform data', in this case: part of a SPECTRUM.
                    iMsgType = RIGCTRL_MSGTYPE_SPECTRUM; // not spectrum-CONTROL but -DATA !
                    fIsUnsolicitedMsg = TRUE;
                    pRC->iScopeDataCountdown_ms = RIGCTRL_SPECTRUM_TIMEOUT_MS; // reload 'Scope waveform' timer (for timeout monitoring)
                    // Do we want to receive spectra at all (configurable) ?
                    if( pRC->iRigControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA )
                     { // Yes...
                       RigCtrl_ParseSpectrumData_CIV( pRC, pbMessage+6/*payload*/, iMsgLength-7/*payloadLength*/, cpComment );
                       // See ASCII graphics in RigCtrl_ParseSpectrumData_CIV()
                       // about what is considered the 'payload' for a SPECTRUM !
                     } // end if < ..ControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA >
                    else // option RIGCTRL_FLAG_WANT_SPECTRUM_DATA *not* set :
                     { // Turn off the unsolicited transmission of scope data.
                       // This reduces the 'bus load', so other messages
                       // (like VFO frequency control) have a smaller latency.
                       // Also prevents flooding the CAT traffic monitor .
                       pRC->iPostponedSetMessages |= (1<<RIGCTRL_MSGTYPE_C_SPECTRUM);
                       // The rest happens in RigControl_Handler(): As soon as
                       // there's an empty slot (in the heavy bus traffic),
                       // it will call RigCtrl_StartStopSpectrumData( pRC, TRUE ).
                     }
                    break;

                 case 0x10:  // "Send/read the Scope ON/OFF status (00=OFF, 01=ON)"
                    // If the scope (on the Icom's own display) is OFF,
                    // the radio also won't send "waveform data" via CI-V !
                    break;

                 case 0x11:  // "Send/read the Scope wave data output setting (00=OFF, 01=ON)"
                    // (should get here after RigCtrl_StartStopSpectrumData() )
                    // When "RS-BA1" connected through client port #1, it sent:
                    // > 21:22:57.3 r1 008 FE FE 94 E0 27 11 00 FD
                    //                                       |__ 0x00 = off, 0x01 = on
                    // No byte to distinguish between "main scope" and "sub scope" here !
                    if( iMsgLength < 8 ) // look mum, no DATA .. must be a QUERY .
                     { strcpy( cpComment, "Scope data on ?" );
                     }
                    else
                     { if( pbMessage[6] == 0x00 )
                        { strcpy( cpComment, "Scope data off" );
                          pRC->iScopeMode = RIGCTRL_SCOPE_MODE_OFF;
                          // When terminating RS-BA1 (connected to an IC-7300 through SL's client port),
                          //       it STOPPED the periodic transmission of scope waveform data :
                          // > 20:08:53.1 r1 008 FE FE 94 E0 27 11 00 FD (scope) ; Scope data off (from client)
                          // > 20:08:53.1 TX 008 FE FE 94 E0 27 11 00 FD (scope) ; Scope data off (forwarding to radio)
                          // > 20:08:53.2 RX 006 FE FE E0 94 FB FD       (OK)    ; response from radio
                          // > 20:08:53.2 t1 006 FE FE E0 94 FB FD       (OK)    ; forwarding response to client
                          // The TIMEOUT MONITOR for spectrum data in RigCtrl_Handler()
                          // will automatically try to kick the scope alive again.
                        }
                       else
                        { strcpy( cpComment, "Scope data on" );
                          if( pRC->iScopeMode == RIGCTRL_SCOPE_MODE_OFF )
                           { // cannot be "OFF" not, so assume "fixed frequency" mode:
                             pRC->iScopeMode = RIGCTRL_SCOPE_MODE_FIXED;
                           }
                          // Note: RIGCTRL_SCOPE_MODE_CENTER / FIXED set further below
                        }
                     }
                    break;

                 case 0x12:  // "Send/read the Scope MAIN/SUB setting (00=Fixed to MAIN)"
                    break;

                 case 0x13:  // "Send/read the Scope Single/Dual setting (00=Fixed to Single)"
                    break;

                 case 0x14:  // "Send/read the Scope Center mode or Fixed mode setting"
                    // pbMessage[6] : 0x00 = main scope,  0x01 = sub scope (IC-7851) ; IC-7300 : "fixed" 0x00
                    // pbMessage[7] : 0x00 = center mode, 0x01 = fixed mode
                    // (looks like in contrast to Yaesu, Icom cares for compatibility .. in most cases)
                    if( iMsgLength < 9 ) // again, no DATA .. must be a QUERY .
                     { strcpy( cpComment, "Scope C/F ?" );
                     }
                    else
                     {
                       if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { if( pbMessage[7] == 0x00 )
                           { pRC->iScopeMode = RIGCTRL_SCOPE_MODE_CENTER;
                             strcpy( cpComment, "Scope:Center" );
                           }
                          else
                           { pRC->iScopeMode = RIGCTRL_SCOPE_MODE_FIXED;
                             strcpy( cpComment, "Scope:Fixed" );
                           }
                        }
                     }
                    break;

                 case 0x15:  // "Send/read the span setting in the Center mode Scope"
                    //        pbMesssage[0  1  2  3  4  5  6  7]
                    // RS-BA1: > r1 008 FE FE 94 E0 27 15 00 FD (scope) ; span ?
                    //               |                    |
                    //           iMsgLength               0=main scope, 1=sub scope
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY .
                     { strcpy( cpComment, "Scope span ?" );
                     }
                    else  // iMsgLength > 8 -> not a query ("get") but a SET-command :
                     { // Response from IC-7300 :
                       //     pbMesssage[0  1  2  3  4  5  6  7  8  9 10 11 12]
                       //      > RX 00D FE FE E0 94 27 15 00 00 25 00 00 00 FD (scope) ; span=2500 Hz
                       //            |                       |____________|
                       //        iMsgLength=13               BCD value, HERE: LEAST significant digits first !
                       if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->dblScopeSpan_Hz = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbMessage+7,
                               /* number of digits: */ 2*(iMsgLength/*13*/-8) );
                          sprintf( cpComment, "span=%ld Hz", (long)pRC->dblScopeSpan_Hz );
                        }
                     }
                    break;

                 case 0x16:  // "Edge number setting in the Fixed mode Scope"
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY .
                     { strcpy( cpComment, "Scope edge# ?" );
                     }
                    else
                     { if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->iScopeEdgeNumber = pbMessage[7]; // range : 1..3 (per BAND.. what a strange concept)
                          sprintf( cpComment, "edge=%d", pRC->iScopeEdgeNumber );
                        }
                     }
                    break;

                 case 0x17:  // "Send/read the Scope hold function status"
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY .
                     { strcpy( cpComment, "Scope hold ?" );
                     }
                    else
                     { if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->iScopeHoldFlag = pbMessage[7]; // 0=hold OFF, 1=hold ON
                          sprintf( cpComment, "hold=%d", pRC->iScopeHoldFlag );
                        }
                     }
                    break;

                 case 0x18:  // "Send/read the Scope Attenuator setting" (IC-7851, n/a on IC-7300)
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY .
                     { strcpy( cpComment, "Scope att ?" );
                     }
                    else
                     { if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->dblScopeAttenuator_dB = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbMessage+7,
                               /* number of digits: */ 2*(iMsgLength-8) );
                          sprintf( cpComment, "att=%.1lf dB", pRC->dblScopeAttenuator_dB );
                        }
                     }
                    break;

                 case 0x19:  // "Send/read the Scope Reference level setting"
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY .
                     { strcpy( cpComment, "Scope ref ?" );
                     }
                    else
                     { // Response seen from an IC-7300 :
                       // pbMessage[0  1  2  3  4  5  6  7  8  9 10]
                       // > RX 00B FE FE E0 94 27 19 00 10 50 00 FD   (scope)    ; ref=+10.5 dB
                       if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->dblScopeRefLevel_dB = 0.01f * (float)RigCtrl_ParseBCD_CIV_MSDigitFirst( pbMessage+7, 4/*digits*/ );
                          if( pbMessage[9] == 0x01 )  // 0x00 = plus, 0x01 = minus
                           { pRC->dblScopeRefLevel_dB = -pRC->dblScopeRefLevel_dB;
                           }
                          sprintf( cpComment, "ref=%.1lf dB", pRC->dblScopeRefLevel_dB );
                        }
                     }
                    break;

                 case 0x1A: // "Send/read the Sweep speed setting"
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY .
                     { strcpy( cpComment, "Scope speed ?" );
                       // Seen from RS-BA1 :
                       //  > r1 008 FE FE 94 E0 27 1A 00 FD (scope)    ; Scope speed ?
                     }
                    else
                     { // Response seen from an IC-7300 :
                       // pbMessage[0  1  2  3  4  5  6  7  8  9 10]
                       //  > RX 009 FE FE E0 94 27 1A 00 00 FD  (scope) ; Speed:FAST
                       if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { switch( pbMessage[7] )  // 0x00 = fast, 0x01 = mid, 0x02 = slow
                           { case 0x00 : pRC->iScopeSpeed = RIGCTRL_SCOPE_SPEED_FAST;
                                  strcpy( cpComment, "Scope:FAST" );
                                  break;
                             case 0x01 : pRC->iScopeSpeed = RIGCTRL_SCOPE_SPEED_MID;
                                  strcpy( cpComment, "Scope:MID" );
                                  break;
                             case 0x02 : pRC->iScopeSpeed = RIGCTRL_SCOPE_SPEED_SLOW;
                                  strcpy( cpComment, "Scope:SLOW" );
                                  break;
                             default   : // ???
                                  break;
                           }
                        }
                     }
                    break;

                 case 0x1B: // "Send/read the Scope indication during TX in the Center mode" (0=off, 1=on)
                    if( iMsgLength <= 7/*!*/ ) // no DATA .. must be a QUERY .
                     { // Seen from RS-BA1 : > r1 007 FE FE 94 E0 27 1B FD
                       strcpy( cpComment, "Scope on TX ?" );
                     }
                    else
                     { // Response seen from an IC-7300 :
                       // pbMessage[0  1  2  3  4  5  6  7  8  9 10]
                       //  > RX 008 FE FE E0 94 27 1B 00 FD
                       //                             |
                       //                          Here, this isn't the "main/sub"-scope-flag
                       //                          but the OFF/ON status !
                       sprintf( cpComment, "Scope on TX:%s", (pbMessage[6]==0) ? "OFF":"ON" );
                     }
                    break;

                 case 0x1C: // "Send/read scope center frequency setting in the Center mode"
                    // 00=Filter center, 01=Carrier point center, 02=Carrier point center (Abs. Freq.))
                    // Seen shortly after launching RS-BA1 :
                    //          [0  1  2  3  4  5  6  7]
                    // > TX 007 FE FE 94 E0 27 1C FD    (scope) ; CenterMode ?
                    // > RX 008 FE FE E0 94 27 1C 01 FD (scope) ; CenterMode=1
                    if( iMsgLength <= 7 ) // no DATA .. must be a QUERY .
                     { strcpy( cpComment, "CenterMode ?" );
                     }
                    else      // message WITH data ..
                     { pRC->iScopeCenterMode = pbMessage[7]; // 0=RIGCTRL_SCOPE_CM_FILTER_CENTER / 1=CARRIER_POINT / 2=CARRIER_ABS (whatever that means)
                       sprintf( cpComment, "CenterMode=%d", pRC->iScopeCenterMode );
                     }
                    break;

                 case 0x1D: // "Send/read the Scope 'Video Band Width' setting" (narrow/wide)
                    // Seen shortly after launching RS-BA1 :
                    //          [0  1  2  3  4  5  6  7  8]
                    // > TX 008 FE FE 94 E0 27 1D 00 FD    (scope) ; CenterMode ?
                    // > RX 009 FE FE E0 94 27 1D 00 00 FD (scope) ; CenterMode=1
                    //                             |  |__ 0=narrow, 1=wide
                    //                             0=main scope, 1=sub scope
                    if( iMsgLength <= 7 ) // no DATA .. must be a QUERY .
                     { strcpy( cpComment, "VideoBW ?" );
                     }
                    else      // message WITH data ..
                     {
                       if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->iScopeVideoBW = pbMessage[7]; // 0=RIGCTRL_SCOPE_VBW_NARROR, 1=.._WIDE
                          sprintf( cpComment, "VideoBW=%d", pRC->iScopeVideoBW );
                        }
                     }
                    break;

                 case 0x1E: // "Send/read the Scope Fixed edge frequencies"
                    // There may be up to 12 selectable frequency ranges.
                    // Each range has THREE "edge numbers" (already in pRC->iScopeEdgeNumber).
                    // The command (read-response) contains RANGE, EDGE NUMBER,
                    //     and the two BCD-encoded edge frequencies (lower, higher).
                    // Which of the 12 frequency ranges is currenty in use, depends on the VFO frequency.
                    // To avoid having to read this repeatedly, only read the
                    // entire "table" ONCE, and store it in an array.
                    // Format :   (1) (2) (3) (4) (5) (6) (7) (8) (9) (10)(11)(12)
                    //           |X:X|0:X|X:X|X:X|X:X|X:X|0:0|X:X|X:X|X:X|X:X|0:0|
                    // range nr ---'   |   |   |   |   |   | | |   |   |   |   |
                    // edge nr (1..3)--' |<--- lower edge -->|<-- higher edge -->|
                    //                     |   |   |   |   | | |   |   |   |   |
                    //                    10+1 .........1GHZ |10+1 ...........1GHz
                    //                     Hz            100M| Hz              100MHz
                    // (expect those ZEROs in the 100 MHz- and 1 GHz to be used
                    //  sooner or later .. thinking about an IC-9700 with V/U/SHF)
                    // Icom's RS-BA1 reads out ALL these frequencies when starting:
                    // > r1 009 FE FE 94 E0 27 1E 01 01 FD  (scope)    ; 16..0 kHz
                    iRange = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbMessage+6, 2/*digits*/ );
                    iEdge  = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbMessage+7, 2/*digits*/ );
                    if( iMsgLength <= 9/*!*/ ) // no DATA, only RANGE- and EDGE-number .. must be a QUERY .
                     { sprintf( cpComment, "rng[%d],edg[%d] ?",(int)iRange, (int)iEdge );
                     }
                    else
                     {
                       iRange = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbMessage+6, 2/*digits*/ );
                       iEdge  = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbMessage+7, 2/*digits*/ );
                       if( (iRange>=1) && (iRange<=RIGCTRL_MAX_FIXED_EDGE_SCOPE_FREQ_RANGES)
                         &&(iEdge >=1) && (iEdge <=RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE) )
                        { pRC->ScopeFreqRange[iRange-1][iEdge-1].dblFmin_Hz = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbMessage+8, 10/*digits*/ );
                          pRC->ScopeFreqRange[iRange-1][iEdge-1].dblFmax_Hz = RigCtrl_ParseBCD_CIV_LSDigitFirst( pbMessage+13,10/*digits*/ );
                          sprintf( cpComment, "%ld..%ld kHz",
                             (long)(1e-3 * pRC->ScopeFreqRange[iRange-1][iEdge-1].dblFmin_Hz),
                             (long)(1e-3 * pRC->ScopeFreqRange[iRange-1][iEdge-1].dblFmax_Hz) );
                        }
                     }
                    break; // end case 0x27 0x1E ("Scope Fixed edge frequencies")

                 default : // purpose unknown ... future extensions ?
                    break;
               } // end switch < byte after 0x27 >
            } // end if < valid message length >
           break; // end case 0x27

        case 0x28: // Voice TX Memory (0=stop, 1 to 8 = memory #)
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           break; // end case 0x28

        case 0x29: // "directly specify main or sub band" (IC-7610)
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           break; // end case 0x29

        case 0xFA: // "NG" ("No-Good" or "FAilure"), indicates an error
           iMsgType = RIGCTRL_MSGTYPE_NOT_OK;
           if( iRigCtrlPort==RIGCTRL_PORT_RADIO ) // only care for "OK" / "Not OK" *FROM THE RADIO*...
            {
              pRC->iResponseCountdown_ms = 0; // don't wait any longer (there was a response, even though a BAD one)
              // Will usually get here when asking for something not supported by the radio.
              // Example: Asking an IC-7300 for the current SCOPE ATTENUATOR LEVEL :
              // > 21:50:22.7 TX 008 FE FE 94 E0 27 18 00 FD   (scope)  ; atten. ?
              // > 21:50:22.9 RX 006 FE FE E0 94 FA FD         (Not OK)
              if( pRC->iExpectedResponseOnRadioPort >= 0 )
               { sprintf( cpComment, "Rejected cmd 0x%X", (int)pRC->iExpectedResponseOnRadioPort );
                 // That's only the default comment. Human-friendly output below.
               }
              switch( pRC->iExpectedResponseOnRadioPort ) // we know the answer ("NG"), but what was the question ?
               { case 0x2714 : // have asked for the spectrum scope's "Center/Fixed" setting
                    strcpy( cpComment, "no scope-C/F" );
                    break;
                 case 0x2715 : // have asked for the spectrum scope's "Frequency Span"
                    strcpy( cpComment, "no scope-span" );
                    break;
                 case 0x2716 : // have asked for the spectrum scope's "Edge Number"
                    strcpy( cpComment, "no scope edges" );
                    break;
                 case 0x2717 : // have asked for the "Scope hold function status"
                    strcpy( cpComment, "no scope-hold" );
                    break;
                 case 0x2718 : // have asked for the spectrum scope's "Attenuator Level",
                    // which for example the IC-7300 doesn't support.
                    strcpy( cpComment, "no scope-att" );
                    break;
                 case 0x271E : // have asked for one of the scope's "frequency ranges":
                    // When deliberately exceeding the number of "scope frequency ranges"
                    //  ( > 13 for an IC-7300 ), the radio responded with "Not Good" :
                    // > 23:23:00.0 TX 009 FE FE 94 E0 27 1E 14 01 FD (scope)    ; edge[14][1] ?
                    // > 23:23:00.1 RX 006 FE FE E0 94 FA FD          (Not OK)
                    // If this happens, limit pRC->iScopeNumFreqRanges
                    //   and/or pRC->iScopeNumEdgesPerFreqRange
                    //   to what the radio *really* supports. But WHAT caused the "Not OK" ?
                    break; // end case "Not OK" for a SCOPE FREQUENCY RANGE (0x271E)
                 default:
                    sprintf( cpComment, "NotOK for 0x%X", (int)pRC->iExpectedResponseOnRadioPort );
                    break;
              } // end switch( pRC->iExpectedResponseOnRadioPort )
              pRC->iExpectedResponseOnRadioPort = -1;  // after "Not OK" from the radio,
              // we're not expecting to receive the "wanted" parameter value or whatever.
            } // end if < "Not OK" received FROM THE RADIO (not on one of the client ports) >
           break; // end case < receive "No-Good" / "Not OK">

        case 0xFB: // "OK" (success, "Fine Business") from the rig
           iMsgType = RIGCTRL_MSGTYPE_OK;
           // We will never know (exactly) which commands will be answered
           // with the SAME command-byte (in the response), and which will be
           // answered with this simple "OK" (especially not when a CLIENT had
           // sent the request) .. the client may know more about CI-V than RigControl itself,
           // especially if a new CI-V command will be answered with "OK"
           // or the value of a read-out parameter, or whatever.
           // Solution: Similar as already used for response "No-Good" (above):
           if( iRigCtrlPort==RIGCTRL_PORT_RADIO ) // only care for "OK" / "Not OK" *FROM THE RADIO*...
            { fIsExpectedResponse = TRUE;  // unspecific "OK" treated like an "expected response" !
            }
           break;

        default:  // anything else is ignored here
           iMsgType = RIGCTRL_MSGTYPE_UNKNOWN; // this parser couldn't process the message !
           break;
      } // end switch < command code, pbMessage[4] >

     if( fIsExpectedResponse )
      { // RigControl_Handler() has been waiting for THIS message..
        if( iRigCtrlPort==RIGCTRL_PORT_RADIO )
         { pRC->iResponseCountdown_ms = 0;   // don't wait any longer !
           pRC->iExpectedResponseOnRadioPort = -1;   // avoid misinterpreting the next unspecific "OK" or "NOT OK".
           // (only RigControl_Handler() decides what to ask NEXT)
         }
      } // end if < parsed the EXPECTED response >

   } // end if( ! fIsEcho )
  else // received message is an ECHO, not a response, not a command :
   { iMsgType = RIGCTRL_MSGTYPE_ECHO; // for filtering via pRC->dwRejectMessages4Log
   }

  //--------------------------------------------------------------------------
  // At this point, finished parsing the RECEIVED message, and extracted some
  //                info (like message type, expected response type, etc).
  // Next: Decide "what to do" on reception, e.g. forward to another port, etc.
  //--------------------------------------------------------------------------

  if( pRC->dwRejectMessages4Log & (1<<iMsgType) )
   { fShowInLog = FALSE; // avoid flooding the CAT traffic log by certain message types, e.g. spectrum scope waveforms
   }
  if( fShowInLog )  // BEFORE retransmitting anything, show the *received* message in the log:
   { RigCtrl_AddToTrafficLog( pRC, iRigCtrlPort, pbMessage, iMsgLength,
          iMsgType | RIGCTRL_MSGTYPE_FLAG_RX | (fIsExpectedResponse ? RIGCTRL_MSGTYPE_FLAG_EXPECTED : 0 ), // turn 'em green ?
          szComment );
   }

  if( iRigCtrlPort == RIGCTRL_PORT_RADIO ) // response received on the "RADIO" port ?
   { // May have to "forward" this message to one or more of the clients.
     // But some clients (like 'Hamlib' used by WSJT) are irritated by ANYTHING
     //       THEY DO NOT EXPECT (-> WSJT-X : "Ploink... RIG CONTROL ERROR").
     if( ! fIsEcho )
      { RigCtrl_CheckAndForwardResponseToClients( pRC, pbMessage, iMsgLength,
                      &msg_filter, iMsgType, fIsUnsolicitedMsg, szComment );
      }
   }
  else // iRigCtrlPort != RIGCTRL_PORT_RADIO -> command from one of the CLIENT ports ?
   { // "Echo-back" (Icom slang) the RECEIVED message (on the port on which it was received) ?
     //   Some applications developed when CI-V was a simple SINGE-WIRE interface
     //   still expect their transmitted data echoed back, possibly for collision detection .
     if( pPort->fEmulateCIVEcho ) // <- this may have been turned on/off further above via CI-V itself !
      { RigCtrl_SendAndLogMessage( pRC, iRigCtrlPort, pbMessage, iMsgLength,
              FALSE/*don't wait for completion*/,
              RIGCTRL_MSGTYPE_ECHO | RIGCTRL_MSGTYPE_FLAG_TX, szComment );
      } // end if( pPort->fEmulateCIVEcho )

     // Forward a client's command to the RADIO port immediately,
     //  or store it in a FIFO (to retransmit it LATER), ignore it, or what else ?
     // Here's ONE example of forwarding CI-V messages, with WSJT-X connected to client #2 :
     // > r2 006 FE FE 94 E0 03 FD                 (VFO freq)                [WSJT-X to SpecLab] -> RIGCTRL_CLIENT_ACTION_RETRANSMIT
     // > TX 006 FE FE 94 E0 03 FD                 (VFO freq)                [SpecLab to IC7300]
     // > RX 00B FE FE E0 94 03 00 70 56 03 00 FD  (VFO freq) ; 3567000.0 Hz [IC7300 to SpecLab] -> RIGCTRL_CLIENT_ACTION_RETRANSMIT
     // > t2 00B FE FE E0 94 03 00 70 56 03 00 FD  (VFO freq) ; 3567000.0 Hz [SpecLab to WSJT-X]
     if(  (iClientAction==RIGCTRL_CLIENT_ACTION_RETRANSMIT )
       && (pRC->iExpectedResponseOnRadioPort >= 0 ) ) // oops, cannot IMMEDIATELY retransmit command from client to radio !
      { iClientAction = RIGCTRL_CLIENT_ACTION_ENTER_FIFO; // .. so send whatever the client wants a few milliseconds LATER .
        // 2019-01-05 : Got here with pRC->iExpectedResponseOnRadioPort = 0x0701,
        //     shortly after being "connected" by WSJT-X on a client port :
        // > r2 007 FE FE 94 E0 07 01 FD (other)    ; SelVFO:01 [WSJT-X to SpecLab]
        // > t2 007 FE FE 94 E0 07 01 FD (echo)     ; SelVFO:01 [immediate echo to make Hamlib happy]
        // > TX 007 FE FE 94 E0 07 01 FD (other)    ; SelVFO:01 [SpecLab to IC7300, iExpectedResponseOnRadioPort now 0x0701]
        // > RX 007 FE FE 94 E0 07 01 FD (echo)                 [IC7300 to SpecLab, not the expected response yet]
        // > RX 006 FE FE E0 94 FB FD    (OK)                   [IC7300 to SpecLab, response to cmd 0x07]
        // > t2 006 FE FE E0 94 FB FD    (OK)                   [SpecLab to WSJT-X]
        // > r2 006 FE FE 94 E0 03 FD    (VFO freq)
        // Solution: See 'case 0xFB' (="OK") in RigCtrl_ParseCIV(), further above.
      } // end if < "wanted" to retransmit, but cannot because still waiting for a response on the RADIO port >
     switch( iClientAction )
      { case RIGCTRL_CLIENT_ACTION_NONE      : // "do nothing" (igore command from 3rd party client)
           break;
        case RIGCTRL_CLIENT_ACTION_RETRANSMIT: // immediately retransmit command from client to radio,
           // but also REGISTER THIS TYPE OF MESSAGE to forward the response to the client later:
           RigCtrl_RegisterCIVReponseForClient( pRC, iRigCtrlPort/*client-port*/, &msg_filter );
           if( pRC->pRadioPortTxFifo != NULL )
            { // Mark the "radio port" as occupied, until the response
              //      for the forwarded command (sent below) arrives :
              pRC->iExpectedResponseOnRadioPort = iCombinedCmdAndSubcode;
              RigCtrl_SendAndLogMessage( pRC, RIGCTRL_PORT_RADIO, pbMessage, iMsgLength,
                    TRUE/* wait for completion (*) ? */, iMsgType, szComment );
                    // (*) Tried WITH and WITHOUT "waiting for completion",
                    //     because WITH waiting, the worker thread for RigControl
                    //     was seriously slowed down when forwarding SPECTRUM DATA
                    //     to RS-BA1, *AFTER* RS-BA1 had disconnected from the
                    //     other end of the virtual Null-modem cable (com0com).
                    //     Since we RECEIVED these data over a serial port,
                    //     we should be able to RETRANSMIT *without* waiting.
                    //     But in the horrible "serial port API" in Windows,
                    //     things are not as simple as they should be,
                    //     and a thread can only wait for ONE outstanding
                    //     'FileWrite'-call (here: for a single CI-V message).
                    // With fWait=FALSE above, Spectrum Lab's waterfall wasn't
                    //     slowed down anymore, but the waterfall in RS-BA1
                    //     didn't work anymore, most likely because the
                    //     TRANSMITTED data (sent from here) were corrupted.
                    //   Got to emulate a nice simple reliable CIRCULAR TRANSMIT FIFO
                    //   (as in a microcontroller firmware) for Windoze one fine day !
            } // end if < address of the transmit-function available > ?
           break;
        case RIGCTRL_CLIENT_ACTION_ENTER_FIFO: // store command in FIFO, retransmit or execute later
           // NOT YET: RigCtrl_RegisterCIVReponseForClient( pRC, iRigCtrlPort/*client-port*/, &msg_filter );
           // NOT YET: RigCtrl_SendAndLogMessage( pRC, RIGCTRL_PORT_RADIO, ..) // forward to radio
           RigCtrl_AppendMessageToRadioPortTxFifo( pRC,
                iRigCtrlPort/*"transmit on behalf of THIS client"*/,
                pbMessage, iMsgLength );
           break;
        case RIGCTRL_CLIENT_ACTION_FAKE_RESPONSE: // immediately send a "fake response" to the client without forwarding anything to the radio
           RigCtrl_SendAndLogMessage( pRC, iRigCtrlPort/*client-port*/,
                    pPort->bTxMsg, pPort->iTxMsgLength,
                    TRUE/*wait for completion*/, iMsgType, szComment );
           break;
        default:
           break;
      } // end switch( iClientAction )
   }   // end else < command received on one of the CLIENT ports >


  return iMsgType;

} // end RigCtrl_ParseCIV()

//--------------------------------------------------------------------------
void RigControl_ProcessRxData(   // tries to de-packetize a 'serial bytestream'.
        // For CI-V, looks for the 0xFE 0xFE preamble, and the 0xFD postamble, etc..
        T_RigCtrlInstance *pRC,  // [in] 'Rig Control' instance (so far, only one, but who knows..)
        int  iRigCtrlPort )  // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 ?
{ // Called from the main threads ("GUI thread") after reception of a block of data.
  //        Initially only for data received from a serial port ("COM"),
  //        later (but not HERE) also used for CI-V embedded in UDP packets for LAN or WLAN.
  int iSource, nBytesReceived = 0;
  BYTE bTemp[1024], b;
  T_RigCtrl_PortInstance *pPortInstance;

  if( pRC == NULL ) // oops.. instance deleted but some thread is still calling us !
   { return;
   }
  if( pRC->initState != initState_Opened ) // oops.. didn't call RigControl_Init() yet, or called RigControl_Close()
   { return;
   }

  if( (iRigCtrlPort<0) || (iRigCtrlPort>/*!*/RIGCTRL_MAX_CLIENT_PORTS) )
   { // iRigCtrlPort = 0 = RIGCTRL_PORT_RADIO = communication port that "talks to the radio",
     // uses the same subroutine to process received data !
     return;  // not a valid 'port identifier' (not to be confused with a COM-port-number) !
   }

  pPortInstance = &pRC->PortInstance[iRigCtrlPort];

  // the keyer thread may have placed received bytes in a thread-safe FIFO.
  // Pass the content of that FIFO to RigControl, with a CI-V parser, etc:
  if( (iRigCtrlPort==RIGCTRL_PORT_RADIO) && (pRC->pRadioPortRxFifo!=NULL) )
   { nBytesReceived = CFIFO_Read( pRC->pRadioPortRxFifo, bTemp, sizeof(bTemp) );
   }

  // Convert the received data from the radio ("CI-V" or whatever protocol)
  //    into the an internal buffer, and parse it as soon as a "message" is complete:
  for(iSource=0; iSource<nBytesReceived; ++iSource)
   {
      b = bTemp[iSource];  // get the next BYTE (!) from the serial input stream
      // Collect all received bytes in an extra buffer until the end of the message is reached:
      if( pPortInstance->iRcvdMsgLength < RIGCTRL_MAX_MESSAGE_LENGTH )
       {  pPortInstance->bRcvdMsg[pPortInstance->iRcvdMsgLength++] = b;
       }
      else // couldn't append the new byte, count it as "garbage"
       {  ++pPortInstance->dwNumGarbageBytes;
       }

      // Check for end-of-message (depends on the protocol; radio-specific)
      switch( pRC->iRadioCtrlProtocol )
       {
        case RIGCTRL_PROTOCOL_NONE :   // "no" protocol, assume "normal text"..
        default :
             // in this case, "zero" bytes, carriage return, or new line are treated as end-markers:
             if( b==0 || b==10 || b==13 )
              {
                pPortInstance->iRcvdMsgLength = 0;   // begin receiving next frame
              }
             break;
        case RIGCTRL_PROTOCOL_ICOM_CI_V  :
             if( b==0xFD )  // received CI-V code for "End Of Message"
              {
               if( (pPortInstance->bRcvdMsg[0]==0xFE) && (pPortInstance->bRcvdMsg[1]==0xFE) )
                { // preamble code ok ?
                  if(  (pPortInstance->bRcvdMsg[2]==pRC->iRadioMasterAddr)
                     ||(pPortInstance->bRcvdMsg[2]==pRC->iRadioDeviceAddr) // listen to the "echo" too !
                     ||(pPortInstance->bRcvdMsg[2]==0x00) ) // directed to the "broadcast" address
                   { // Messages generated by a rig by turning the dial or by changing
                     // the mode are sent to this adress $00 (provided that "CI-V transceive"
                     // is set to ON. If OFF, no messages are generated by the rig).
                     // don't care for IO_b255MsgFromRadio[2] = transceiver's address (0x48 = IC706)
                     // and  for IO_b255MsgFromRadio[2] = controller's address (often 0xE0) .
                     // The "from"-address (=IO_b255MsgFromRadio[3] is ignored here .
                     ++pPortInstance->dwNumMessagesRcvd;
                     RigCtrl_ParseCIV( pRC, iRigCtrlPort, pPortInstance->bRcvdMsg, pPortInstance->iRcvdMsgLength );
                   } // end if < matching "address" >
                  else // pre- and postamble ok, but something wrong with the addresses
                   { pPortInstance->dwNumGarbageBytes += pPortInstance->iRcvdMsgLength;
                   }
                 } // end if < preamble code ok >
                pPortInstance->iRcvdMsgLength = 0;   // begin receiving next frame
              } // end if < received "end-of-message" code >
             else if( (b==0xFE) && (pPortInstance->bPrevRcvdByte==0xFE) )
              { // "Preamble code (fixed)" :
                // Expect the NEXT byte (at offset 2) to be the
                //        "Transceiver's default address"
                // or the "Controller's default address" .
                //      (more on those "default addresses", including a LIST, in RigCtrl_RadioInfo_CIV[] )
                //
                // The IC-7610 CI-V REFERENCE GUIDE, A7380-7EX-1 available from Icom,
                // clearly states that MORE THEN TWO of these 'preamble codes'
                // may be sent. Example:
                // > When sending the power ON command (18 01), you need
                // > to repeatedly send "FE" before the standard format.
                // (at 115200 bits/second, the receiver expects 150 (!) "FE"s)
                // When Icom's "RS-BA1" is at the other end of a com0com connection
                // (thinking it's talking to an Icom radio but in fact talking to us),
                // got here with a huge pile of 0xFE's before the first "real" message.
                if( pPortInstance->iRcvdMsgLength > 3 ) // oops.. something in between that was NOT 0xFE
                 {  pPortInstance->dwNumGarbageBytes += (pPortInstance->iRcvdMsgLength-3);
                 }
                pPortInstance->iRcvdMsgLength = 2; // discard anything before the "Preamble code"
              }

             break; // end case C_RADIO_CTRL_PROT_ICOM_CI_V
       } // end switch( pRC->iRadioCtrlProtocol )
      // For "decoder" of received messages (from radio, most likely CI-V protocol)

     pPortInstance->bPrevRcvdByte = b;

   }  // end for <all bytes from serial data block>


} // end RigControl_ProcessRxData()




//--------------------------------------------------------------------------
char* RigCtrl_MsgTypeToString(int iMsgType)
{ // Mainly used for dislay in Spectrum Lab's "CAT traffic log".
  // Keep the returned strings short (ideally 7 or 8 characters).
  static char sz10Result[16]; // If it's one of the common types, don't use this evil static buffer !

  switch( iMsgType & RIGCTRL_MSGTYPE_STRIP_FLAGS )
   {
     case RIGCTRL_MSGTYPE_INFO     :
        return "info";
     case RIGCTRL_MSGTYPE_ECHO     :
        return "echo";
     case RIGCTRL_MSGTYPE_RADIO_ID :
        return "radio ID";
     case RIGCTRL_MSGTYPE_FREQUENCY: // "frequency data" (usually with the VFO frequency)
        return "VFO freq";
     case RIGCTRL_MSGTYPE_OP_MODE  : // "mode data" (often unsolicited, "on change")
        return "op mode";
     case RIGCTRL_MSGTYPE_RIT_XIT  : // RIT or XIT data (often unsolicited, "on change")
        return "RIT/XIT";
     case RIGCTRL_MSGTYPE_RF_GAIN  : // anything that affects the RF gain (preamp, attenuator, "AIP", ..)
        return "RF gain";
     case RIGCTRL_MSGTYPE_C_SPECTRUM: // "waveform data" (in fact, for Icom's spectrum scope)
        return "scope";              // "spectrum scope control" (not WAVEFORM DATA)
     case RIGCTRL_MSGTYPE_SPECTRUM : // "waveform data" (in fact, for Icom's spectrum scope)
        return "wform";              // short info because the hex field is long
     case RIGCTRL_MSGTYPE_POWER_ON_OFF:
        return "pwr ctl";
     case RIGCTRL_MSGTYPE_OTHER    : // acknowledge for 'set', etc pp.
        return "other";
     case RIGCTRL_MSGTYPE_OK       : // in CI-V : unspecific "OK", cmd 0xFB
        return "OK";
     case RIGCTRL_MSGTYPE_NOT_OK   : // in CI-V : unspecific "NG" / "Not Good", cmd 0xFA
        return "Not OK";
     case RIGCTRL_MSGTYPE_UDP_CONTROL:
        return "UDP Ctrl";
     case RIGCTRL_MSGTYPE_UDP_SERIAL :
        return "UDP Serial";
     case RIGCTRL_MSGTYPE_UDP_AUDIO:
        return "UDP Audio";
     case RIGCTRL_MSGTYPE_UNKNOWN  : // unknown CI-V or in fact UDP (used by Icom radios via LAN or WLAN)
        return "UNKNWN";
     default:
        sprintf( sz10Result, "?%d?", (int)iMsgType );
        return sz10Result;
   } // end switch( iMsgType )

} // end RigCtrl_MsgTypeToString()

//---------------------------------------------------------------------------
T_RigCtrl_RadioInfo *RigCtrl_GetRadioInfoByDefaultAddress( int iRadioCtrlProtocol, int iDefaultAddress )
{
  T_RigCtrl_RadioInfo *pRadioInfo = NULL;

  // For future extensions (Yaesu CAT protocol ?), there may be
  // different tables, not only the one for radios with Icom-compatible CI-V !
  switch( iRadioCtrlProtocol )
   {
     case RIGCTRL_PROTOCOL_NONE      : // if protocol not set yet, assume it's an Icom
     case RIGCTRL_PROTOCOL_ICOM_CI_V :
        pRadioInfo = (T_RigCtrl_RadioInfo *)RigCtrl_RadioInfo_CIV;
        break;
     case RIGCTRL_PROTOCOL_KENWOOD   : // Sorry folks... Kenwood not supported yet
     case RIGCTRL_PROTOCOL_YAESU_CAT : // Yeasu's "CAT" is a nightmare.. every radio is incompatible
         // (They don't give a shit for compatibility between Yasu radios :
         // > Unfortunately, there is no "master list" of Yaesu CAT commands.
         // > Most Yeasu models have unique command sets; in some cases these
         // > are similar to those of predecessors, in other cases
         // > they don't even use the same bit ordering.
         // > There are also ROM updates that introduce new commands or change
         // > the behaviour of existing commands. Yeasu's CAT documentation
         // > is rarely comprehensive or error free. (..)
         // > My current theory is that Yaesu engineers learned about computers
         // > and software from reading the backs of cereal boxes.
         //     Looks like Dave, AA6YQ, has also had a lot of fun with Yaesu.
         //     Or, as Simon HB9DRV writes in his 'Command Tester' user guide:
         // > Be aware that there are many errors in the Yaesu
         // > documentation, so the data returned may not agree with the
         // > handbook and the radios current settings.
         break;
     case RIGCTRL_PROTOCOL_AD9832_I2C  : // not supported HERE but somewhere else (in SpecLab)
         break;
   } // end  switch( iRadioCtrlProtocol )


  if( pRadioInfo != NULL )
   {
     // The table isn't sorted by anything, so use a linear search:
     while( pRadioInfo->pszName != NULL )  // NULL marks the end of the table
      { if( pRadioInfo->iDefaultAddress == iDefaultAddress )
         {  return pRadioInfo; // bingo, found what we're looking for
         }
        ++pRadioInfo; // check next table entry
      }
   }

  // Arrived here ? Out of luck...
  return NULL;
} // end  RigCtrl_GetRadioInfoByDefaultAddress()

//---------------------------------------------------------------------------
char* RigCtrl_DefaultAddressToString( int iRadioCtrlProtocol, int iDefaultAddress )
{
  static char sz7Result[8];
  T_RigCtrl_RadioInfo *pRadioInfo = RigCtrl_GetRadioInfoByDefaultAddress(
                                          iRadioCtrlProtocol, iDefaultAddress );
  if( pRadioInfo != NULL )
   { return pRadioInfo->pszName;
   }
  else
   { sprintf( sz7Result, "<%02X>", iDefaultAddress );
     return sz7Result;
   }
} // end RigCtrl_DefaultAddressToString()

//---------------------------------------------------------------------------
char* RigCtrl_RigControlPortNrToString( int iRigCtrlPort )
{ switch( iRigCtrlPort )
   { case RIGCTRL_PORT_RADIO    : return "Radio Port";
     case RIGCTRL_PORT_CLIENT_1 : return "Client Port #1";
     case RIGCTRL_PORT_CLIENT_2 : return "Client Port #2";
#ifdef    RIGCTRL_PORT_CLIENT_3
     case RIGCTRL_PORT_CLIENT_3 : return "Client Port #3";
#endif
     default : return "Invalid port";
   }
} // end RigCtrl_RigControlPortNrToString()

//---------------------------------------------------------------------------
char *RigCtrl_ControlConnStateToString( T_RigCtrlInstance *pRC ) // more verbose than RigCtrl_ConnStateToString(), but mainly for the "CONTROL"-UDP-port
  // [in] pRC->iConnState    : RIGCTRL_CONNSTATE_OFF/CONNECTING/CONNECTED
  // [in] pRC->iConnSubstate : RIGCTRL_CONN_SUBSTATE_SEND_CONN_REQUEST, etc^10 ..
  // [return] e.g. "Connected to IC-705" (ideally).
  // Called from TDebugForm::Btn_GetReportOnCommsClick(),
  //    and from TSpectrumLab::MI_StartStopClick() to show the current state in SL's MAIN MENU.
{ static char sz80Result[84];
  char *cp = sz80Result, *pszEndstop = sz80Result+80;

  if( pRC==NULL )
   { return "Deleted";
   }
  switch( pRC->initState )
   { case initState_ColdStart : return "NoInit";
     case initState_Closed    : return "Stopped";  // e.g. RigCtrl_Close() called from the main menu, "Start/Stop".."Rig Control".."Stop" (thus this result)
     default : break;   // e.g. initState = 1 = "Opened" ...
   }

  // If a "real" serial port (RS-232 or single-wire CI-V adapter), or a VIRTUAL COM PORT on USB is used,
  // there's no such complex "Connection" state machine as for UDP (with the "three ports").
  // In this case, we consider ourselves "connected" as soon as we RECEIVED
  // the first properly formatted CI-V message from an Icom radio.
  SL_AppendString( &cp, pszEndstop, RigCtrl_ParameterPollingStateToString(pRC->iParameterPollingState) );
  // ,--------------------------------'
  // '--> e.g. "passive" /  "start reading" / "read VFO freq" / "read op-mode" / "start spectrum" /  "running"
  if( pRC->PortInstance[RIGCTRL_PORT_RADIO].dwNumMessagesRcvd > 0 )
   { SL_AppendPrintf( &cp, pszEndstop, ", rcvd %ld msgs",
            (long)pRC->PortInstance[RIGCTRL_PORT_RADIO].dwNumMessagesRcvd );
   }
  else
   { SL_AppendString( &cp, pszEndstop, ", no RX yet" );
   }
  return sz80Result; // end case INET_PROTOCOL_COM_PORT
} // end RigCtrl_ControlConnStateToString()

//---------------------------------------------------------------------------
char *RigCtrl_ParameterPollingStateToString(int iParameterPollingState)
  // Added 2023-06 when, despite being successfully "connected" to all three
  // UDP ports in an IC-705, none of the essential parameters became valid.
  // Called from, and results displayed in, Spectrum Lab's "CAT traffic monitor".
{ static char sz15Result[16];
  switch( iParameterPollingState )
   { case RIGCTRL_POLLSTATE_PASSIVE             : return "passive";        // initial state: have neither READ nor WRITTEN any setting from/to the remote rig
     case RIGCTRL_POLLSTATE_START_RD            : return "start reading";  // begin reading all we need to know about/from the remote rig
     case RIGCTRL_POLLSTATE_FREQUENCY           : return "read VFO freq";  // (try to) read the current VFO frequency
     case RIGCTRL_POLLSTATE_OPMODE              : return "read op-mode";   // (try to) read the operating mode (USB,LSB,CW,..)
     case RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM  : return "start spectrum"; // (try to) activate transmission of "spectrum scope waveform data"
     case RIGCTRL_POLLSTATE_DONE                : return "running";        // finished polling "all we need to know" about the radio; periodic polling is "running" (S-meter, VFO, Spectrum, ?)
     default : sprintf( sz15Result, "?%d?", (int)iParameterPollingState );
               return sz15Result;
   }
} // end RigCtrl_ParameterPollingStateToString()


//---------------------------------------------------------------------------
// Built-in "traffic log" (circular FIFO for the display in e.g. DebugU1.cpp)
//    Initially used for CI-V via serial port, but in the meantime,
//    can optionally store the first few bytes in any kind of UDP packets,
//    packetized analog samples received from external A/D converters, etc.
//    (Why that ? Certain Icom radios send / receive audio via UDP too,
//     thus module RigControl.c must be able to exchange audio streams
//     with Spectrum Lab anyway. Thus, to develop microcontroller firmware
//     driving external A/D or D/A converters, the former "CAT Traffic
//     Monitor" shall be able to display other kinds of 'packets', too.
//
//  -> file:///C:/cbproj/SpecLab/html/troubles.htm#CAT_traffic_monitor
//
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
CPROT void RigCtrl_ClearTrafficLog( T_RigCtrlInstance *pRC )
{
  pRC->iTrafficLogHeadIndex = 0; // circular FIFO head index (post-incremented, wraps from RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES-1 to zero)
  pRC->fTrafficLogFull  = FALSE; // TRUE when all <RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES> entries are IN USE.
  // (but in most cases, the circular head index 'iTrafficLogHeadIndex' keeps incrementing,
  //  unless RigControl_iTrafficMonitorDisplayOptions.RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_FULL_FIFO is set.)


} // end RigCtrl_ClearTrafficLog()

//---------------------------------------------------------------------------
void RigCtrl_AddToTrafficLog( T_RigCtrlInstance *pRC, // for 'decoded CI-V' messages (*)
         int  iRigCtrlPort, // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 / .. ?
         BYTE *pbMessage, int iMsgLength, // [in] CI-V message as text, including pre- and postamble,
                            //      or just an 'info' about an important internal state change,
                            //      in complete plain text, with a few parameters.
         int  iMsgType,     // [in] e.g. RIGCTRL_MSGTYPE_RADIO_ID | RIGCTRL_MSGTYPE_NO_OK, etc;
                            //      plus some bitwise combineable flags like RIGCTRL_MSGTYPE_FLAG_UDP .
         char *pszComment)  // [in] optional comment (*not* a format string)
  // (*) Since 2021-09, RigCtrl_AddToTrafficLog() is also used to show
  //     a few Icom-specific UDP-"control" messages (with UDP payload only).
  //     The COLOUR used for the display (in a Rich Edit control) is selected
  //     later, in RigCtrl_ReadTrafficLog(), depending on "Message Type" flags like
  //       RIGCTRL_MSGTYPE_ERROR, RIGCTRL_MSGTYPE_FLAG_TX, RIGCTRL_MSGTYPE_FLAG_RX, ...
  //     Examples (copied and pasted from the RichEdit control into HTML *as text*)
  //     are in the Spectrum Lab manual [SpecLab_CAT_Monitor] .
  //
  // Non-CI-V-entries (but UDP) are flagged by RIGCTRL_MSGTYPE_CONTROL_VIA_UDP in iMsgType,
  // and are formatted slightly different than CI-V in RigCtrl_ReadTrafficLog().
{
  T_RigCtrlTrafficLogEntry *pNewEntry;
  double dblUnixTime = UTL_GetCurrentUnixDateAndTime();
  int iHeadIndex = pRC->iTrafficLogHeadIndex % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
  int i,n;

  // Since 2023-07, not only the DISPLAY may be paused, but also the acquisition
  //  of TRAFFIC LOG DATA. Controllable via 'More'-button -> menu in DebugU1.cpp:
  if( RigControl_fTrafficMonitorPaused )  // one of the 'configurable triggers' has PAUSED the traffic log ->
   { return;  // avoid those "important captured data" from being overwritten,
     // while the user scrolls back and forth, and SWITCHES THE DISPLAY MODE,
     // which is only possible if the 'important' entries are still in pRC->TrafficLog[] !
   } // end if( RigControl_fTrafficMonitorPaused )

  pNewEntry = &pRC->TrafficLog[iHeadIndex];
  memset( pNewEntry, 0, sizeof(T_RigCtrlTrafficLogEntry) );  // provides zero-terminated strings
  if( iMsgLength > 0 )  // Only increment the 'message number' when logging REAL PACKETS,
   { ++RigControl_i32MessageNrForTrafficLog; // .. because only 'real packets' are NUMBERED.
     // Info strings like "Opening 'Serial' port: Hostname 'IC-705' resolved into IP 192.168.178.28"
     // are no PACKETS TRAVELLING ON THE NETWORK, and will not have a number in the first column
     // when converted into text in RigCtrl_ReadTrafficLog() .
     // Despite that, if a log entry WITH packet data (i.e. T_RigCtrlTrafficLogEntry.iMsgLength > 0)
     //  is followed by a couple of 'info lines WITHOUT packet data',
     //  those subsequent lines will have the same T_RigCtrlTrafficLogEntry.i32MessageNr .
   }
  pNewEntry->i32MessageNr = RigControl_i32MessageNrForTrafficLog; // unique "message identifier",
    // used in the GUI (DebugU1.cpp) when the operator clicked on a certain line
    // in the RichText control. The "Message Number" (or "No.", as in Wireshark)
    // can be used to retrieve more info than currently visible on the screen,
    // as long as the circular FIFO (

  pNewEntry->dblUnixTime = dblUnixTime;
  n = iMsgLength; // number of "raw message bytes" that can be stored in a log-entry
  if( n>80 )
   {  n=80;
   }
  if( pbMessage != NULL )  // may be NULL for special messages like "Spectrum timeout", etc
   { for( i=0; i<n; ++i )
      { pNewEntry->b80Message[i] = pbMessage[i];  // copy up to 80 bytes of the message data
      }
   }
  pNewEntry->iMsgLength   = iMsgLength;
  pNewEntry->iMsgType     = iMsgType;  // RIGCTRL_MSGTYPE_FLAG_UDP, etc ..
  pNewEntry->iRigCtrlPort = iRigCtrlPort;
  if( pszComment != NULL )
   { SL_strncpy( pNewEntry->szComment, pszComment, RIGCTRL_MAX_COMMENT_LENGTH );
   }

  ++iHeadIndex;
  if( RigControl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_FULL_FIFO )
   { if( iHeadIndex == (RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES-1) )
      { RigControl_fTrafficMonitorPaused = TRUE;
        pRC->fTrafficLogFull = TRUE; // all <RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES> are valid now !
        // In the VERY LAST FIFO entry, leave a note that we stopped because the FIFO is "full":
        pNewEntry = &pRC->TrafficLog[iHeadIndex];
        memset( pNewEntry, 0, sizeof(T_RigCtrlTrafficLogEntry) );  // provides zero-terminated strings
        pNewEntry->i32MessageNr = RigControl_i32MessageNrForTrafficLog; // unique "message identifier",
        pNewEntry->dblUnixTime = dblUnixTime;
        SL_strncpy( pNewEntry->szComment, "traffic monitor paused on 'full FIFO' (as configured)", RIGCTRL_MAX_COMMENT_LENGTH );
        pRC->iTrafficLogHeadIndex = 0;
        RigControl_fUpdateTrafficLog = TRUE; // now "ready for being UPDATED on-screen"
        return;
      }
   } // RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_FULL_FIFO ?
  if( iHeadIndex >= RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES )
   {  iHeadIndex = 0;
      pRC->fTrafficLogFull = TRUE; // all <RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES> are valid now !
      // Because there may be an unknown number of READERS,
      // T_RigCtrlInstance contains no FIFO *TAIL* INDEX itself !
      // An example for a "traffic log display" (reader) is in
      // C:\cbproj\SpecLab\DebugU1.cpp .
   }
  pRC->iTrafficLogHeadIndex = iHeadIndex;
  RigControl_fUpdateTrafficLog = TRUE; // now "ready for being UPDATED on-screen"


} // end RigCtrl_AddToTrafficLog()


//---------------------------------------------------------------------------
int  RigCtrl_GetNumEntriesInTrafficLog( T_RigCtrlInstance *pRC,
             int iTailIndex )     // [in] FIFO tail index
{
  int iHeadIndex = pRC->iTrafficLogHeadIndex % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
  int iOldestEntry, nEntries;
  if( pRC->fTrafficLogFull )
   {  // ALL entries, pRC->TrafficLog[0..RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES-1], are valid.
      iOldestEntry = (iHeadIndex+1) % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
   }
  else // only entries in pRC->TrafficLog[0..iHeadIndex-1] are valid
   {  // Reading would BEGIN at index zero .
      // No more entries are available at index iHeadIndex .
      iOldestEntry = 0;
   }
  if( iTailIndex < 0 ) // caller would read the FIRST entry (when available)
   {  iTailIndex = iOldestEntry;
   }
  else
   {  iTailIndex %= RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
   }
  nEntries = iHeadIndex - iTailIndex;
  if( nEntries < 0 ) // circular wrap..
   {  nEntries += RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
   }
  return nEntries;
} // end RigCtrl_GetNumEntriesInTrafficLog(


//---------------------------------------------------------------------------
BOOL RigCtrl_ReadTrafficLog( // .. as 'human readable text' with colour highlighting
             T_RigCtrlInstance *pRC,   // [in] rig control instance (just in case we support more than one)
             int  *piTrafficTailIndex, // [in,out] FIFO tail index (incremented in each call)
             char *pszDest,            // [out] human readable text, capacity at least <iWantedCharsPerLine+4> bytes
             int  iWantedCharsPerLine, // [in] typically 80 characters per line, but maybe more (see TDebugForm::UpdateCATTrafficLog() )
             int  *piMsgType,          // [out,optional] RIGCTRL_MSGTYPE_FREQUENCY (protocol independent)
             DWORD *pdwColor )         // [out,optional] colour, from MSBIT to LSBIT: BLUE, GREEN, RED (8 bit each)
  // Returns TRUE as long as more entries are available (and the TAIL INDEX is incremented),
  //     or  FALSE when no more entries are available in the CAT message FIFO.
  //
  // Since 2023-07, RigCtrl_ReadTrafficLog(!) may decide HOW TO DISPLAY packets,
  //    depending on RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD  .
  //    This was easy to achieve because if RIGCTRL_MSGTYPE_FLAG_UDP is set in iMsgType,
  //    pbMessage points to THE FIRST BYTE of the *UDP* payload;
  //    not just the CI-V payload (see RigCtrl_SendCIVDataViaUDP() -> RigCtrl_AddToTrafficLog() ).
  //
{
  int iHeadIndex = pRC->iTrafficLogHeadIndex % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
  int iTailIndex = *piTrafficTailIndex;
  int iOldestEntry, i, iType2, n, iSlen, nCharsForMsgTypeAndComment;
  T_RigCtrlTrafficLogEntry *pEntry;
  char *cp = pszDest, *pszMessageTypeAsString;
  char *pszEndstop = pszDest + iWantedCharsPerLine;
  DWORD dwColor;
  BOOL  fHavePacketData; // TRUE : got a log entry with "packet data";  FALSE : "just an INFO or missing data"
  BYTE  *pbSource;

  *cp = '\0';

  if( pRC->fTrafficLogFull )
   {   // ALL entries, pRC->TrafficLog[0..RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES-1], are valid.
       // Reading must BEGIN at iHeadIndex+1 because that's THE OLDEST entry
       //   (==the next to be overwritten in RigCtrl_AddToTrafficLog() .
       // When the tail index reaches iHeadIndex, no more entries are available.
       iOldestEntry = (iHeadIndex+1) % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
   }
  else // only entries in pRC->TrafficLog[0..iHeadIndex-1] are valid
   {   // Reading must BEGIN at index zero .
       // No more entries are available at index iHeadIndex .
       iOldestEntry = 0;
   }
  if( iTailIndex < 0 ) // caller wants to read the FIRST, aka OLDEST entry (when available)
   {  iTailIndex = iOldestEntry;
   }
  else
   {  iTailIndex %= RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
   }
  if( iHeadIndex == iTailIndex )
   { return FALSE;
   }
  else
   { pEntry = &pRC->TrafficLog[iTailIndex];
     *piTrafficTailIndex = (iTailIndex+1) % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
     iType2 = pEntry->iMsgType & RIGCTRL_MSGTYPE_STRIP_FLAGS;

     // Which background colour (that's the easy part...) ?
     if( pEntry->iMsgType & RIGCTRL_MSGTYPE_ERROR ) // e.g. a response-reception-TIMEOUT
      { dwColor = 0x7F00FF; // PINK(ish) .. from MSBIT to LSBIT: BLUE, GREEN, RED (8 bit each)
      }
     else if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_TX ) // this message was SENT (in CI-V speak, from "controller" to "radio")
      { dwColor = 0x00FFFF; // YELLOW .. from MSBIT to LSBIT: BLUE, GREEN, RED (8 bit each)
      }
     else if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_RX ) // this message contains 'real DATA' (UDP or CI-V), and it was RECEIVED
      { // RECEIVED message : light cyan background (unless it's a "NoGo"-response, or some other "special case" :
        if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_UDP )
         { // Added 2021-09 : not only the normal CI-V payload but UDP-payload :
           // If the message type was recognized, the "meaning" of this message
           // is in pEntry->szComment originating from RigControl_ProcessIcomUDP().
           // Show "CI-V less" messages in YELLOW .
           // Otherwise (UDP Control packet from an Icom radio that we DON'T understand),
           // show the message in ORANGE colour. At least, that was the plan...
           if( pEntry->szComment[0] != '\0' ) // ok, RigControl_ProcessIcomUDP() had recognized the received type ->
            { dwColor = 0xC0FF7F; // B,G,R for "aqua"(?) = RECEIVED UDP (not yet reduced to "pure CI-V" for the 'Traffic Display")
            }
           else // not recognized ?
            { dwColor = 0x007FFF; // orange = received UDP, but content not recognized
            }
         }   // RIGCTRL_MSGTYPE_CONTROL_VIA_UDP ?
        else // not an entry with UDP "payload" but most likely a CI-V message:
         {
           dwColor = 0xFFFF00; // 100 % BLUE  + 100 % GREEN + 0 % RED (8 bit each) : lightcyan = RECEIVED CiV (not UDP)
           if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_EXPECTED )
            { dwColor = 0x7FFF7F; // more green-ish than Cyan : "received the EXPECTED response"
            }
         }
      }
     else // none of the above, no "real DATA packet" (neither RX- nor TX-data), so maybe...
      { // .. a 'state transition', like the ones emitted from RigControl_EnterState_Done(),
        // with only the bit 'RIGCTRL_MSGTYPE_OK' set (or not even that) :
        dwColor = 0xC0C0C0; // B,G,R for "lightgray"
      }

     switch( iType2 )
      { case RIGCTRL_MSGTYPE_FREQUENCY: // "frequency data" (usually with the VFO frequency)
        case RIGCTRL_MSGTYPE_OP_MODE  : // "mode data" (often unsolicited, "on change")
        case RIGCTRL_MSGTYPE_RIT_XIT  : // RIT or XIT data (often unsolicited, "on change")
        case RIGCTRL_MSGTYPE_RF_GAIN  : // anything that affects the RF gain (preamp, attenuator, "AIP", ..)
        case RIGCTRL_MSGTYPE_C_SPECTRUM: // configure transmission of spectrum data
        case RIGCTRL_MSGTYPE_SPECTRUM : // "waveform data" (in fact, for Icom's spectrum scope)
        case RIGCTRL_MSGTYPE_POWER_ON_OFF: // power on/off (special case, needs long preamble)
        case RIGCTRL_MSGTYPE_OTHER    : // acknowledge for 'set', etc pp.
        case RIGCTRL_MSGTYPE_OK       : // in CI-V : unspecific "OK", cmd 0xFB
             break;
        case RIGCTRL_MSGTYPE_NOT_OK   : // in CI-V : unspecific "NG" / "Not Good", cmd 0xFA
             dwColor = 0xFF00FF; // colour format, from MSBIT to LSBIT: BLUE, GREEN, RED (8 bit each)
             break;
        case RIGCTRL_MSGTYPE_INFO     : // just an INFO, e.g. about an important state transition, NO PACKET DATA.
             dwColor = 0xC0C0C0; // colour format, from MSBIT to LSBIT: BLUE, GREEN, RED (8 bit each)
             break;
        default:
             break;
      }


     // Check the widths for the optional 'message type indicator' (in parentheses)
     //                  and the optional 'comment' in advance, to limit the
     //                  the number of bytes emitted as hex dump further below.
     if( (pEntry->iMsgType & RIGCTRL_MSGTYPE_STRIP_FLAGS) != RIGCTRL_MSGTYPE_UNKNOWN )
      { pszMessageTypeAsString = RigCtrl_MsgTypeToString( pEntry->iMsgType );
      }
     else
      { pszMessageTypeAsString = "";
      }

     // Begin each line in the "traffic log" with the 'Number'
     //   ("No." as in Wireshark, with leading spaces, and RIGHT-ALIGNED.
     //    Four digits should be sufficient, because 'the trouble usually starts early'):
     if( (pEntry->iMsgLength > 0) && ( iType2 != RIGCTRL_MSGTYPE_INFO) ) // traffic log entry with REAL PACKET DATA ?
      { sprintf( cp, "%4ld", (long)( pEntry->i32MessageNr % 10000 ) ); // Show their "No." in the first column.
        fHavePacketData = TRUE;
      }
     else // traffic log entry WITHOUT real packet data -> leave several columns away further below !
      { fHavePacketData = FALSE;
      }
     SL_PadSpaces( &cp, pszEndstop, 5/*iFixedFieldLength*/ );

     // Next (optional) column: "time of day" in in UTC .. also makes sense for RIGCTRL_MSGTYPE_INFO (without PACKET DATA)
     if( RigControl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_TIME )
      { UTL_FormatDateAndTime( "hh:mm:ss.s", pEntry->dblUnixTime, cp );
        SL_SkipToEndOfString( &cp );
        SL_AppendChar( &cp, pszEndstop, ' ' );  // TEN characters plus one column-separating SPACE
      }

     if( fHavePacketData ) // most columns that follow are only displayed if there are 'packet data' ...
      {

        // Next (optional) column: show "TX" / "RX" (transmit/receive flag; useful if this cannot be indicated by COLOURS)
        if( RigControl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_TX_RX )
         { switch( pEntry->iRigCtrlPort ) // message on the RADIO PORT or one of the "duplicator" ports ?
            { case RIGCTRL_PORT_RADIO :   // message on the port that connects TO THE RADIO...
                 if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_TX )
                  { strcpy( cp, "TX " );
                  }
                 else
                  { strcpy( cp, "RX " );
                  }
                 break;
              default :
                 if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_TX )
                  { sprintf( cp, "t%d ", (int)pEntry->iRigCtrlPort );  // "t1".."t4" : transmitted to the 1st .. 4th "CLIENT PORT"
                  }
                 else
                  { sprintf( cp, "r%d ", (int)pEntry->iRigCtrlPort );  // "r1".."r4" : received from the 1st .. 4th "CLIENT PORT"
                  }
                 break;
            } // end switch( iRigCtrlPort )
           SL_PadSpaces( &cp, pszEndstop, 3/*iFixedFieldLength*/ ); // TWO characters plus one column-separating SPACE
         } // end if < .. RIGCTRL_TRAFFIC_EXTRA_COLUMN_TX_RX > ?

        // Next (optional) column: show PAYLOAD LENGTH (three hex digits without prefix) ?
        if( RigControl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_PAYLOAD_LENGTH )
         { sprintf( cp, "%03X", (int)pEntry->iMsgLength ); // length (3 digits hex, should be ok for up to 4095 bytes from LAN)
           SL_PadSpaces( &cp, pszEndstop, 4/*iFixedFieldLength*/ ); // THREE hex digits plus one column-separating SPACE
         } // end if < .. RIGCTRL_TRAFFIC_EXTRA_COLUMN_PAYLOAD_LENGTH > ?


        // Convert the (mostly binary) entry into human-readable text,
        //   unless the "traffic log entry" contains no data but only a TEXT MESSAGE:
        switch( pEntry->iRigCtrlPort ) // message on the RADIO PORT or one of the "duplicator" ports ?
         { case RIGCTRL_PORT_RADIO :   // message on the port that connects TO THE RADIO...
              // leave dwColor as-is (not "weaker than usual".. see below)
              break;
           default :
              dwColor &= 0xBFBFBF; // make the three colour components a bit (3) darker to indicate a "duplicator client port"
              break;
         } // end switch( iRigCtrlPort )
        n = pEntry->iMsgLength; // number of "raw message bytes" that can be stored in a log-entry
        if( n>80 )
         {  n=80;
         }
        nCharsForMsgTypeAndComment = 0;
        if(  ( pEntry->szComment[0] != '\0' )
          && ( ( RigControl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_COMMENTS ) // show "machine-generated comments, parameter names, etc" ? ...
             ||( ! fHavePacketData ) ) // without "packet data" to show, all we can show is the "comment" or "info line" !
          )
         { nCharsForMsgTypeAndComment += strlen(pEntry->szComment) + 2/*strlen("; ")*/;
         }
        if( pszMessageTypeAsString[0] != '\0' )
         { nCharsForMsgTypeAndComment += strlen(pszMessageTypeAsString) + 3/*strlen("() ")*/;
         }
        if( nCharsForMsgTypeAndComment == 0 )
         { nCharsForMsgTypeAndComment = 32;  // use the 'old, fixed default'
         }

        for( i=0; (i<n) && (cp<(pszDest+iWantedCharsPerLine-nCharsForMsgTypeAndComment-5)); ++i )
         { sprintf( cp, "%02X ", (int)pEntry->b80Message[i] ); // show data (CI-V or UDP) as 2-digit HEX
           cp += strlen(cp);
         }
        if( i<n ) // couldn't dump everything (e.g. with waveform data) ?
         { *cp++ = '.';
           *cp++ = '.'; // marker for omitted hex bytes
           *cp++ = ' ';
         }
        while( cp<(pszDest+iWantedCharsPerLine-nCharsForMsgTypeAndComment) ) // pad spaces for short messages, incl. VFO frequency
         { *cp++ = ' '; // (for rare long messages, columns may be misaligned but that's ok)
         }
        *cp = '\0';
        if( pszMessageTypeAsString[0] != '\0' )
         { sprintf( cp, "(%s) ", pszMessageTypeAsString );
           cp += strlen(cp);
         }
        iSlen = strlen(pEntry->szComment);
        if( (iSlen>0) && ( (cp+iSlen)<(pszDest+iWantedCharsPerLine) ) )
         {
           // if there's enough space, pad up to the next column
           while( (cp<(pszDest+iWantedCharsPerLine-iSlen-2) )
               && (cp<(pszDest+iWantedCharsPerLine-16) )  )
            { *cp++ = ' ';
            }
           sprintf( cp, "; %s", pEntry->szComment );
           cp += strlen(cp);
         } // end if < append the COMMENT > ?
      } // end if( fHavePacketData )
     else // "coloured text" only, no message data:
      { sprintf( cp, " %s", pEntry->szComment );
        cp += strlen(cp);
      } // end else < !fHavePacketData >
     while( cp<(pszDest+iWantedCharsPerLine) )  // for the coloured background, pad even more spaces..
      { *cp++ = ' '; // formerly 80 chars per line, now depends on width of the window
      }
     *cp = '\0';

     if( piMsgType != NULL )
      { *piMsgType = pEntry->iMsgType; // [out,optional] RIGCTRL_MSGTYPE_FREQUENCY (protocol independent)
      }
     if( pdwColor != NULL )
      { *pdwColor = dwColor;
      }
     return TRUE;
   }
} // end RigCtrl_ReadTrafficLog()


//---------------------------------------------------------------------------
void RigCtrl_GetHeadlineForTrafficLog(int iMsgType, char *pszDest, int iWantedCharsPerLine)
  // Generates a 'headline' to simplify interpreting the (hex) message payload
  // in "Icom UDP" packets, any maybe other message types.
  // Occasionally called from TDebugForm::UpdateCATTrafficLog() (!) .
  // The FORMAT must match the strings assembled by RigCtrl_ReadTrafficLog().
  //
  //  [in] iMsgType : bit combination including flags like  ..
  //       RIGCTRL_MSGTYPE_FLAG_UDP : UDP (datagram, usually Icom RS-BA compatible)
  //  [out] pszDest : capactiy should be iWantedCharsPerLine + 4 (!) chars or more.
{
  int byte_index;
  char *cp         = pszDest;
  char *pszEndstop = pszDest + iWantedCharsPerLine; // <- points to the STRING TERMINATOR
  int  nCharsRemaining;


  // First (non-optional) column: "No. " (with FOUR digits plus a column-separating SPACE) ..
  SL_AppendString( &cp, pszEndstop, "; Nr " );

  // Next (optional) column: "time of day" in in UTC ..
  if( RigControl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_TIME )
   { SL_AppendString( &cp, pszEndstop, "hh:mm:ss.s " );
   }

  // Next (optional) column:  header for the "TX" / "RX" indicator
  if( RigControl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_TX_RX )
   { SL_AppendString( &cp, pszEndstop, "tr " );
   }

  // Next (optional) column: packet length (3 digits hex)
  if( RigControl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_PAYLOAD_LENGTH )
   { SL_AppendString( &cp, pszEndstop, "len " );
   } // end if < .. RIGCTRL_TRAFFIC_EXTRA_COLUMN_PAYLOAD_LENGTH > ?

  // Next (NON-OPTIONAL) column : "legend for the hex dump"
  if( iMsgType & RIGCTRL_MSGTYPE_FLAG_UDP ) // line in the traffic log with *UDP payload* :
   {
     // Example (line with UDP hex payload emitted by RigCtrl_ReadTrafficLog() ) :
     //                                 10 00 00 00 03 00 00 00 D7 E9 00 00 00 00 00 00
     SL_AppendString( &cp, pszEndstop, "|-IcomLen-| |typ| |seq| |-sent_id-| |-rcvd_id-|" );
     nCharsRemaining = pszEndstop - cp;
     if( nCharsRemaining > 3 )  // beginning at this character-index,
      { // just add HEXADECIMAL BYTE INDICES, as used in the RS-BA-compatible
        // structure definitions in RigControl.h :
        cp = pszDest + strlen(pszDest);
        byte_index = 0x10;
        while( (cp+2)<pszEndstop )
         { sprintf( cp, " %02X", (int)(byte_index++) );
           SL_SkipToEndOfString( &cp );
         }
      }
   }
  else if(iMsgType & RIGCTRL_MSGTYPE_FLAG_CIV ) // only CI-V, without overhead from any "transport layer"
   { //               byte-index----------> 00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F 10 11 12 13 14 15 16 17 ...
     //               3-digit LENGTH--->010 FE FE A4 E0 27 .. .. .. .. .. .. .. FD
     SL_AppendString( &cp, pszEndstop, "CiV |pre| to fm parameters..  postamble=FD   " );
   }
  else
   { SL_AppendString( &cp, pszEndstop, "len generic hex data, type unknown" );
   }
  // Pad with spaces for the wanted length (just in case the headline has a coloured background) :
  cp = pszDest + SL_strnlen( pszDest, iWantedCharsPerLine );
  while(cp<pszEndstop)
   { *cp++ = ' ';
   }
  *pszEndstop = '\0';  // <- one of the reasons why the capacity of pszDest must exceed iWantedCharsPerLine (by at least ONE BYTE)

} // end RigCtrl_GetHeadlineForTrafficLog()


//---------------------------------------------------------------------------
// Lookup tables and similar ...
//---------------------------------------------------------------------------

const T_RigCtrl_RadioInfo RigCtrl_RadioInfo_CIV[] =
{ // { iDefaultAddress,            pszName,        iCapabilities (if known)

  { RIGCTRL_DEF_ADDR_IC_271      , "IC-271"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_275      , "IC-275"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_375      , "IC-375"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_471      , "IC-471"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_475      , "IC-475"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_575      , "IC-575"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_703      , "IC-703"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_706      , "IC-706"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_706MkII  , "IC-706MkII"  , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_706MkIIG , "IC-706MkIIG" , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_707      , "IC-707"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_718      , "IC-718"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_725      , "IC-725"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_726      , "IC-726"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_728      , "IC-728"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_729      , "IC-729"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_735      , "IC-735"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_736      , "IC-736"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_737      , "IC-737"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_738      , "IC-738"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_746      , "IC-746"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_751A     , "IC-751A"     , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_756      , "IC-756"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_756Pro   , "IC-756Pro"   , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_756ProII , "IC-756ProII" , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_756ProIII, "IC-756ProIII", 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_761      , "IC-761"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_765      , "IC-765"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_775      , "IC-775"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_78       , "IC-78"       , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_7000     , "IC-7000"     , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_7100     , "IC-7100"     , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_7200     , "IC-7200"     , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_7300     , "IC-7300"  , RIGCTRL_CAPS_SPECTRUM_VIA_CAT
                                              | RIGCTRL_CAPS_HAS_NO_IQ_OUTPUT },
  { RIGCTRL_DEF_ADDR_IC_7400     , "IC-7400"     , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_7600     , "IC-7600"     , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_7610     , "IC-7610"  , RIGCTRL_CAPS_SPECTRUM_VIA_CAT
                                              | RIGCTRL_CAPS_HAS_IQ_OUTPUT    },
  { RIGCTRL_DEF_ADDR_IC_7700     , "IC-7700"     , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_7800     , "IC-7800"     , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_781      , "IC-781"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_820      , "IC-820"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_821      , "IC-821"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_910      , "IC-910"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_970      , "IC-970"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_9700     , "IC-9700"  , RIGCTRL_CAPS_SPECTRUM_VIA_CAT
                                              | RIGCTRL_CAPS_HAS_NO_IQ_OUTPUT },
  { RIGCTRL_DEF_ADDR_IC_705      , "IC-705"   , RIGCTRL_CAPS_SPECTRUM_VIA_CAT },
  { RIGCTRL_DEF_ADDR_IC_905      , "IC-905"   , RIGCTRL_CAPS_SPECTRUM_VIA_CAT },
  { RIGCTRL_DEF_ADDR_IC_1271     , "IC-1271"     , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_1275     , "IC-1275"     , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_R10      , "IC-R10"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_R20      , "IC-R20"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_R71      , "IC-R71"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_R72      , "IC-R72"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_R75      , "IC-R75"      , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_R7000    , "IC-R7000"    , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_R7100    , "IC-R7100"    , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_R8500    , "IC-R8500"    , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_R8600    , "IC-R8600" , RIGCTRL_CAPS_SPECTRUM_VIA_CAT },
  { RIGCTRL_DEF_ADDR_IC_R9000    , "IC-R9000"    , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_R9500    , "IC-R9500"    , 0 /*unknown capabilities*/ },
  { RIGCTRL_DEF_ADDR_IC_RX7      , "IC-RX7"      , 0 /*unknown capabilities*/ },
  { 0, NULL, 0 } // <- and entry with "all zeroes" marks the end of the table
}; // end RigCtrl_RadioInfo_CIV[]


const T_RigCtrl_ParamInfo RigCtrl_ParameterInfo[] =
{ // iUnifiedPN,                   pszToken,              pszUnit,
  //  iByteOffsetIntoRigCtrlInstance            , iRigCtrlDataType, iAccessFlags:

 { RIGCTRL_PN_FREQUENCY,           "VFOfreq",             "Hz",
      offsetof( T_RigCtrlInstance, dblVfoFrequency), RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_OP_MODE,             "OpMode",              "",
      offsetof( T_RigCtrlInstance, iOpMode),      RIGCTRL_DT_INT,  RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_FILTER_BANDWIDTH,    "FilterBW",            "Hz",
      offsetof( T_RigCtrlInstance, iFilterBW_Hz), RIGCTRL_DT_INT,  RIGCTRL_AF_READWRITE},
 { RIGCTRL_PN_DATA_MODE,           "DataMode",            "",
      offsetof( T_RigCtrlInstance, iDataMode),     RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_TRANSMITTING,        "transmitting",        "",
      offsetof( T_RigCtrlInstance, iTransmitting), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_ATU_ENABLED,         "ATU_enabled",         "",
      offsetof( T_RigCtrlInstance, iATU_enabled ), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_XFC_PRESSED,         "XFC_pressed",         "",
      offsetof( T_RigCtrlInstance, iXFC_pressed ), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_TRANSMIT_FREQ,       "TX_freq",             "Hz",
      offsetof( T_RigCtrlInstance, iTX_frequency), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_REPEATER_TONE_FREQ,  "RptrTone",            "Hz",
      offsetof( T_RigCtrlInstance, dblRepeaterTone_Hz),RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_TONE_SQUELCH_FREQ,   "ToneSquelch",         "Hz",
      offsetof( T_RigCtrlInstance, dblToneSquelch_Hz), RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE },


 // --- SPECTRUM SCOPE related (with CI-V, exchanged via command 0x27) ------>
 { RIGCTRL_PN_SCOPE_MODE         , "ScopeMode",           "",
      offsetof( T_RigCtrlInstance, iScopeMode),    RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SCOPE_SPAN         , "ScopeSpan",           "Hz",
      offsetof( T_RigCtrlInstance, dblScopeSpan_Hz), RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SCOPE_ATTEN        , "ScopeAtt",            "dB",
      offsetof( T_RigCtrlInstance, dblScopeAttenuator_dB), RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SCOPE_REF          , "ScopeRef",            "dB",
      offsetof( T_RigCtrlInstance, dblScopeRefLevel_dB), RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SCOPE_EDGE_NR      , "ScopeEdge",           "",
      offsetof( T_RigCtrlInstance, iScopeEdgeNumber), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SCOPE_HOLD_FLAG    , "ScopeHold",           "",
      offsetof( T_RigCtrlInstance, iScopeHoldFlag), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SCOPE_SPEED        , "ScopeSpeed",          "",
      offsetof( T_RigCtrlInstance, iScopeSpeed), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 //--- Various "poti" settings (with CI-V, exchanged via command 0x14) ------>
 { RIGCTRL_PN_AUDIO_VOLUME_PERCENT, "AudioVolume",        "%",
      offsetof( T_RigCtrlInstance, iAudioVolume_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_RF_GAIN_PERCENT    , "RFGain",              "%",
      offsetof( T_RigCtrlInstance, iRFGain_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SQUELCH_LEVEL_PERCENT, "SquelchLevelSet",   "%",
      offsetof( T_RigCtrlInstance, iSquelchLevel_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_NOISE_REDUCTION_PERCENT, "NoiseReduction",  "%",
      offsetof( T_RigCtrlInstance, iNoiseReduction_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_PASSBAND_TUNING_POS1, "PassbandT1",         "%", // "inner ring" of "TWIN PBT", 50 % = center
      offsetof( T_RigCtrlInstance, iPassbandTuningPos1), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_PASSBAND_TUNING_POS2, "PassbandT2",         "%", // "outer ring" of "TWIN PBT", 50 % = center
      offsetof( T_RigCtrlInstance, iPassbandTuningPos2), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_CW_PITCH_HZ        , "CW_Pitch",            "Hz",
      offsetof( T_RigCtrlInstance, iCWPitch_Hz), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_RF_POWER_SETTING_PERCENT, "PowerSet",       "%", // "setting" to avoid confusion with the POWER METER MEASUREMENT !
      offsetof( T_RigCtrlInstance, iRFPowerSetting_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_MIC_GAIN_PERCENT   , "MicGain",             "%",
      offsetof( T_RigCtrlInstance, iMicGain_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_KEYER_SPEED_PERCENT, "KeyerSpeed",          "%", // 0 = often 6 WPM, 100 % = 48 WPM (at least for Icom)
      offsetof( T_RigCtrlInstance, iKeyerSpeed_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_NOTCH_POS_PERCENT  , "NotchPos",            "%",
      offsetof( T_RigCtrlInstance, iNotchPos_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_COMP_SETTING_PERCENT, "CompSet",            "%", // audio compressor SETTING (not a "measurement")
      offsetof( T_RigCtrlInstance, iCompSetting_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_BREAK_IN_DELAY_PERCENT, "BkInDelay",        "%",
      offsetof( T_RigCtrlInstance, iBreakInDelay_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_NBL_SETTING_PERCENT, "NoiseBlanker",        "%",
      offsetof( T_RigCtrlInstance, iNoiseBlankerSetting_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_MONITOR_GAIN_PERCENT, "MonitorGain",        "%",
      offsetof( T_RigCtrlInstance, iMonitorGain_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_VOX_GAIN_PERCENT   , "VoxGain",             "%",
      offsetof( T_RigCtrlInstance, iVoxGain_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_ANTI_VOX_GAIN_PERCENT, "AntiVoxGain",       "%",
      offsetof( T_RigCtrlInstance, iAntiVoxGain_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_BRIGHTNESS_PERCENT , "Brightness",          "%",
      offsetof( T_RigCtrlInstance, iBrightness_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 //------ Various "meter" readings and related statuses --------------------->
 { RIGCTRL_PN_SQUELCH_STATUS     , "SquelchStatus",       "",
      offsetof( T_RigCtrlInstance, iSquelchStatus), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY },
 { RIGCTRL_PN_S_METER_LEVEL_DB   , "SMeterLevel",         "dB", // here: "decibel over S0",  not the strange CI-V internal unit
      offsetof( T_RigCtrlInstance, iSMeterLevel_dB), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY },
 { RIGCTRL_PN_POWER_METER_PERCENT, "PowerMeter",          "%",
      offsetof( T_RigCtrlInstance, iPowerMeterLevel_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY },
 { RIGCTRL_PN_SWR_METER_LEVEL    , "SWRMeter",            "%", // unit: "percent of the TX's tolerance", Icom: 100% = SWR 3.0
      offsetof( T_RigCtrlInstance, iSWRMeterLevel), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY },
 { RIGCTRL_PN_ALC_METER_LEVEL    , "ALCMeter",            "%",
      offsetof( T_RigCtrlInstance, iALCMeterLevel), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY },
 { RIGCTRL_PN_COMP_METER_LEVEL_DB, "CompMeter",           "dB", // decibel, 0=no compression, 30 dB = maximum reading for IC-7300)
      offsetof( T_RigCtrlInstance, iCompMeterLevel_dB), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY },
 { RIGCTRL_PN_SUPPLY_VOLTATE_mV  , "SupplyVoltage",       "mV",
      offsetof( T_RigCtrlInstance, iSupplyVoltage_mV), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY },
 { RIGCTRL_PN_DRAIN_CURRENT_mA   , "DrainCurrent",        "mA",
      offsetof( T_RigCtrlInstance, iDrainCurrent_mA), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY },

 //-- extended Preamp, AGC speed, Noise blanker, Audio peak filter,
 //   noise reduction, auto notch, etc. With CI-V, via command 0x16) -----
 { RIGCTRL_PN_PREAMP_SETTING,      "PreampSet",           "", // CI-V : 0x16 0x02
      offsetof( T_RigCtrlInstance, iPreampSet), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_AGC_SPEED,           "AGCSpeed",            "", // CI-V : 0x16 0x12
      offsetof( T_RigCtrlInstance, iAGCSpeed), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_NOISE_BLANKER_ON,    "NoiseBlankerON",      "", // CI-V : 0x16 0x22
      offsetof( T_RigCtrlInstance, iNoiseBlankerON),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_NOISE_REDUCT_ON,     "NoiseReductON",       "", // CI-V : 0x16 0x40
      offsetof( T_RigCtrlInstance, iNoiseReductON),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_AUTO_NOTCH_ON,       "AutoNotchON",         "", // CI-V : 0x16 0x41
      offsetof( T_RigCtrlInstance, iAutoNotchON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_RPTR_TONE_ON,        "RptrToneON",          "", // CI-V : 0x16 0x42
      offsetof( T_RigCtrlInstance, iRptrToneON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_TONE_SQUELCH_ON,     "ToneSquelchON",       "", // CI-V : 0x16 0x43
      offsetof( T_RigCtrlInstance, iToneSquelchON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_COMP_ON,             "CompressON",          "", // CI-V : 0x16 0x44
      offsetof( T_RigCtrlInstance, iCompressON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_MONITOR_ON,          "MonitorON",           "", // CI-V : 0x16 0x45
      offsetof( T_RigCtrlInstance, iMonitorON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_VOX_ON,              "VoxON",               "", // CI-V : 0x16 0x46
      offsetof( T_RigCtrlInstance, iVoxON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_BK_IN_ON,            "BreakInON",           "", // CI-V : 0x16 0x47
      offsetof( T_RigCtrlInstance, iBreakInON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_MANUAL_NOTCH_ON,     "ManualNotchON",       "", // CI-V : 0x16 0x48
      offsetof( T_RigCtrlInstance, iManualNotchON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_TWIN_PEAK_FILTER_ON, "TwinPeakFilterON",    "", // CI-V : 0x16 0x4F
      offsetof( T_RigCtrlInstance, iTwinPeakFilterON),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_DIAL_LOCK_ON,        "DialLockON",          "", // CI-V : 0x16 0x50
      offsetof( T_RigCtrlInstance, iDialLockON), RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_DSP_SHARP_SOFT,      "DSP_SharpSoft",       "", // CI-V : 0x16 0x56
      offsetof( T_RigCtrlInstance, iDSP_SharpSoft),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_MANUAL_NOTCH_WIDTH,  "ManualNotchWidth",    "", // CI-V : 0x16 0x57
      offsetof( T_RigCtrlInstance, iManualNotchWidth),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SSB_TX_BANDWIDTH,    "SSB_Tx_BW",           "", // CI-V : 0x16 0x58
      offsetof( T_RigCtrlInstance, iSSB_Tx_BW),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE },


 //-- "Special" settings, audio equalizer, connector settings, etc ---------->
 { RIGCTRL_PN_SSB_RX_HPF_LPF,      "SSB_RX_HPF_LPF",      "",
      offsetof( T_RigCtrlInstance, iSSB_RX_HPF_LPF), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SSB_RX_BASS_LEVEL,   "SSB_RX_BASS_LEVEL",   "",
      offsetof( T_RigCtrlInstance, iSSB_RX_BASS_LEVEL), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SSB_RX_TREBLE_LEVEL, "SSB_RX_TREBLE_LEVEL", "",
      offsetof( T_RigCtrlInstance, iSSB_RX_TREBLE_LEVEL), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SSB_CW_SYNC_TUNING,  "SSB_CW_sync_tuning",  "",
      offsetof( T_RigCtrlInstance, iSSB_CW_sync_tuning), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_CALIBRATION_MARKER,  "calib_marker",        "",
      offsetof( T_RigCtrlInstance, iCalibMarker), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_AF_OUTPUT_LEVEL_TO_ACC, "AF2ACC_level",     "%",
      offsetof( T_RigCtrlInstance,    iAF2ACC_level), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SQUELCH_FOR_AF_ON_ACC, "squelch_for_ACC",   "%",
      offsetof( T_RigCtrlInstance, iSquelch_for_ACC), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_BEEP_AND_SPEECH_ON_ACC, "beep_on_ACC",      "",
      offsetof( T_RigCtrlInstance, iBeep_on_ACC), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_IF_OUTPUT_LEVEL_TO_ACC, "IF_level_on_ACC",  "%",
      offsetof( T_RigCtrlInstance, iIF_level_on_ACC), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_MOD_INPUT_LEVEL_FROM_ACC, "MOD_level_ACC",  "%",
      offsetof( T_RigCtrlInstance, iMOD_level_ACC), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_MOD_INPUT_LEVEL_FROM_USB, "MOD_level_USB",  "%",
      offsetof( T_RigCtrlInstance, iMOD_level_USB), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_MOD_IN_CONN_FOR_DATA_OFF, "MOD_conn_NoData","",
      offsetof( T_RigCtrlInstance, iMOD_conn_NoData), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_MOD_IN_CONN_FOR_DATA, "MOD_conn_Data",      "",
      offsetof( T_RigCtrlInstance, iMOD_conn_Data), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_CIV_TRANSCEIVE,      "CIV_transceive",      "",
      offsetof( T_RigCtrlInstance, iCIV_transceive), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_CIV_ADDR_USB_TO_REMOTE, "CIV_addr_USB_Rem", "",
      offsetof( T_RigCtrlInstance,    iCIV_addr_USB_Rem), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_CIV_OUTPUT_FOR_ANT,  "CIV_for_ANT",         "",
      offsetof( T_RigCtrlInstance, iCIV_for_ANT), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_CIV_UNLINK_FROM_REMOTE,"CIV_unlink_Rem",    "",
      offsetof( T_RigCtrlInstance, iCIV_unlink_Remote), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_CIV_USB_ECHO,        "CIV_USB_echo",        "",
      offsetof( T_RigCtrlInstance, iCIV_USB_echo), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_IP_PLUS,             "IP_Plus",             "",
      offsetof( T_RigCtrlInstance, iIP_Plus),      RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_NTP_SERVER_ACCESS,   "NTP_access",          "",
      offsetof( T_RigCtrlInstance, iNTP_Server_Access), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_NTP_ACCESS_RESULT,   "NTP_result",          "",
      offsetof( T_RigCtrlInstance, iNTP_Access_Result), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_AF_MUTE,             "AF_mute",             "",
      offsetof( T_RigCtrlInstance, iAF_mute),      RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 //-- Even more settings (in CI-V, controlled via command 0x0E..0x10) --
 { RIGCTRL_PN_SCAN_MODE,           "ScanMode",            "",
      offsetof( T_RigCtrlInstance, iScanMode),     RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_SPLIT_MODE,          "SplitMode",           "",
      offsetof( T_RigCtrlInstance, iSplitMode),    RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_TUNING_STEP,         "TuningStep",          "Hz",
      offsetof( T_RigCtrlInstance, iTuningStep_Hz),RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },
 { RIGCTRL_PN_ATTENUATOR_DB,       "Attenuator",          "dB",
      offsetof( T_RigCtrlInstance, iAttenuator_dB),RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE },

 // As usual, the end of the table is marked by an entry with "all zeroes" :
 { 0, NULL, 0, 0, 0 }

}; // end RigCtrl_ParameterInfo[]


/* EOF < C:\cbproj\Remote_CW_Keyer\RigControl.c > */


